using System;
using System.Windows.Forms;

using Microsoft.Xna.Framework;
using Microsoft.Xna.Framework.Graphics;

namespace AnimatedTextureGenerator
{
    public class AnimatedTextureGenerator
    {
        private GraphicsDevice device;

        public AnimatedTextureGenerator()
        {
            Form form = new Form();
            this.device = new GraphicsDevice(
                GraphicsAdapter.DefaultAdapter,
                DeviceType.Hardware,
                form.Handle,
                CreateOptions.HardwareVertexProcessing,
                new PresentationParameters()
                );
        }

        public Texture3D Generate()
        {
            //Texture3D smokeTexture = Texture3D.FromFile(this.device, "Content/Smoke.dds");
            //Color[] smokeColors = new Color[smokeTexture.Width * smokeTexture.Height * smokeTexture.Depth];
            //smokeTexture.GetData(smokeColors);

            Texture2D sourceTexture = Texture2D.FromFile(this.device, "Content/Explosion2.png");
            Color[] sourceColors = new Color[sourceTexture.Width * sourceTexture.Height];
            sourceTexture.GetData(sourceColors);

            Texture3D turbulenceTexture = Texture3D.FromFile(this.device, "Content/Turbulence128_2.dds");
            Vector4[] turbulenceColors =
                new Vector4[turbulenceTexture.Width * turbulenceTexture.Height * turbulenceTexture.Depth];
            turbulenceTexture.GetData(turbulenceColors);
            
            int width = sourceTexture.Width;
            int height = sourceTexture.Height;
            int depth = turbulenceTexture.Depth;

            Color[] newSmokeColors = new Color[width * height * depth];
            for (int z = 0; z < depth; z++) {
                for (int y = 0; y < height; y++) {
                    for (int x = 0; x < width; x++) {
                        Color oldSmokeColor = sourceColors[y * width + x];
                        
                        float turbulenceColor1 = GetColorAt(turbulenceColors, height, width, x, y, z).X;
                        float turbulenceColor2 = GetColorAt(turbulenceColors, height, width, x-1, y, z).X;
                        float turbulenceColor3 = GetColorAt(turbulenceColors, height, width, x+1, y, z).X;
                        float turbulenceColor4 = GetColorAt(turbulenceColors, height, width, x, y-1, z).X;
                        float turbulenceColor5 = GetColorAt(turbulenceColors, height, width, x, y+1, z).X;

                        float turbulenceColor =
                            CalulcateMeanValue(
                                turbulenceColor1,
                                turbulenceColor2,
                                turbulenceColor3,
                                turbulenceColor4,
                                turbulenceColor5);

                        byte r = (byte) (oldSmokeColor.R * turbulenceColor);
                        byte g = (byte) (oldSmokeColor.G * turbulenceColor);
                        byte b = (byte) (oldSmokeColor.B * turbulenceColor);
                        byte a = (byte) (oldSmokeColor.A * turbulenceColor);
                        Color newColor = new Color(r, g, b, a);

                        newSmokeColors[z * width * height + y * width + x] = newColor;
                    }
                }
            }

            Texture3D newSmokeTexture = new Texture3D(
                this.device,
                sourceTexture.Width,
                sourceTexture.Height,
                turbulenceTexture.Depth,
                1,
                ResourceUsage.None,
                SurfaceFormat.Color
                );
            newSmokeTexture.SetData(newSmokeColors);

            return newSmokeTexture;
        }

        public Texture3D Normalize(Texture3D texture)
        {
            int width = texture.Width;
            int height = texture.Height;
            int depth = texture.Depth;

            Color[] colors = new Color[width * height * depth];
            texture.GetData(colors);

            float maxColor = 0;
            float maxAlpha = 0;

            for (int z = 0; z < depth; z++) {
                for (int y = 0; y < height; y++) {
                    for (int x = 0; x < width; x++) {
                        Color oldColor = colors[z * width * height + y * width + x];

                        float r = oldColor.R / 255f;
                        float g = oldColor.G / 255f;
                        float b = oldColor.B / 255f;
                        float a = oldColor.A / 255f;

                        maxColor = Math.Max(maxColor, Math.Max(r, Math.Max(g, b)));
                        maxAlpha = Math.Max(maxAlpha, a);
                    }
                }
            }

            for (int z = 0; z < depth; z++) {
                for (int y = 0; y < height; y++) {
                    for (int x = 0; x < width; x++) {
                        Color oldColor = colors[z * width * height + y * width + x];

                        float r = oldColor.R / 255f;
                        float g = oldColor.G / 255f;
                        float b = oldColor.B / 255f;
                        float a = oldColor.A / 255f;

                        r /= maxColor;
                        g /= maxColor;
                        b /= maxColor;
                        a /= maxAlpha;

                        colors[z * width * height + y * width + x] =
                            new Color(
                                (byte) (r * 255),
                                (byte) (g * 255),
                                (byte) (b * 255),
                                (byte) (a * 255));
                    }
                }
            }
            
            Texture3D newTexture = new Texture3D(
                this.device, width, height, depth, 1, ResourceUsage.None, SurfaceFormat.Color
                );
            newTexture.SetData(colors);

            return newTexture;
        }

        private static Vector4 GetColorAt(Vector4[] turbulenceColors, int height, int width, int x, int y, int z) {
            x = Math.Max(x, 0);
            x = Math.Min(x, width - 1);
            
            y = Math.Max(y, 0);
            y = Math.Min(y, height - 1);

            z = Math.Max(z, 0);
            
            return turbulenceColors[z * width * height + y * width + x];
        }


        private static float CalulcateMeanValue(params float[] values)
        {
            if (values.Length == 0) {
                return 0f;
            }

            float meanValue = 0f;
            foreach (float value in values) {
                meanValue += value;
            }
            meanValue /= values.Length;
            return meanValue;
        }
    }
}
