#include "stdafx.h"

#include "DXUT.h"
#include "DXUTmisc.h"

#include "MainFrame.h"

#include <iostream>
#include <ostream>

//--------------------------------------------------------------------------------------
// Forward declarations 
//--------------------------------------------------------------------------------------
bool    CALLBACK IsD3D10DeviceAcceptable( UINT Adapter, UINT Output, D3D10_DRIVER_TYPE DeviceType, DXGI_FORMAT BufferFormat, bool bWindowed, void* pUserContext );
HRESULT CALLBACK OnD3D10CreateDevice( ID3D10Device* pd3dDevice, const DXGI_SURFACE_DESC* pBufferSurfaceDesc, void* pUserContext );
HRESULT CALLBACK OnD3D10ResizedSwapChain( ID3D10Device* pd3dDevice, IDXGISwapChain *pSwapChain, const DXGI_SURFACE_DESC* pBufferSurfaceDesc, void* pUserContext );
void    CALLBACK OnD3D10FrameRender( ID3D10Device* pd3dDevice, double fTime, float fElapsedTime, void* pUserContext );
void    CALLBACK OnD3D10ReleasingSwapChain( void* pUserContext );
void    CALLBACK OnD3D10DestroyDevice( void* pUserContext );

LRESULT CALLBACK MsgProc( HWND hWnd, UINT uMsg, WPARAM wParam, LPARAM lParam, bool* pbNoFurtherProcessing, void* pUserContext );
void    CALLBACK OnKeyboard( UINT nChar, bool bKeyDown, bool bAltDown, void* pUserContext );
void    CALLBACK OnFrameMove( double fTime, float fElapsedTime, void* pUserContext );
bool    CALLBACK ModifyDeviceSettings( DXUTDeviceSettings* pDeviceSettings, void* pUserContext );


//--------------------------------------------------------------------------------------
// Entry point to the program. Initializes everything and goes into a message processing 
// loop. Idle time is used to render the scene.
//--------------------------------------------------------------------------------------
int WINAPI wWinMain( HINSTANCE hInstance, HINSTANCE hPrevInstance, LPWSTR lpCmdLine, int nCmdShow )
{
    // Enable run-time memory check for debug builds.
#if defined(DEBUG) | defined(_DEBUG)
    _CrtSetDbgFlag( _CRTDBG_ALLOC_MEM_DF | _CRTDBG_LEAK_CHECK_DF );
#endif


	MainFrame* mainFrame = MainFrame::Instance();


    // Set DXUT callbacks
    DXUTSetCallbackD3D10DeviceAcceptable( IsD3D10DeviceAcceptable );
    DXUTSetCallbackD3D10DeviceCreated( OnD3D10CreateDevice, mainFrame);
    DXUTSetCallbackD3D10SwapChainResized( OnD3D10ResizedSwapChain, mainFrame );
    DXUTSetCallbackD3D10SwapChainReleasing( OnD3D10ReleasingSwapChain, mainFrame );
    DXUTSetCallbackD3D10DeviceDestroyed( OnD3D10DestroyDevice, mainFrame );
    DXUTSetCallbackD3D10FrameRender( OnD3D10FrameRender, mainFrame );

    DXUTSetCallbackMsgProc( MsgProc, mainFrame );
    DXUTSetCallbackKeyboard( OnKeyboard, mainFrame );
    DXUTSetCallbackFrameMove( OnFrameMove, mainFrame );
    DXUTSetCallbackDeviceChanging( ModifyDeviceSettings, mainFrame );

	DXUTInit( true, true, NULL ); // Parse the command line, show msgboxes on error, no extra command line params
    DXUTSetCursorSettings( true, true ); // Show the cursor and clip it when in full screen
    DXUTCreateWindow( L"Playing along with D3D10" );
    DXUTCreateDevice( true, 640, 480 );

	try {
	
	DXUTMainLoop(); // Enter into the DXUT render loop
	} catch(std::exception e) {
		MessageBox( NULL, L"Exception", L"Error", MB_OK );		
	}
    return DXUTGetExitCode();
}



int main(int argc, char* argv[])

{
	wWinMain(NULL,NULL,L"",SW_SHOWNORMAL);
}


//--------------------------------------------------------------------------------------
// Reject any D3D10 devices that aren't acceptable by returning false
//--------------------------------------------------------------------------------------
bool CALLBACK IsD3D10DeviceAcceptable( UINT Adapter, UINT Output, D3D10_DRIVER_TYPE DeviceType, DXGI_FORMAT BufferFormat, bool bWindowed, void* pUserContext )
{
    return true;
}


//--------------------------------------------------------------------------------------
// Create any D3D10 resources that aren't dependant on the back buffer
//--------------------------------------------------------------------------------------
HRESULT CALLBACK OnD3D10CreateDevice( ID3D10Device* pd3dDevice, const DXGI_SURFACE_DESC *pBufferSurfaceDesc, void* pUserContext )
{
	if (!pUserContext) return S_OK;
	MainFrame *mainFrame = (MainFrame*) pUserContext;

	return mainFrame->OnDeviceCreated(pd3dDevice, pBufferSurfaceDesc);
}


//--------------------------------------------------------------------------------------
// Create any D3D10 resources that depend on the back buffer
// Create and set the depth stencil texture if needed
//--------------------------------------------------------------------------------------
HRESULT CALLBACK OnD3D10ResizedSwapChain( ID3D10Device* pd3dDevice, IDXGISwapChain *pSwapChain, const DXGI_SURFACE_DESC* pBufferSurfaceDesc, void* pUserContext )
{
	if (!pUserContext) return S_OK;
	MainFrame *mainFrame = (MainFrame*) pUserContext;

	return mainFrame->OnSwapChainResized(pd3dDevice, pSwapChain, pBufferSurfaceDesc);	
}

//--------------------------------------------------------------------------------------
// Render the scene using the D3D10 device
//--------------------------------------------------------------------------------------
void CALLBACK OnD3D10FrameRender( ID3D10Device* pd3dDevice, double fTime, float fElapsedTime, void* pUserContext )
{
	if (!pUserContext) return;
	MainFrame *mainFrame = (MainFrame*) pUserContext;

	mainFrame->OnFrameRender(pd3dDevice, fTime, fElapsedTime);
}




//--------------------------------------------------------------------------------------
// Release D3D10 resources created in OnD3D10ResizedSwapChain 
//--------------------------------------------------------------------------------------
void CALLBACK OnD3D10ReleasingSwapChain( void* pUserContext )
{
	if (!pUserContext) return;
	MainFrame *mainFrame = (MainFrame*) pUserContext;

	mainFrame->OnSwapChainReleasing();
}


//--------------------------------------------------------------------------------------
// Release D3D10 resources created in OnD3D10CreateDevice 
//--------------------------------------------------------------------------------------
void CALLBACK OnD3D10DestroyDevice( void* pUserContext )
{
	if (!pUserContext) return;
	MainFrame *mainFrame = (MainFrame*) pUserContext;

	mainFrame->OnDeviceDestroyed();
}

//--------------------------------------------------------------------------------------
// Called right before creating a D3D9 or D3D10 device, allowing the app to modify the device settings as needed
//--------------------------------------------------------------------------------------
bool CALLBACK ModifyDeviceSettings( DXUTDeviceSettings* pDeviceSettings, void* pUserContext )
{
	pDeviceSettings->d3d10.SyncInterval = 0;
	return true;
}


//--------------------------------------------------------------------------------------
// Handle updates to the scene.  This is called regardless of which D3D API is used
//--------------------------------------------------------------------------------------
void CALLBACK OnFrameMove( double fTime, float fElapsedTime, void* pUserContext )
{
	if (!pUserContext) return;
	MainFrame *mainFrame = (MainFrame*) pUserContext;

	mainFrame->OnFrameUpdate(fTime, fElapsedTime);
}


//--------------------------------------------------------------------------------------
// Handle messages to the application
//--------------------------------------------------------------------------------------
LRESULT CALLBACK MsgProc( HWND hWnd, UINT uMsg, WPARAM wParam, LPARAM lParam, bool* pbNoFurtherProcessing, void* pUserContext )
{
	if (!pUserContext) return 0;
	MainFrame *mainFrame = (MainFrame*) pUserContext;

	mainFrame->MsgProc( hWnd, uMsg, wParam, lParam );
	return 0;
}


//--------------------------------------------------------------------------------------
// Handle key presses
//--------------------------------------------------------------------------------------
void CALLBACK OnKeyboard( UINT nChar, bool bKeyDown, bool bAltDown, void* pUserContext )
{
	if (!pUserContext) return;
	MainFrame *mainFrame = (MainFrame*) pUserContext;

	mainFrame->OnKeyDown(nChar, bKeyDown, bAltDown);

}
