#include "stdafx.h"

#include "CalculateVelocityForcesOperation2d.h"
#include "EffectManager.h"
#include "ExplosionParameters.h"
#include "VertexBufferUtils.h"
#include "VertexDefinition.h"

CalculateVelocityForcesOperation2d::CalculateVelocityForcesOperation2d(ID3D10Device* device, TextureStore* textureStore,
                                                                       int velocityId, int velocityForcesId, int densityId)
    :
    effectName("ExplosionSimulator2d.fx"),
    textureStore(textureStore),
    velocityId(velocityId),
    velocityForcesId(velocityForcesId),
    densityId(densityId),
    vorticityId(vorticityId),
    vortexCount(ExplosionParameters::Instance().vortexParticleCount),
    drawFromVortexBuffer(NULL),
    streamOutToVortexBuffer(NULL),
    inputLayout(NULL)
{
    InitCalculateVorticityOperation(device);
    InitCalculateVelocityForcesOperation(device);
    
    if (vortexCount <= 0) return;

    InitMoveVortexParticlesOperation(device);
    InitCalculateVortexParticleForcesOperation(device);
    InitVortexParticleBuffers(device);
}

CalculateVelocityForcesOperation2d::~CalculateVelocityForcesOperation2d()
{
    SAFE_RELEASE(drawFromVortexBuffer);
    SAFE_RELEASE(streamOutToVortexBuffer);
    SAFE_RELEASE(inputLayout);
}

void CalculateVelocityForcesOperation2d::InitCalculateVorticityOperation(ID3D10Device* device)
{
    vorticityId = textureStore->AddTexture("Vorticity", device);

    ID3D10Effect* effect = EffectManager::Instance().GetEffect(effectName, device);
    calculateVorticityTechnique = auto_ptr<TechniqueOperation>(
                                    new TechniqueOperation(effect, "CalculateVorticityTechnique", textureStore, device));
    calculateVorticityOperation = auto_ptr<HlslTextureOperation2d>(new HlslTextureOperation2d(device,
        ExplosionParameters::Instance().GetSimulationWidth(),
        ExplosionParameters::Instance().GetSimulationHeight(),
        calculateVorticityTechnique.get(), calculateVorticityTechnique.get(), USE_DISPLACED_BORDER ));

    calculateVorticityTechnique->AddSetShaderResourceTask("velocity", velocityId);
    calculateVorticityOperation->SetOutputTextureID(vorticityId);
}

void CalculateVelocityForcesOperation2d::InitCalculateVelocityForcesOperation(ID3D10Device* device)
{
    ID3D10Effect* effect = EffectManager::Instance().GetEffect(effectName, device);

    calculateVelocityForcesTechnique = auto_ptr<TechniqueOperation>(
        new TechniqueOperation(effect, "CalculateVelocityForcesTechnique", textureStore, device)
    );
    calculateVelocityForcesOperation = auto_ptr<HlslTextureOperation2d>(
        new HlslTextureOperation2d(
            device,
            ExplosionParameters::Instance().GetSimulationWidth(),
            ExplosionParameters::Instance().GetSimulationHeight(),
            calculateVelocityForcesTechnique.get(),
            calculateVelocityForcesTechnique.get(),
            USE_DISPLACED_BORDER
        )
    );

    calculateVelocityForcesTechnique->AddSetShaderResourceTask("density", densityId);
    calculateVelocityForcesTechnique->AddSetShaderResourceTask("velocity", velocityId);
    calculateVelocityForcesTechnique->AddSetShaderResourceTask("vorticity", vorticityId);

    calculateVelocityForcesOperation->SetOutputTextureID(velocityForcesId);
}

void CalculateVelocityForcesOperation2d::InitMoveVortexParticlesOperation(ID3D10Device* device)
{
    ID3D10Effect* effect = EffectManager::Instance().GetEffect(effectName, device);
    moveVortexParticlesTechnique = auto_ptr<TechniqueOperation>(
        new TechniqueOperation(effect, "MoveVorticesTechnique", textureStore, device)
    );
    moveVortexParticlesTechnique->AddSetShaderResourceTask("velocity", velocityId);
}

void CalculateVelocityForcesOperation2d::InitCalculateVortexParticleForcesOperation(ID3D10Device* device)
{
    ID3D10Effect* effect = EffectManager::Instance().GetEffect(effectName, device);
    calculateVortexParticleForcesTechnique = auto_ptr<TechniqueOperation>(
        new TechniqueOperation(effect, "RenderVorticesTechnique", textureStore, device)
    );
}

float GetRandom(float min, float max)
{
    float r = static_cast<float>(rand()) / RAND_MAX;
    return min + (max-min) * r;
}

void CalculateVelocityForcesOperation2d::InitVortexParticleBuffers(ID3D10Device* device)
{
    VertexPositionRadiusStrength* vortices = new VertexPositionRadiusStrength[vortexCount];

    const float minX = -0.3f;
    const float maxX = 0.3f;
    const float minY = 0.0f;
    const float maxY = 0.8f;

    const float texelSize = 1.0f / ExplosionParameters::Instance().GetSimulationWidth();
    const float radius = ExplosionParameters::Instance().vortexParticleMaxRadius;

    const float maxStrength = ExplosionParameters::Instance().vortexParticleMaxStrength;

    for (int i = 0; i < this->vortexCount; i++) {
        vortices[i] = VertexPositionRadiusStrength(
            D3DXVECTOR3(GetRandom(minX,maxX),GetRandom(minY,maxY),0),
            GetRandom(0, radius) * texelSize,
            GetRandom(-maxStrength, maxStrength));
    }

    ID3D10Buffer* initialBuffer = VertexBufferUtils::CreateDefaultVertexBuffer(device, vortices, vortexCount);

    this->drawFromVortexBuffer = VertexBufferUtils::CreateSOVertexBuffer<VertexPositionRadiusStrength>(device, vortexCount);
    this->streamOutToVortexBuffer = VertexBufferUtils::CreateSOVertexBuffer<VertexPositionRadiusStrength>(device, vortexCount);

    this->inputLayout = VertexBufferUtils::CreateInputLayout<VertexPositionRadiusStrength>(
        device, calculateVortexParticleForcesTechnique->GetFirstPass()
    );

    VertexBufferUtils::SetSingleVertexBuffer<VertexPositionRadiusStrength>(device, initialBuffer);
    VertexBufferUtils::SetSingleSOVertexBuffer(device, drawFromVortexBuffer);

    device->IASetInputLayout(inputLayout);
    device->IASetPrimitiveTopology(D3D10_PRIMITIVE_TOPOLOGY_POINTLIST);

    moveVortexParticlesTechnique->Execute(vortexCount, device);

    device->SOSetTargets(0,0,0);

    SAFE_RELEASE(initialBuffer);
    delete [] vortices;
}

void CalculateVelocityForcesOperation2d::Run(float timestep, ID3D10Device* device)
{
    calculateVorticityOperation->Execute(device);
    calculateVelocityForcesOperation->Execute(device);
    
    if (vortexCount > 0) {
        // <Common to all draw calls in this method>
        device->IASetInputLayout(inputLayout);
        device->IASetPrimitiveTopology(D3D10_PRIMITIVE_TOPOLOGY_POINTLIST);
        // </Common to all draw calls in this method>

        // <Render the contents in the drawFromBuffer into the streamOutToBuffer after moving the vortices>
        device->OMSetRenderTargets(0,0,0);
        VertexBufferUtils::SetSingleVertexBuffer<VertexPositionRadiusStrength>(device, drawFromVortexBuffer);
        VertexBufferUtils::SetSingleSOVertexBuffer(device, streamOutToVortexBuffer);
        moveVortexParticlesTechnique->Execute(vortexCount, device);

        device->SOSetTargets(0, 0, 0);
        std::swap(drawFromVortexBuffer, streamOutToVortexBuffer);
        // </Render the contents in the drawFromBuffer into the streamOutToBuffer after moving the vortices>

        textureStore->SetAsRenderTarget(velocityForcesId, false, device);
        VertexBufferUtils::SetSingleVertexBuffer<VertexPositionRadiusStrength>(device, drawFromVortexBuffer);
        calculateVortexParticleForcesTechnique->Execute(vortexCount, device);
    }
}
