#include "stdafx.h"

#include "DensityOperation2d.h"
#include "EffectManager.h"
#include "ExplosionParameters.h"
#include "ExplosionSimulator2d.h"
#include "HlslTextureOperation2d.h"
#include "TechniqueOperation.h"
#include "TextureStore2D.h"
#include "TextureUtils.h"
#include "VelocityOperation2d.h"
#include "VertexBufferUtils.h"
#include "VertexDefinition.h"

#pragma warning (disable:4995)
#include <iomanip>
#include <sstream>
#include <math.h>
#include <memory>

using namespace std;

ExplosionSimulator2d::ExplosionSimulator2d()
    :   
    textureStore(NULL),
    densityId(-1), velocityId(-1), thermalExpansionId(-1),
    calculateDensityChangesOperation(NULL),
    calculateVelocityForcesOperation(NULL),
    calculateThermalExpansionOperation(NULL),
    velocityOperation(NULL),
    densityOperation(NULL),
    framesToSimulate(1),
    simulateFreely(ExplosionParameters::Instance().GetSimulateFreely())
{
}

void ExplosionSimulator2d::Draw( ID3D10Device* device, double time, float elapsedTime )
{
    if (!ShouldSimulate()) return;

    if (calculateDensityChangesOperation) calculateDensityChangesOperation->Run(elapsedTime, device);
    if (calculateVelocityForcesOperation) calculateVelocityForcesOperation->Run(elapsedTime, device);
    if (calculateThermalExpansionOperation) calculateThermalExpansionOperation->Run(elapsedTime, device);

    if (velocityOperation) velocityOperation->Run(elapsedTime, device);
    if (densityOperation) densityOperation->Run(elapsedTime, device);
}

bool ExplosionSimulator2d::LoadGeneralGraphicsContent(ID3D10Device* device, const DXGI_SURFACE_DESC *bufferSurfaceDesc)
{
    InitializeEffectVariables(device);
    if (!InitializeTextures(device)) return false;
    if (!InitializeSimulationOperations(device)) return false;
    if (!InitializeCommonEffectVariables(device)) return false;
    return true;
}

void ExplosionSimulator2d::ReleaseGeneralGraphicsContent()
{
    SAFE_DELETE(densityOperation);
    SAFE_DELETE(velocityOperation);
    SAFE_DELETE(calculateThermalExpansionOperation);
    SAFE_DELETE(calculateVelocityForcesOperation);
    SAFE_DELETE(calculateDensityChangesOperation);
    SAFE_DELETE(textureStore);
}

void ExplosionSimulator2d::InitializeEffectVariables(ID3D10Device* device)
{
    ID3D10Effect* effect = EffectManager::Instance().GetEffect("ExplosionSimulator2d.fx", device);
    ExplosionParameters& p = ExplosionParameters::Instance();

    effect->GetVariableByName("LowerIgnitionLimit")->AsScalar()->SetFloat(p.GetLowerIgnitionLimit());
    effect->GetVariableByName("BurningRate")->AsScalar()->SetFloat(p.GetBurningRate());
    effect->GetVariableByName("StoichiometricMixture")->AsScalar()->SetFloat(p.GetStoichiometricMixture());
    effect->GetVariableByName("OutputHeat")->AsScalar()->SetFloat(p.GetOutputHeat());

    effect->GetVariableByName("ThermalExpansionRate")->AsScalar()->SetFloat(p.GetThermalExpansionRate());

    effect->GetVariableByName("Buoyancy")->AsScalar()->SetFloat(p.GetBuoyancy());
    effect->GetVariableByName("Gravity")->AsScalar()->SetFloat(p.GetGravity());

    effect->GetVariableByName("VorticityConfinementStrength")->AsScalar()->SetFloat(p.GetVorticityConfinementStrength());

    effect->GetVariableByName("Dissipation")->AsVector()->SetFloatVector(p.GetDissipation());
}

bool ExplosionSimulator2d::InitializeTextures(ID3D10Device* device)
{
    textureStore = new TextureStore2D(
        ExplosionParameters::Instance().GetSimulationWidth(),
        ExplosionParameters::Instance().GetSimulationHeight(), 
        device, 
        2
    );

    densityId = textureStore->AddTexture("Density", device);
    velocityId = textureStore->AddTexture("Velocity", device);
    thermalExpansionId = textureStore->AddTexture("Thermal Expansion", device);
    densitySourcesId = textureStore->AddTexture("Density Sources", device);
    velocityForcesId = textureStore->AddTexture("Velocity Forces", device);

    textureStore->ZeroTexture(densityId, device);
    textureStore->ZeroTexture(velocityId, device);
    textureStore->ZeroTexture(thermalExpansionId, device);
    textureStore->ZeroTexture(densitySourcesId, device);
    textureStore->ZeroTexture(velocityForcesId, device);

    InitializeFuel(device);
    InitializeSpark(device);

    return true;
}

bool ExplosionSimulator2d::InitializeCommonEffectVariables(ID3D10Device* device)
{
    ExplosionParameters& p = ExplosionParameters::Instance();

    ID3D10Effect *effect = EffectManager::Instance().GetEffect("ExplosionSimulator2d.fx", device);
    effect->GetVariableByName("texelSizes")->AsVector()->SetFloatVector(p.GetTexelSizes());
    effect->GetVariableByName("TimeStep")->AsScalar()->SetFloat(p.GetTimeStep());

    effect = EffectManager::Instance().GetEffect("FluidSolver2d.fx", device);
    effect->GetVariableByName("texelSizes")->AsVector()->SetFloatVector(p.GetTexelSizes());
    effect->GetVariableByName("TimeStep")->AsScalar()->SetFloat(p.GetTimeStep());

    return true;
}

bool ExplosionSimulator2d::InitializeSimulationOperations(ID3D10Device* device)
{
    calculateDensityChangesOperation = new CalculateDensityChangesOperation2d(
        device, textureStore, densityId, densitySourcesId
    );
    calculateVelocityForcesOperation = new CalculateVelocityForcesOperation2d(
        device, textureStore, velocityId, velocityForcesId, densityId
    );
    calculateThermalExpansionOperation = new CalculateThermalExpansionOperation2d(
        device, textureStore, densitySourcesId, thermalExpansionId
    );
    velocityOperation = new VelocityOperation2d(
        device, textureStore, velocityId, thermalExpansionId, velocityForcesId
    );
    densityOperation = new DensityOperation2d(
        device, textureStore, densityId, velocityId, densitySourcesId
    );

    return true;
}

void ExplosionSimulator2d::InitializeFuel(ID3D10Device* device)
{
    ID3D10Effect *effect = EffectManager::Instance().GetEffect("ExplosionSimulator2d.fx", device);
    effect->GetVariableByName("InitialFuelDensity")->AsScalar()->SetFloat(ExplosionParameters::Instance().GetInitialFuelDensity());
    effect->GetVariableByName("FuelAddScale")->AsScalar()->SetFloat(1.0f);

    auto_ptr<TechniqueOperation> renderFuelOperation(new TechniqueOperation(effect, "RenderFuelTechnique", textureStore, device));
    
    textureStore->ZeroTexture(densityId, device);
    textureStore->SetAsRenderTarget(densityId, false, device);

    float centerX = ExplosionParameters::Instance().GetFuelAreaCenterX();
    float centerY = ExplosionParameters::Instance().GetFuelAreaCenterY();
    float radiusX = static_cast<float>(ExplosionParameters::Instance().GetFuelAreaRadiusInTexels()) /
                    ExplosionParameters::Instance().GetSimulationWidth();
    float radiusY = static_cast<float>(ExplosionParameters::Instance().GetFuelAreaRadiusInTexels()) /
                    ExplosionParameters::Instance().GetSimulationHeight();

    float left   = centerX - radiusX*2;
    float right  = centerX + radiusX*2;
    float top    = centerY - radiusY*2;
    float bottom = centerY + radiusY*2;

    VertexPositionTexture vertices[] =
    {
        VertexPositionTexture(D3DXVECTOR3(left, top,   0), D3DXVECTOR2(-1, 1)),
        VertexPositionTexture(D3DXVECTOR3(left, bottom,0), D3DXVECTOR2(-1,-1)),
        VertexPositionTexture(D3DXVECTOR3(right,top,   0), D3DXVECTOR2( 1, 1)),
        VertexPositionTexture(D3DXVECTOR3(right,bottom,0), D3DXVECTOR2( 1,-1))
    };
    int vertexCount = sizeof(vertices)/sizeof(VertexPositionTexture);
    ID3D10Buffer* vertexBuffer = VertexBufferUtils::CreateDefaultVertexBuffer(device, vertices, vertexCount);
    ID3D10InputLayout* inputLayout = VertexBufferUtils::CreateInputLayout<VertexPositionTexture>(
        device, renderFuelOperation->GetFirstPass()
    );
    D3D10_PRIMITIVE_TOPOLOGY primitiveTopology = D3D10_PRIMITIVE_TOPOLOGY_TRIANGLESTRIP;
    
    VertexBufferUtils::SetSingleVertexBuffer<VertexPositionTexture>(device, vertexBuffer);
    device->IASetInputLayout(inputLayout);
    device->IASetPrimitiveTopology(primitiveTopology);

    renderFuelOperation->Execute(vertexCount, device);

    SAFE_RELEASE(vertexBuffer);
    SAFE_RELEASE(inputLayout);
}

void ExplosionSimulator2d::InitializeSpark(ID3D10Device* device)
{
    ID3D10Effect *effect = EffectManager::Instance().GetEffect("ExplosionSimulator2d.fx", device);
    effect->GetVariableByName("sparkTemp")->AsScalar()->SetFloat(ExplosionParameters::Instance().GetSparkTemp());

    auto_ptr<TechniqueOperation> renderSparkOperation(new TechniqueOperation(effect, "RenderSparkTechnique", textureStore, device));
    
    textureStore->SetAsRenderTarget(densityId, false, device);

    float posX = ExplosionParameters::Instance().GetSparkPositionX();
    float posY = ExplosionParameters::Instance().GetSparkPositionY();

    VertexPosition vertices[] =
    {
        VertexPosition(D3DXVECTOR3(posX,posY,0))
    };
    int vertexCount = sizeof(vertices)/sizeof(VertexPosition);
    ID3D10Buffer* vertexBuffer = VertexBufferUtils::CreateDefaultVertexBuffer(device, vertices, vertexCount);
    ID3D10InputLayout* inputLayout = VertexBufferUtils::CreateInputLayout<VertexPosition>(device, renderSparkOperation->GetFirstPass());
    D3D10_PRIMITIVE_TOPOLOGY primitiveTopology = D3D10_PRIMITIVE_TOPOLOGY_POINTLIST;

    VertexBufferUtils::SetSingleVertexBuffer<VertexPositionTexture>(device, vertexBuffer);
    device->IASetInputLayout(inputLayout);
    device->IASetPrimitiveTopology(primitiveTopology);

    renderSparkOperation->Execute(vertexCount, device);

    SAFE_RELEASE(vertexBuffer);
    SAFE_RELEASE(inputLayout);
}

TextureStore* ExplosionSimulator2d::GetTextureStore()
{
    return textureStore;
}

int ExplosionSimulator2d::GetDensityResultID()
{
    return densityId;
}

int ExplosionSimulator2d::GetObstaclesID()
{
    return -1;
}

int ExplosionSimulator2d::GetVelocityResultID()
{
    return velocityId;
}

void ExplosionSimulator2d::HandleKeyEvent(UINT nChar, bool keyDown, bool altDown)
{
    if (!keyDown) return;

    if (nChar >= (int)'1' && nChar <= (int)'9') {
        this->framesToSimulate = nChar-(int)'1' + 2;
        this->simulateFreely = false;
    } else if (nChar == VK_RETURN ) {
        this->simulateFreely = !this->simulateFreely;
    }
}

bool ExplosionSimulator2d::ShouldSimulate()
{
    return simulateFreely || framesToSimulate>0;
}

void ExplosionSimulator2d::Update(double time, float elapsedTime)
{
    framesToSimulate--;
}
