#include "stdafx.h"

#include <iomanip>
#include <sstream>
#include <math.h>
#include "ExplosionSimulator3d.h"
#include "TextureStore3D.h"
#include "HlslTextureOperation2d.h"
#include "TechniqueOperation.h"
#include "TextureUtils.h"
#include "EffectManager.h"
#include "ExplosionParameters.h"
#include "DensityOperation3d.h"
#include "ExplosionSpecificOperation3d.h"
#include "VelocityOperation3d.h"
#include "VortexParticleOperation3d.h"
#include "TechniqueOperation.h"
#include "HlslTextureOperation3d.h"
#include "VertexBufferUtils.h"

ExplosionSimulator3d* ExplosionSimulator3d::Instance;


ExplosionSimulator3d::ExplosionSimulator3d()
:   textureStore(NULL),
    densityId(-1), velocityId(-1), thermalExpansionId(-1), obstaclesId(-1),
    explosionSpecificOperation(NULL),
    velocityOperation(NULL),
    densityOperation(NULL),
    vortexParticleOperation(NULL),
    framesToSimulate(1),
    simulateFreely(false)
{
    simulateFreely = ExplosionParameters::Instance().GetSimulateFreely();
    zSliceToDraw = ExplosionParameters::Instance().GetSimulationDepth() / 2;
    Instance = this;
}

void ExplosionSimulator3d::Draw( ID3D10Device* device, double time, float elapsedTime )
{
    if (!ShouldSimulate()) return;

    explosionSpecificOperation->Run(elapsedTime, device);

    //vortexParticleOperation->Run(elapsedTime, device);
    velocityOperation->Run(elapsedTime, device);
    densityOperation->Run(elapsedTime, device);
}

bool ExplosionSimulator3d::LoadGeneralGraphicsContent(ID3D10Device* device, const DXGI_SURFACE_DESC *bufferSurfaceDesc)
{
    if (!InitTextures(device)) return false;
    if (!InitSimulationOperations(device)) return false;
    if (!InitCommonEffectVariables(device)) return false;
    return true;
}

void ExplosionSimulator3d::ReleaseGeneralGraphicsContent()
{
    SAFE_DELETE(vortexParticleOperation);
    SAFE_DELETE(densityOperation);
    SAFE_DELETE(velocityOperation);
    SAFE_DELETE(explosionSpecificOperation);
    SAFE_DELETE(textureStore);
}

bool ExplosionSimulator3d::InitTextures(ID3D10Device* device)
{
    textureStore = new TextureStore3D(ExplosionParameters::Instance().GetSimulationWidth(),
                                    ExplosionParameters::Instance().GetSimulationHeight(),
                                    ExplosionParameters::Instance().GetSimulationWidth(),
                                    device, 2);

    velocityId = textureStore->AddTexture("Velocity", device);
    densityId = textureStore->AddTexture("Density", device);
    thermalExpansionId = textureStore->AddTexture("Thermal Expansion", device);
    obstaclesId = textureStore->AddTexture("Obstacles", device);

    textureStore->ZeroTexture(velocityId, device);
    textureStore->ZeroTexture(thermalExpansionId, device);

    InitializeObstacles(device);
    InitializeDensities(device);

    return true;
}

void ExplosionSimulator3d::DrawCube(ID3D10Device* device,
                                    float xMin, float xMax,
                                    float yMin, float yMax,
                                    int zMin, int zMax,
                                    TechniqueOperation* technique)
{
    int sliceCount = zMax - zMin + 1;
    int vertexCount = sliceCount*6;
    VertexPositionTexture3D* vertices = new VertexPositionTexture3D[vertexCount];

    int startIndex = 0;
    for (int z = zMin; z <= zMax; z++) {
        VertexPositionTexture3D bottomleft(  D3DXVECTOR3(xMin,yMin,0), D3DXVECTOR3(-1, 1, (float)z));
        VertexPositionTexture3D topleft(     D3DXVECTOR3(xMin,yMax,0), D3DXVECTOR3(-1, -1, (float)z));
        VertexPositionTexture3D bottomright( D3DXVECTOR3(xMax,yMin,0), D3DXVECTOR3(1, 1, (float)z));
        VertexPositionTexture3D topright(    D3DXVECTOR3(xMax,yMax,0), D3DXVECTOR3(1, -1, (float)z));

        vertices[startIndex++] = bottomleft;
        vertices[startIndex++] = topleft;
        vertices[startIndex++] = bottomright;
        vertices[startIndex++] = bottomright;
        vertices[startIndex++] = topleft;
        vertices[startIndex++] = topright;
    }

    ID3D10Buffer* buffer = VertexBufferUtils::CreateDefaultVertexBuffer(device, vertices, vertexCount);
    ID3D10InputLayout* layout = VertexBufferUtils::CreateInputLayout<VertexPositionTexture3D>(device, technique->GetFirstPass());


    VertexBufferUtils::SetSingleVertexBuffer<VertexPositionTexture3D>(device, buffer);
    device->IASetInputLayout(layout);
    device->IASetPrimitiveTopology(D3D10_PRIMITIVE_TOPOLOGY_TRIANGLELIST);

    technique->Execute(vertexCount, device);

    buffer->Release();
    layout->Release();
    delete [] vertices;

}

void ExplosionSimulator3d::InitializeFuel(ID3D10Device* device)
{
    ID3D10Effect *effect = EffectManager::Instance().GetEffect("ExplosionSimulator3d.fx", device);
    effect->GetVariableByName("InitialFuelDensity")->AsScalar()->SetFloat(ExplosionParameters::Instance().GetInitialFuelDensity());

    TechniqueOperation* renderFuel = new TechniqueOperation(effect, "RenderFuel", textureStore, device);

    textureStore->ZeroTexture(densityId, device);

    // The radius gets multiplied by 2 because the vertices are in [-1,1] and not [0,1]
    float xMin = ExplosionParameters::Instance().GetSparkPositionX() -
        2*(float)ExplosionParameters::Instance().GetFuelAreaRadiusInTexels() / ExplosionParameters::Instance().GetSimulationWidth();
    float xMax = ExplosionParameters::Instance().GetSparkPositionX() +
        2*(float)ExplosionParameters::Instance().GetFuelAreaRadiusInTexels() / ExplosionParameters::Instance().GetSimulationWidth();
    float yMin = ExplosionParameters::Instance().GetSparkPositionY() -
        2*(float)ExplosionParameters::Instance().GetFuelAreaRadiusInTexels() / ExplosionParameters::Instance().GetSimulationHeight();
    float yMax = ExplosionParameters::Instance().GetSparkPositionY() +
        2*(float)ExplosionParameters::Instance().GetFuelAreaRadiusInTexels() / ExplosionParameters::Instance().GetSimulationHeight();

    // Expects a GetSparkPositionZ in [0,1]
    int zMin = (int)(ExplosionParameters::Instance().GetSparkPositionZ()*ExplosionParameters::Instance().GetSimulationDepth())
        - ExplosionParameters::Instance().GetFuelAreaRadiusInTexels();
    int zMax = (int)(ExplosionParameters::Instance().GetSparkPositionZ()*ExplosionParameters::Instance().GetSimulationDepth())
        + ExplosionParameters::Instance().GetFuelAreaRadiusInTexels();

    if (zMin<0) zMin = 0;
    if (zMax >= ExplosionParameters::Instance().GetSimulationDepth())
        zMax = ExplosionParameters::Instance().GetSimulationDepth()-1;
    assert(zMin <= zMax);

    textureStore->SetAsRenderTarget(densityId, false, device);
    DrawCube(device, xMin, xMax, yMin, yMax, zMin, zMax, renderFuel);

    delete renderFuel;
}

void ExplosionSimulator3d::InitializeSpark(ID3D10Device* device)
{
    ID3D10Effect *effect = EffectManager::Instance().GetEffect("ExplosionSimulator3d.fx", device);
    effect->GetVariableByName("sparkTemp")->AsScalar()->SetFloat(ExplosionParameters::Instance().GetSparkTemp());
    TechniqueOperation* renderSpark = new TechniqueOperation(effect, "RenderSpark", textureStore, device);

    float x = ExplosionParameters::Instance().GetSparkPositionX();
    float y = ExplosionParameters::Instance().GetSparkPositionY();
    float z = ExplosionParameters::Instance().GetSparkPositionZ() * ExplosionParameters::Instance().GetSimulationDepth();
    VertexPositionTexture3D vertex(D3DXVECTOR3(x,y,0), D3DXVECTOR3(0,0,z));

    ID3D10Buffer* buffer = VertexBufferUtils::CreateDefaultVertexBuffer(device, &vertex, 1);
    ID3D10InputLayout* layout = VertexBufferUtils::CreateInputLayout<VertexPositionTexture3D>(device, renderSpark->GetFirstPass());

    VertexBufferUtils::SetSingleVertexBuffer<VertexPositionTexture3D>(device, buffer);
    device->IASetInputLayout(layout);
    device->IASetPrimitiveTopology(D3D10_PRIMITIVE_TOPOLOGY_POINTLIST);
    textureStore->SetAsRenderTarget(densityId, false, device);

    renderSpark->Execute(1, device);

    buffer->Release();
    layout->Release();
    delete renderSpark;
}

void ExplosionSimulator3d::InitializeDensities(ID3D10Device* device)
{
    InitializeFuel(device);
    InitializeSpark(device);
}

void ExplosionSimulator3d::InitializeObstacles(ID3D10Device* device)
{
    ID3D10Effect *effect = EffectManager::Instance().GetEffect("ExplosionSimulator3d.fx", device);
    TechniqueOperation* renderBlack = new TechniqueOperation(effect, "BlackTechnique", textureStore, device);
    HlslTextureOperation* textureOperation = new HlslTextureOperation3d(device,
        ExplosionParameters::Instance().GetSimulationWidth(),
        ExplosionParameters::Instance().GetSimulationHeight(),
        ExplosionParameters::Instance().GetSimulationDepth(),
        renderBlack, NULL);
    textureOperation->SetOutputTextureID(obstaclesId);

    float white[] = {1,1,1,1};
    textureStore->ClearRenderTarget(obstaclesId, white, device);
    textureOperation->Execute(device);

    delete textureOperation;
    delete renderBlack;

    // Number of pixels
    float xLenHalf = 2;
    float yLenHalf = 12;
    int   zLenHalf = 4;

    float xCenter = 0.0f;
    float yCenter = 0.2f;
    int zCenter   = ExplosionParameters::Instance().GetSimulationDepth()/2;

    float xMin = xCenter - xLenHalf / ExplosionParameters::Instance().GetSimulationWidth();
    float xMax = xCenter + xLenHalf / ExplosionParameters::Instance().GetSimulationWidth();
    float yMin = yCenter - yLenHalf / ExplosionParameters::Instance().GetSimulationHeight();
    float yMax = yCenter + yLenHalf / ExplosionParameters::Instance().GetSimulationHeight();
    int zMin = zCenter - zLenHalf;
    int zMax = zCenter + zLenHalf;
    int depth = ExplosionParameters::Instance().GetSimulationDepth();

    TechniqueOperation* renderWhite = new TechniqueOperation(effect, "WhiteTechnique", textureStore, device);
    textureStore->SetAsRenderTarget(obstaclesId, false, device);

    //DrawCube(device, xMin, xMax, yMin, yMax, zMin, zMax, renderWhite);
    //DrawCube(device, xMin, xMax+0.2f, yMin-0.1f, yMin, zMin, zMax, renderWhite);

    //////////
    // Arch
    //////////
    //DrawCube(device,
    //         -0.25f, 0.25f,
    //       0.5f, 1.0f,
    //       0.60f*depth, 0.9f*depth,
    //       renderWhite);
    //DrawCube(device,
    //         -0.25f, 0.25f,
    //       0.5f, 1.0f,
    //       0.1f*depth, 0.40f*depth,
    //       renderWhite);
    //DrawCube(device,
    //         -0.25f, 0.25f,
    //       0.4f, 0.66f,
    //       0.1f*depth, 0.9f*depth,
    //       renderWhite);

    //DrawCube(device,
    //         -1.0f, 0.1f,
    //       0.5f, 0.6f,
    //       0, (int) (0.6 * depth),
    //       renderWhite);

    //DrawCube(device,
    //         -1.0f, 0.1f,
    //       0.5f, 0.6f,
    //       0, (int) (0.6 * depth),
    //       renderWhite);

    delete renderWhite;
}

bool ExplosionSimulator3d::InitCommonEffectVariables(ID3D10Device* device)
{
    ExplosionParameters& p = ExplosionParameters::Instance();

    ID3D10Effect *effect = EffectManager::Instance().GetEffect("ExplosionSimulator3d.fx", device);
    effect->GetVariableByName("simulationDims")->AsVector()->SetFloatVector(p.GetSimulationDims());
    effect->GetVariableByName("texelSizes")->AsVector()->SetFloatVector(p.GetTexelSizes());
    effect->GetVariableByName("TimeStep")->AsScalar()->SetFloat(p.GetTimeStep());

    effect = EffectManager::Instance().GetEffect("FluidSolver3d.fx", device);
    effect->GetVariableByName("simulationDims")->AsVector()->SetFloatVector(p.GetSimulationDims());
    effect->GetVariableByName("texelSizes")->AsVector()->SetFloatVector(p.GetTexelSizes());
    effect->GetVariableByName("TimeStep")->AsScalar()->SetFloat(p.GetTimeStep());

    return true;
}

bool ExplosionSimulator3d::InitSimulationOperations(ID3D10Device* device)
{

    //vortexParticleOperation = new VortexParticleOperation3d(device, textureStore, velocityId);
    explosionSpecificOperation = new ExplosionSpecificOperation3d(device, textureStore, densityId, velocityId, thermalExpansionId, obstaclesId);
    velocityOperation = new VelocityOperation3d(device, textureStore, velocityId, thermalExpansionId, obstaclesId);
    densityOperation = new DensityOperation3d(device, textureStore, velocityId, densityId, obstaclesId);

    return true;
}

TextureStore* ExplosionSimulator3d::GetTextureStore()
{
    return textureStore;
}

int ExplosionSimulator3d::GetDensityResultID()
{
    return densityId;
}

int ExplosionSimulator3d::GetVelocityResultID()
{
    return velocityId;
}

int ExplosionSimulator3d::GetObstaclesID()
{
    return obstaclesId;
}

void ExplosionSimulator3d::HandleKeyEvent(UINT nChar, bool keyDown, bool altDown)
{
    if (!keyDown) return;

    if (nChar >= (int)'1' && nChar <= (int)'9') {
        this->framesToSimulate = nChar-(int)'1' + 2;
        this->simulateFreely = false;
    } else if (nChar == VK_RETURN ) {
        this->simulateFreely = !this->simulateFreely;
    } else if (nChar == VK_ADD) {
        this->zSliceToDraw++;
        if (this->zSliceToDraw < 0) this->zSliceToDraw = 0;
    } else if (nChar == VK_SUBTRACT) {
        int zMax = ExplosionParameters::Instance().GetSimulationDepth() - 1;
        this->zSliceToDraw--;
        if (this->zSliceToDraw > zMax) this->zSliceToDraw = zMax;
    }
}

bool ExplosionSimulator3d::ShouldSimulate()
{
    return simulateFreely || framesToSimulate>0;
}

void ExplosionSimulator3d::Update(double time, float elapsedTime)
{
    framesToSimulate--;
}
