//--------------------------------------------------------------------------------------
// Structures
//--------------------------------------------------------------------------------------

struct VS_Pos3
{
	float3 Position : POS;
};

struct PS_Pos4
{
	float4 Position : SV_POSITION;
};

struct VS_Pos3Tex2
{
	float3 Position : POS;
	float2 Texcoord : TEXCOORD0;	
};

struct VS_Pos3Tex2Offset2
{
	float3 Position : POS;
	float2 Texcoord : TEXCOORD0;	
	float2 Offset : OFFSET;	
};

struct PS_Pos4Tex2
{
	float4 Position : SV_POSITION;	
	float2 Texcoord : TEXCOORD0;	
};

struct PS_Pos4Tex2Offset2
{
	float4 Position : SV_POSITION;	
	float2 Texcoord : TEXCOORD0;	
	float2 Offset : OFFSET;	
};

struct VS_VortexData
{
	float3 Pos : POS;
	float Radius : RADIUS;
	float Strength : STRENGTH;
};

struct GS_VortexData
{
	float4 Pos : POS;
	float Radius : RADIUS;
	float Strength : STRENGTH;
};

struct PS_VortexData
{
	float4 Pos : SV_POSITION;
	float2 Tex : TEXCOORD0;
	float Strength : STRENGTH;
};

//--------------------------------------------------------------------------------------
// States
//--------------------------------------------------------------------------------------

SamplerState samPoint
{
    AddressU = Clamp;
    AddressV = Clamp;
};

SamplerState samLinear
{
	Filter = MIN_MAG_MIP_LINEAR;
    AddressU = Clamp;
    AddressV = Clamp;
};

BlendState NoBlending
{
    AlphaToCoverageEnable = FALSE;
    BlendEnable[0] = FALSE;
};

BlendState AdditiveBlending
{
    AlphaToCoverageEnable = FALSE;
    BlendEnable[0] = TRUE;
    SrcBlend = ONE;
    DestBlend = ONE;
    BlendOp = ADD;
    SrcBlendAlpha = ZERO;
    DestBlendAlpha = ZERO;
    BlendOpAlpha = ADD;
    RenderTargetWriteMask[0] = 0x0F;
};

DepthStencilState DisableDepth
{
	DepthEnable = FALSE;
	DepthWriteMask = ZERO;
}; 

//--------------------------------------------------------------------------------------
// Textures
//--------------------------------------------------------------------------------------

Texture2D density;
Texture2D vorticity;
Texture2D velocity;
Texture2D densitySources;

//--------------------------------------------------------------------------------------
// Constant Buffers
//--------------------------------------------------------------------------------------

cbuffer cbNeverChanges{
	float TimeStep;
	float2 texelSizes;

	float InitialFuelDensity;
	float sparkTemp;

	float Buoyancy;
	float Gravity;

	float LowerIgnitionLimit;
	float BurningRate;
	float StoichiometricMixture;
	float OutputHeat;

	float4 Dissipation;

	float ThermalExpansionRate;
	float VorticityConfinementStrength;	
};

cbuffer cbChangesPerFrame{
	float FuelAddScale;
};

//--------------------------------------------------------------------------------------
// General Vertex Shader
//--------------------------------------------------------------------------------------

PS_Pos4Tex2 VS_General(VS_Pos3Tex2 input) {
	PS_Pos4Tex2 output = (PS_Pos4Tex2)0;
	output.Position = float4(input.Position,1);
	output.Texcoord = input.Texcoord;
	return output;
}

PS_Pos4 VS_PosGeneral(VS_Pos3 input) {
	PS_Pos4 output = (PS_Pos4)0;
	output.Position = float4(input.Position,1);
	return output;
}

//--------------------------------------------------------------------------------------
// Render fuel
//--------------------------------------------------------------------------------------

float4 PS_RenderFuel(PS_Pos4Tex2 input) : SV_Target
{
	float4 ret = 0;
	float2 sqr = input.Texcoord*input.Texcoord;
	float dist = sqr.x + sqr.y;
	if (dist < 1.0f) {
		ret.x = (1 - sqrt(dist)) * InitialFuelDensity;
	}
	return ret * FuelAddScale;
}

technique10 RenderFuelTechnique
{
	pass P0 
	{
		SetVertexShader( CompileShader( vs_4_0, VS_General() ) );
        SetGeometryShader( NULL );
        SetPixelShader( CompileShader( ps_4_0, PS_RenderFuel() ) );

        SetBlendState( AdditiveBlending, float4( 0.0f, 0.0f, 0.0f, 0.0f ), 0xFFFFFFFF );        
	}
}

//--------------------------------------------------------------------------------------
// Render spark
//--------------------------------------------------------------------------------------

float4 PS_RenderSpark(PS_Pos4 input) : SV_Target
{
	return float4(0,0,sparkTemp,0); 
}

technique10 RenderSparkTechnique
{
	pass P0 
	{
		SetVertexShader( CompileShader( vs_4_0, VS_PosGeneral() ) );
        SetGeometryShader( NULL );
        SetPixelShader( CompileShader( ps_4_0, PS_RenderSpark() ) );

        SetBlendState( AdditiveBlending, float4( 0.0f, 0.0f, 0.0f, 0.0f ), 0xFFFFFFFF );        
	}
}

//--------------------------------------------------------------------------------------
// Calculate density sources
//--------------------------------------------------------------------------------------

float4 PS_CalculateDensitySources(PS_Pos4Tex2 input) : SV_Target
{
	float4 den = density.Sample(samPoint, input.Texcoord);

	float4 ret = 0.0f;
	if (den.z > LowerIgnitionLimit) {
		float C = BurningRate * den.x;
		ret.x = -C;
		ret.y = C * (StoichiometricMixture + 1);
		ret.z = C * OutputHeat;
	}
	ret.xyz -= Dissipation * den.xyz;

	return ret;
}

technique10 CalculateDensitySourcesTechnique
{
	pass 
	{
		SetVertexShader( CompileShader( vs_4_0, VS_General() ) );
        SetGeometryShader( NULL );
        SetPixelShader( CompileShader( ps_4_0, PS_CalculateDensitySources() ) );

        SetBlendState( NoBlending, float4( 0.0f, 0.0f, 0.0f, 0.0f ), 0xFFFFFFFF );        
	}
}

//--------------------------------------------------------------------------------------
// Calculate velocity forces
//--------------------------------------------------------------------------------------

float2 CalcVorticityForces(float2 coords) {

    float4 wLeft   = abs(vorticity.Sample(samPoint, coords - float2(texelSizes.x, 0.0f)));
    float4 wRight  = abs(vorticity.Sample(samPoint, coords + float2(texelSizes.x, 0.0f)));
    float4 wBottom = abs(vorticity.Sample(samPoint, coords - float2(0.0f, texelSizes.y)));
    float4 wTop    = abs(vorticity.Sample(samPoint, coords + float2(0.0f, texelSizes.y)));

    float2 dw;
    dw.x = 0.5f * (wRight.x - wLeft.x); // dwdx
    dw.y = 0.5f * (wTop.x - wBottom.x); // dwdy

    float lensqr = dw.x*dw.x + dw.y*dw.y;
    float len = length(dw);
	
    if (lensqr > 1e-7f) {

        float4 w = vorticity.Sample(samPoint, coords);
        float4 force;
        force.x = dw.y *  w.x;
        force.y = dw.x * -w.x;
        force.zw = float2(0, 0);

        force *= VorticityConfinementStrength / len;

        return force;
    } else {
        return 0;
    }
}

float4 PS_CalculateVelocityForces(PS_Pos4Tex2 input) : SV_Target
{
    float3 den = density.Sample(samPoint, input.Texcoord).xyz;

	float4 force = 0.0f;
	force.y = den.z * Buoyancy - den.y * Gravity;
	force.y -= 0.5 * velocity.Sample(samPoint, input.Texcoord).y;
	force.xy += CalcVorticityForces(input.Texcoord);
	return force;
}

technique10 CalculateVelocityForcesTechnique
{
	pass 
	{
		SetVertexShader( CompileShader( vs_4_0, VS_General() ) );
        SetGeometryShader( NULL );
        SetPixelShader( CompileShader( ps_4_0, PS_CalculateVelocityForces() ) );

        SetBlendState( NoBlending, float4( 0.0f, 0.0f, 0.0f, 0.0f ), 0xFFFFFFFF );        
	}
}

//--------------------------------------------------------------------------------------
// Calculate thermal expansion
//--------------------------------------------------------------------------------------

float4 PS_CalculateThermalExpansion(PS_Pos4Tex2 input) : SV_Target
{
    float4 ret = 0.0f;
    ret.x = ThermalExpansionRate * densitySources.Sample(samPoint, input.Texcoord).z;
    return ret;
}

technique10 CalculateThermalExpansionTechnique
{
	pass
	{
		SetVertexShader( CompileShader( vs_4_0, VS_General() ) );
        SetGeometryShader( NULL );
        SetPixelShader( CompileShader( ps_4_0, PS_CalculateThermalExpansion() ) );

        SetBlendState( NoBlending, float4( 0.0f, 0.0f, 0.0f, 0.0f ), 0xFFFFFFFF );        
	}
}

//--------------------------------------------------------------------------------------
// Calculate Vorticity
//--------------------------------------------------------------------------------------

float4 PS_CalculateVorticity(PS_Pos4Tex2 input) : SV_Target
{
    float4 vLeft   = velocity.Sample(samPoint, input.Texcoord - float2(texelSizes.x ,0.0f)); 
    float4 vRight  = velocity.Sample(samPoint, input.Texcoord + float2(texelSizes.x ,0.0f)); 
    float4 uBottom = velocity.Sample(samPoint, input.Texcoord - float2(0.0f, texelSizes.y));
    float4 uTop    = velocity.Sample(samPoint, input.Texcoord + float2(0.0f, texelSizes.y));

    float dvdx = 0.5f * (vRight.y - vLeft.y);
    float dudy = 0.5f * (uTop.x - uBottom.x);
    
    float4 vorticity = 0.0f;
    vorticity.x = dvdx - dudy;

    return vorticity;
}

technique10 CalculateVorticityTechnique
{
	pass P0
	{
		SetVertexShader( CompileShader( vs_4_0, VS_General() ) );
        SetGeometryShader( NULL );
        SetPixelShader( CompileShader( ps_4_0, PS_CalculateVorticity() ) );

        SetBlendState( NoBlending, float4( 0.0f, 0.0f, 0.0f, 0.0f ), 0xFFFFFFFF );        
	}
}

//--------------------------------------------------------------------------------------
// Vortex particle movement
//--------------------------------------------------------------------------------------

VS_VortexData VSVortexPassThrough(VS_VortexData input) 
{
	float2 speed = velocity.SampleLevel(samLinear, (input.Pos.xy+1)/2, 0).xy;
	input.Pos.xy += speed * TimeStep * texelSizes.xy;
	return input;
}

[MaxVertexCount(1)]
void GS_MoveParticles(point VS_VortexData input[1], inout PointStream<VS_VortexData> ParticleOutputStream)
{
	ParticleOutputStream.Append(input[0]);
}

GeometryShader gsVortexStreamOut = ConstructGSWithSO( 
    CompileShader( gs_4_0, GS_MoveParticles() ), 
    "POS.xyz; RADIUS.x; STRENGTH.x" 
);

technique10 MoveVorticesTechnique
{
	pass P0
	{
		SetVertexShader( CompileShader( vs_4_0, VSVortexPassThrough() ) );
        SetGeometryShader( gsVortexStreamOut );
        SetPixelShader( NULL );
        
        SetDepthStencilState(DisableDepth, 0);
	}
}

//--------------------------------------------------------------------------------------
// Vortex particle rendering
//--------------------------------------------------------------------------------------

GS_VortexData VS_Vortex(VS_VortexData input) {
	GS_VortexData output = (GS_VortexData)0;
	output.Pos = float4(input.Pos,1);
	output.Radius = input.Radius;
	output.Strength = input.Strength;
	return output;
}

//
// Geometry shader for creating point sprites
//
[maxvertexcount(4)]
void GS_VortexPointSprite(point GS_VortexData input[1], inout TriangleStream<PS_VortexData> SpriteStream)
{
    PS_VortexData output;
	output.Strength = input[0].Strength;
	float radius = input[0].Radius;
    float4 pos = input[0].Pos;

	output.Pos = pos + float4(-radius, radius, 0, 0);
	output.Tex = float2(-1,1);
	SpriteStream.Append(output);

	output.Pos = pos + float4( radius, radius, 0, 0);
	output.Tex = float2(1,1);
	SpriteStream.Append(output);

	output.Pos = pos + float4(-radius,-radius, 0, 0);
	output.Tex = float2(-1,-1);
	SpriteStream.Append(output);

	output.Pos = pos + float4( radius,-radius, 0, 0);
	output.Tex = float2(1,-1);
	SpriteStream.Append(output);

	SpriteStream.RestartStrip();
}

// Simplified gaussian kernel because the radius is 1.
float GaussianKernel(float distSqr)
{
    return pow(2.71828182845904523536f, -distSqr*3.0f);
}

float4 PS_Vortex(PS_VortexData input) : SV_Target
{
	float4 ret = 0;
	float2 temp = input.Tex*input.Tex;
	float distSqr = temp.x+temp.y;

	float len = length(input.Tex.xy);

	if (distSqr < 1.0f && len > 0)
	{
		float2 vec = input.Tex.xy / len;
		ret.xy = input.Strength * (1-distSqr);
	}

	return ret;
}

technique10 RenderVorticesTechnique
{
	pass P0
	{
		SetVertexShader( CompileShader( vs_4_0, VS_Vortex() ) );
        SetGeometryShader( CompileShader( gs_4_0, GS_VortexPointSprite() ) );
        SetPixelShader( CompileShader( ps_4_0, PS_Vortex() ) );		

        SetBlendState( AdditiveBlending, float4( 0.0f, 0.0f, 0.0f, 0.0f ), 0xFFFFFFFF );
	}
}
