//--------------------------------------------------------------------------------------
// Structures
//--------------------------------------------------------------------------------------

#define CENTERCELL input.Texcoord
#define LEFTCELL float3(input.Texcoord.x-texelSizes.x, input.Texcoord.y, input.Texcoord.z)
#define RIGHTCELL float3(input.Texcoord.x+texelSizes.x, input.Texcoord.y, input.Texcoord.z)
#define TOPCELL float3(input.Texcoord.x, input.Texcoord.y+texelSizes.y, input.Texcoord.z)
#define BOTTOMCELL float3(input.Texcoord.x, input.Texcoord.y-texelSizes.y, input.Texcoord.z)
#define BACKCELL float3(input.Texcoord.x, input.Texcoord.y, input.Texcoord.z-texelSizes.z)
#define FRONTCELL float3(input.Texcoord.x, input.Texcoord.y, input.Texcoord.z+texelSizes.z)

struct VS_Pos3Tex3
{
	float3 Position : POS;
	float3 Texcoord : TEXCOORD;	
};

struct GS_Pos4Tex3SliceID
{
	float4 Position : SV_POSITION;	
	float3 Texcoord : TEXCOORD;	
    uint SliceIndex : SLICE;
};

struct PS_Pos4Tex3SliceID
{
	float4 Position : SV_POSITION;	
	float3 Texcoord : TEXCOORD;	
    uint SliceIndex : SV_RenderTargetArrayIndex;
};

struct ExplosionOutput {
	float4 Density : SV_Target0;
	float4 Velocity: SV_Target1;
	float4 ThermalExpansion: SV_Target2;
};

//--------------------------------------------------------------------------------------
// Constants
//--------------------------------------------------------------------------------------

cbuffer cbNeverChanges
{
	float3 simulationDims;
	float TimeStep;
	float3 texelSizes;

	float InitialFuelDensity;
	float sparkTemp;
	float LowerIgnitionLimit;
	float BurningRate;
	float StoichiometricMixture;
	float OutputHeat;
	float4 Dissipation;
	float Gravity;
	float Buoyancy;
	float ThermalExpansionRate;
	float VorticityConfinementStrength;
};

//--------------------------------------------------------------------------------------
// States
//--------------------------------------------------------------------------------------

SamplerState samPoint
{
	Filter = MIN_MAG_MIP_POINT;
    AddressU = Clamp;
    AddressV = Clamp;
    AddressW = Clamp;
};

SamplerState samLinear
{
	Filter = MIN_MAG_MIP_LINEAR;
    AddressU = Clamp;
    AddressV = Clamp;
    AddressW = Clamp;
};

BlendState NoBlending
{
    AlphaToCoverageEnable = FALSE;
    BlendEnable[0] = FALSE;
};

BlendState AdditiveBlending
{
    AlphaToCoverageEnable = FALSE;
    BlendEnable[0] = TRUE;
    SrcBlend = ONE;
    DestBlend = ONE;
    BlendOp = ADD;
    SrcBlendAlpha = ZERO;
    DestBlendAlpha = ZERO;
    BlendOpAlpha = ADD;
    RenderTargetWriteMask[0] = 0x0F;	 // ALL
};

DepthStencilState DisableDepth
{
	DepthEnable = FALSE;
	DepthWriteMask = ZERO;
}; 

//--------------------------------------------------------------------------------------
// Textures
//--------------------------------------------------------------------------------------

Texture3D velocity;
Texture3D density;
Texture3D obstacles;
Texture3D vorticity;

//--------------------------------------------------------------------------------------
// Vertex Shaders
//--------------------------------------------------------------------------------------

GS_Pos4Tex3SliceID VS_Through(VS_Pos3Tex3 input) 
{
	GS_Pos4Tex3SliceID output;
	output.Position = float4(input.Position, 1.0);
	output.SliceIndex = input.Texcoord.z;
	output.Texcoord = input.Texcoord;
	output.Texcoord.z = (output.Texcoord.z + 0.5) / simulationDims.z;
	return output; 
}

//--------------------------------------------------------------------------------------
// Geometry Shaders
//--------------------------------------------------------------------------------------

[maxvertexcount (3)]
void GS_TriangleThrough(triangle GS_Pos4Tex3SliceID input[3], inout TriangleStream<PS_Pos4Tex3SliceID> triStream)
{
    PS_Pos4Tex3SliceID output;
    output.SliceIndex = input[0].SliceIndex;
    for(int v=0; v<3; v++)
    {
        output.Position     = input[v].Position; 
        output.Texcoord     = input[v].Texcoord;
        triStream.Append( output );
    }
    triStream.RestartStrip( );
}

[maxvertexcount (1)]
void GS_PointThrough(point GS_Pos4Tex3SliceID input[1], inout PointStream<PS_Pos4Tex3SliceID> pointStream)
{
    PS_Pos4Tex3SliceID output;
    output.SliceIndex = input[0].SliceIndex;
    output.Position     = input[0].Position; 
    output.Texcoord     = input[0].Texcoord;
    pointStream.Append( output );
    pointStream.RestartStrip( );
}

//--------------------------------------------------------------------------------------
// Helper Functions
//--------------------------------------------------------------------------------------

bool IsBoundaryCell(float3 tex)
{
	return obstacles.SampleLevel(samPoint, tex, 0).x > 0.0;
}

//--------------------------------------------------------------------------------------
// Pixel Shaders
//--------------------------------------------------------------------------------------

float4 PS_Color(PS_Pos4Tex3SliceID input, uniform float4 color) : SV_TARGET
{
	return color;
}

//--------------------------------------------------------------------------------------
// Render white
//--------------------------------------------------------------------------------------

technique10 WhiteTechnique
{
	pass
	{
        SetVertexShader(CompileShader( vs_4_0, VS_Through() ));
        SetGeometryShader( CompileShader( gs_4_0, GS_TriangleThrough() ));
        SetPixelShader( CompileShader( ps_4_0, PS_Color(float4(1,1,1,1)) ));

        SetBlendState( NoBlending, float4( 0.0f, 0.0f, 0.0f, 0.0f ), 0xFFFFFFFF );                
	}
}

//--------------------------------------------------------------------------------------
// Render black
//--------------------------------------------------------------------------------------

technique10 BlackTechnique
{
	pass
	{
        SetVertexShader(CompileShader( vs_4_0,  VS_Through() ));
        SetGeometryShader( CompileShader( gs_4_0, GS_TriangleThrough() ));
        SetPixelShader( CompileShader( ps_4_0, PS_Color(float4(0,0,0,1)) ));

        SetBlendState( NoBlending, float4( 0.0f, 0.0f, 0.0f, 0.0f ), 0xFFFFFFFF );                
	}
}

//--------------------------------------------------------------------------------------
// Render fuel
//--------------------------------------------------------------------------------------

float4 PS_RenderFuel(PS_Pos4Tex3SliceID input) : SV_TARGET
{
	float len = length(input.Texcoord.xy);
	
	float f = 1.0f - smoothstep(0.9, 1.1, len);
	
	return float4(f*InitialFuelDensity, 0,0,0);
}

technique10 RenderFuel
{
	pass
	{
        SetVertexShader(CompileShader( vs_4_0,  VS_Through() ));
        SetGeometryShader( CompileShader( gs_4_0, GS_TriangleThrough() ));
        SetPixelShader( CompileShader( ps_4_0,  PS_RenderFuel()   ));

        SetBlendState( AdditiveBlending, float4( 0.0f, 0.0f, 0.0f, 0.0f ), 0xFFFFFFFF );        
	}
}

//--------------------------------------------------------------------------------------
// Render spark
//--------------------------------------------------------------------------------------

float4 PS_RenderSpark(PS_Pos4Tex3SliceID input) : SV_TARGET
{
	return float4(0,0,sparkTemp,0);
}

technique10 RenderSpark
{
	pass
	{
        SetVertexShader(CompileShader( vs_4_0,  VS_Through() ));
        SetGeometryShader( CompileShader( gs_4_0, GS_PointThrough() ));
        SetPixelShader( CompileShader( ps_4_0,  PS_RenderSpark()   ));

        SetBlendState( AdditiveBlending, float4( 0.0f, 0.0f, 0.0f, 0.0f ), 0xFFFFFFFF );        
	}
}

//--------------------------------------------------------------------------------------
// Calculate vorticity
//--------------------------------------------------------------------------------------

float4 PS_CalculateVorticity(PS_Pos4Tex3SliceID input) : SV_TARGET
{
	float4 left = velocity.Sample(samPoint, LEFTCELL);
	float4 right = velocity.Sample(samPoint, RIGHTCELL);
	float4 top = velocity.Sample(samPoint, TOPCELL);
	float4 bottom = velocity.Sample(samPoint, BOTTOMCELL);
	float4 front = velocity.Sample(samPoint, FRONTCELL);
	float4 back = velocity.Sample(samPoint, BACKCELL);

    float4 vorticity;
    vorticity.xyz = 0.5 * float3( (( top.z - bottom.z ) - ( top.y - bottom.y )) ,
                                  (( top.x - bottom.x ) - ( right.z - left.z )) ,
                                  (( right.y - left.y ) - ( top.x - bottom.x )) );
                                 
	vorticity.w = length(vorticity.xyz);
	return vorticity;
}

technique10 CalculateVorticityTechnique
{
	pass
	{
        SetVertexShader(CompileShader( vs_4_0,  VS_Through() ));
        SetGeometryShader( CompileShader( gs_4_0, GS_TriangleThrough() ));
        SetPixelShader( CompileShader( ps_4_0,  PS_CalculateVorticity()   ));
        
        SetBlendState( NoBlending, float4( 0.0f, 0.0f, 0.0f, 0.0f ), 0xFFFFFFFF );                
	}
}

//--------------------------------------------------------------------------------------
// Explosion specifics
//--------------------------------------------------------------------------------------

float3 CalculateDensityChanges(float3 den)
{
	float3 ret = float3(0,0,0);
	if (den.z > LowerIgnitionLimit) {
		float C = BurningRate * den.x;
		ret.x = -C;
		ret.y = C * (StoichiometricMixture + 1);
		ret.z = C * OutputHeat;// * StoichiometricMixture;
	}
	ret -= Dissipation * den;
	return ret;
}

float3 CalcVorticityForces(PS_Pos4Tex3SliceID input)
{
    float4 center	 = vorticity.SampleLevel(samPoint, input.Texcoord, 0);

	float leftLen	= vorticity.SampleLevel(samPoint, LEFTCELL, 0).w;
	float rightLen	= vorticity.SampleLevel(samPoint, RIGHTCELL, 0).w;
	float topLen	= vorticity.SampleLevel(samPoint, TOPCELL, 0).w;
	float bottomLen = vorticity.SampleLevel(samPoint, BOTTOMCELL, 0).w;
	float frontLen	= vorticity.SampleLevel(samPoint, FRONTCELL, 0).w;
	float backLen	= vorticity.SampleLevel(samPoint, BACKCELL, 0).w;

    float3 eta = 0.5 * float3( rightLen - leftLen,
                              topLen - bottomLen,
                              frontLen - backLen );

    eta = normalize( eta + float3(0.001,0.001,0.001) );

    float4 force;
    force.xyz = VorticityConfinementStrength * float3(	eta.y * center.z - eta.z * center.y,
														eta.z * center.x - eta.x * center.z,
														eta.x * center.y - eta.y * center.x );
    
    // Note: the result is added to the current velocity at each cell using "additive blending"
    return force;
}

float3 CalculateVelocityChanges(float3 density, float3 velocity, PS_Pos4Tex3SliceID input) 
{
	float3 force  = float3(0, density.z * Buoyancy - density.y * Gravity, 0);
	force += CalcVorticityForces(input);
	return force;
}

ExplosionOutput PS_ExplosionSpecific(PS_Pos4Tex3SliceID input)
{	
	ExplosionOutput output = (ExplosionOutput)0;
	
	if (IsBoundaryCell(input.Texcoord)) 
		return output;

	
	float4 density = density.SampleLevel(samPoint, input.Texcoord, 0);
	float4 velocity = velocity.SampleLevel(samPoint, input.Texcoord, 0);
	
	
	float3 densityChanges = CalculateDensityChanges(density);
	float3 velocityChanges = CalculateVelocityChanges(density, velocity, input);
	
	output.Density.xyz = density + TimeStep * densityChanges;
	output.Velocity.xyz = velocity + TimeStep * velocityChanges;
	output.ThermalExpansion = ThermalExpansionRate * densityChanges.z;	
	
	return output;
}

technique10 ExplosionSpecificsTechnique
{
	pass
	{
        SetVertexShader(CompileShader( vs_4_0,  VS_Through() ));
        SetGeometryShader( CompileShader( gs_4_0, GS_TriangleThrough() ));
        SetPixelShader( CompileShader( ps_4_0,  PS_ExplosionSpecific()   ));

        SetBlendState( NoBlending, float4( 0.0f, 0.0f, 0.0f, 0.0f ), 0xFFFFFFFF );                
	}
}
