#include "stdafx.h"

#include "VortexParticleOperation3d.h"
#include "VertexDefinition.h"
#include "TechniqueOperation.h"
#include "EffectManager.h"
#include "TextureStore3D.h"
#include "VertexBufferUtils.h"
#include "ExplosionParameters.h"
#include "MathUtils.h"
#include <algorithm>

VortexParticleOperation3d::VortexParticleOperation3d(ID3D10Device* device, TextureStore3D* textureStore, int velocityID)
:   vortexCount(9),
    drawFromVortexBuffer(NULL),
    streamOutToVortexBuffer(NULL),
    inputLayout(NULL),
    effectName("ExplosionSimulator3d.fx"),
    textureStore(textureStore),
    velocityID(velocityID)
{
    if (vortexCount <= 0) return;

    InitMoveVorticesOperation(device);
    InitRenderVorticesOperation(device);
    InitVortexParticleBuffers(device);
}

VortexParticleOperation3d::~VortexParticleOperation3d()
{
    SAFE_RELEASE(drawFromVortexBuffer);
    SAFE_RELEASE(streamOutToVortexBuffer);
    SAFE_RELEASE(inputLayout);
}


void VortexParticleOperation3d::InitMoveVorticesOperation(ID3D10Device* device)
{
    ID3D10Effect* effect = EffectManager::Instance().GetEffect(effectName, device);
    moveParticlesTechnique = auto_ptr<TechniqueOperation>(
                                    new TechniqueOperation(effect, "MoveVorticesTechnique", textureStore, device));
    moveParticlesTechnique->AddSetShaderResourceTask("velocity", velocityID);
}

void VortexParticleOperation3d::InitRenderVorticesOperation(ID3D10Device* device)
{
    ID3D10Effect* effect = EffectManager::Instance().GetEffect(effectName, device);
    renderParticlesTechnique = auto_ptr<TechniqueOperation>(
                                    new TechniqueOperation(effect, "RenderVorticesTechnique", textureStore, device));
}

void VortexParticleOperation3d::InitVortexParticleBuffers(ID3D10Device* device)
{
    VertexPositionRadiusStrength* vortices = new VertexPositionRadiusStrength[vortexCount];

    const float minX = 0.7f;
    const float maxX = 0.9f;
    const float minY = -0.1f;
    const float maxY = 0.1f;

    const float texelSize = 1.0f / ExplosionParameters::Instance().GetSimulationWidth();
    const int radius = 4;


    for (int i = 0; i < this->vortexCount; i++) {
        vortices[i] = VertexPositionRadiusStrength(D3DXVECTOR3(MathUtils::randomBetween(minX,maxX),MathUtils::randomBetween(minY,maxY),0), radius*texelSize, 100);
    }

    this->drawFromVortexBuffer = VertexBufferUtils::CreateDefaultVertexBuffer(device, vortices, this->vortexCount, true);
    this->streamOutToVortexBuffer = VertexBufferUtils::CreateSOVertexBuffer<VertexPositionRadiusStrength>(
                                            device, this->vortexCount);

    this->inputLayout = VertexBufferUtils::CreateInputLayout<VertexPositionRadiusStrength>
        (device, renderParticlesTechnique->GetFirstPass());

    delete [] vortices;
}


void VortexParticleOperation3d::Run(float timestep, ID3D10Device* device) {

    if (vortexCount <= 0) return;

    // <Common to all draw calls in this method>
    device->IASetInputLayout(inputLayout);
    device->IASetPrimitiveTopology(D3D10_PRIMITIVE_TOPOLOGY_POINTLIST);
    // </Common to all draw calls in this method>


    // <Render the contents in the drawFromBuffer into the streamOutToBuffer after moving the vortices>
    device->OMSetRenderTargets(0,0,0);
    VertexBufferUtils::SetSingleVertexBuffer<VertexPositionRadiusStrength>(device, drawFromVortexBuffer);
    VertexBufferUtils::SetSingleSOVertexBuffer(device, streamOutToVortexBuffer);
    moveParticlesTechnique->Execute(vortexCount, device);

    device->SOSetTargets(0, 0, 0);
    std::swap(drawFromVortexBuffer, streamOutToVortexBuffer);
    // </Render the contents in the drawFromBuffer into the streamOutToBuffer after moving the vortices>

    textureStore->SetAsRenderTarget(velocityID, false, device);
    VertexBufferUtils::SetSingleVertexBuffer<VertexPositionRadiusStrength>(device, drawFromVortexBuffer);
    renderParticlesTechnique->Execute(vortexCount, device);
}
