#include "stdafx.h"

#include "DXUT.h"
#include "SDKmisc.h"

#include "EffectManager.h"
#include "ExplosionParameters.h"
#include "ObstacleVisualizer.h"
#include "TextureUtils.h"
#include "VertexBufferUtils.h"
#include "VertexDefinition.h"

ObstacleVisualizer::ObstacleVisualizer(ISimulationResultProvider* simulationResultProvider, ICameraProvider* cameraProvider)
	:
	simulationResultProvider(simulationResultProvider),
	cameraProvider(cameraProvider),
	vertexBuffer(NULL),
	inputLayout(NULL),
	renderObstacleTechnique(NULL),
	worldMatrixVariable(NULL),
	viewMatrixVariable(NULL),
	projectionMatrixVariable(NULL),
	stoneWallTextureView(NULL)
{
}

ObstacleVisualizer::~ObstacleVisualizer()
{
}

bool ObstacleVisualizer::LoadGeneralGraphicsContent(ID3D10Device* device, const DXGI_SURFACE_DESC* bufferSurfaceDesc)
{
	InitShader(device);
	InitBuffer(device);

    viewport.Width = bufferSurfaceDesc->Width;
    viewport.Height = bufferSurfaceDesc->Height;
    viewport.MinDepth = 0;
    viewport.MaxDepth = 1;
    viewport.TopLeftX = 0;
    viewport.TopLeftY = 0;

	return true;
}

void ObstacleVisualizer::ReleaseGeneralGraphicsContent()
{
	SAFE_RELEASE(inputLayout);
	SAFE_RELEASE(vertexBuffer);
	SAFE_DELETE(renderObstacleTechnique);

	SAFE_RELEASE_BASE_RESOURCE(stoneWallTextureView);
	SAFE_RELEASE(stoneWallTextureView);
}

void ObstacleVisualizer::Draw(ID3D10Device* device, double time, float elapsedTime)
{
	device->RSSetViewports(1, &viewport);
	ID3D10RenderTargetView* RTV = DXUTGetD3D10RenderTargetView();
	ID3D10DepthStencilView* DSV = DXUTGetD3D10DepthStencilView();
	device->OMSetRenderTargets(1,&RTV, DSV);

	device->IASetInputLayout(inputLayout);
	device->IASetPrimitiveTopology(D3D10_PRIMITIVE_TOPOLOGY_POINTLIST);
	VertexBufferUtils::SetSingleVertexBuffer<VertexPosition>(device, vertexBuffer);

	{
		float yScale = static_cast<float>(ExplosionParameters::Instance().GetSimulationHeight()) / 
										 ExplosionParameters::Instance().GetSimulationWidth();

		D3DXMATRIX temp;
		D3DXMATRIX temp2;

		D3DXMatrixScaling(&temp, 2,2*yScale,2);
		D3DXMatrixTranslation(&temp2, 4, -2, 0);
		D3DXMatrixMultiply(&world, &temp, &temp2);
	}

	worldMatrixVariable->SetMatrix((float*)world);
	viewMatrixVariable->SetMatrix((float*)cameraProvider->GetCamera().GetViewMatrix());
	projectionMatrixVariable->SetMatrix((float*)cameraProvider->GetCamera().GetProjMatrix());

	renderObstacleTechnique->Execute(vertexCount, device);
}

bool ObstacleVisualizer::InitShader(ID3D10Device* device)
{
	ID3D10Effect* effect = EffectManager::Instance().GetEffect("ObstacleVisualizer.fx", device);
	if (effect == NULL) return false;

	TextureStore* textureStore = simulationResultProvider->GetTextureStore();

	renderObstacleTechnique = new TechniqueOperation(effect, "RenderObstacleTechnique", textureStore, device);
	renderObstacleTechnique->AddSetShaderResourceTask("obstacleTexture", simulationResultProvider->GetObstaclesID());

	worldMatrixVariable = effect->GetVariableByName("World")->AsMatrix();
	viewMatrixVariable = effect->GetVariableByName("View")->AsMatrix();
	projectionMatrixVariable = effect->GetVariableByName("Projection")->AsMatrix();
	effect->GetVariableByName("texelSizes")->AsVector()->SetFloatVector(
		ExplosionParameters::Instance().GetTexelSizes());


	int depth = ExplosionParameters::Instance().GetSimulationDepth();
	effect->GetVariableByName("MinTextureCoords")->AsVector()->SetFloatVector(
		(float*) D3DXVECTOR3(-0.3f, -1.0f, 1.0f - 2.0f * 0.9f)
	);
	effect->GetVariableByName("MaxTextureCoords")->AsVector()->SetFloatVector(
		(float*) D3DXVECTOR3(0.3f, -0.5f, 1.0f - 2.0f * 0.1f)
	);

	{
		WCHAR str[MAX_PATH];
		DXUTFindDXSDKMediaFileCch(str, MAX_PATH, L"..\\StoneWallTexture.bmp");

		D3DX10CreateShaderResourceViewFromFile(device, str, NULL, NULL, &stoneWallTextureView);
		effect->GetVariableByName("stoneWallTexture")->AsShaderResource()->SetResource(stoneWallTextureView);
	}

	D3DXMatrixIdentity(&world);
	worldMatrixVariable->SetMatrix((float*)&world);

	return true;
}

bool ObstacleVisualizer::InitBuffer(ID3D10Device* device)
{
	int width  = ExplosionParameters::Instance().GetSimulationWidth();
	int height = ExplosionParameters::Instance().GetSimulationHeight();
	int depth  = ExplosionParameters::Instance().GetSimulationDepth();

	float invWidth  = 1.0f / width;
	float invHeight = 1.0f / height;
	float invDepth  = 1.0f / depth;

	float yScale = height * invWidth;

	vertexCount = width * height * depth;

	VertexPosition* vertices = new VertexPosition[vertexCount];
	for (int z = 0; z < depth; z++) {
		for (int y = 0; y < height; y++) {
			for (int x = 0; x < width; x++) {
				float xPos = -1.0f + (2.0f * x) * invWidth;
				float yPos = -1.0f + (2.0f * y) * invHeight;
				float zPos = -1.0f + (2.0f * z) * invDepth;
				vertices[z*width*height + y*width + x] = VertexPosition(D3DXVECTOR3(xPos, yPos, zPos));
			}
		}
	}

	vertexBuffer = VertexBufferUtils::CreateDefaultVertexBuffer(device, vertices, vertexCount);
	if (vertexBuffer == NULL) return false;

	inputLayout = VertexBufferUtils::CreateInputLayout<VertexPosition>(device, renderObstacleTechnique->GetFirstPass());
	if (inputLayout == NULL) return false;

	delete [] vertices;

	return true;
}
