#include "stdafx.h"

#include "ParticleVisualizer.h"
#include "TextureStore.h"
#include "EffectManager.h"
#include "TechniqueOperation.h"
#include "VertexDefinition.h"
#include "VertexBufferUtils.h"
#include "ExplosionParameters.h"
#include "ICameraProvider.h"
#include "BlackBodyRadiationManager.h"
#include "TextureUtils.h"
#include "MathUtils.h"
#include "DXUT.h"
#include "SDKmisc.h"

#include "ExplosionParameters.h"
#include "ExplosionSimulator3d.h"
#include "TextureStore3D.h"

#pragma warning (disable:4995)
#include <sstream>
#include <iostream>
using namespace std;


ParticleVisualizer::ParticleVisualizer(ParticleSystemType type, ISimulationResultProvider* simulator, ICameraProvider* cameraProvider)
    :
    particleSystemType(type),
    simulator(simulator),
    advanceParticlesTechnique(NULL), renderParticlesTechnique(NULL),
    drawFromBuffer(NULL), streamOutToBuffer(NULL), inputLayout(NULL),
    cameraProvider(cameraProvider), turbulenceInterpolatorCounter(0),
    turbulenceVolumeView1(NULL), turbulenceVolumeView2(NULL), globalTime(0.0f), frame(0),
    isInitialized(false)
{
    switch (particleSystemType)
    {
    case Fire:
        {
            effectFileName = "ParticleVisualizer.fx";
            advanceParticlesTechniqueName = "AdvanceParticles";
            renderParticlesTechniqueName = "RenderFireParticles";

            initialParticleCount = 20000;
            maxParticleCount = 20000;

            break;
        }
    case Smoke:
        {
            effectFileName = "ParticleVisualizer.fx";
            advanceParticlesTechniqueName = "AdvanceParticles";
            renderParticlesTechniqueName = "RenderSmokeParticles";

            initialParticleCount = 20000;
            maxParticleCount = 20000;

            break;
        }
    }
}

void ParticleVisualizer::InitStuff(ID3D10Device* device) {
    if (isInitialized) return;
    isInitialized = true;

    InitEffectVariables(device);
    InitKolmogorovTurbulenceTextures(device);
    InitRandomValuesTable(device);
    InitAdvanceParticlesTechnique(device);
    InitRenderParticlesTechnique(device);
    InitBuffers(device);
    InitBlackBodyRadiationTable(device);
    InitTextures(device);

}

bool ParticleVisualizer::LoadGeneralGraphicsContent(ID3D10Device* device, const DXGI_SURFACE_DESC *bufferSurfaceDesc)
{
    isInitialized = false;
    int defineCount = 0;
    if (ExplosionParameters::Instance().ShouldSimulate3D()) defineCount++;
    if (ExplosionParameters::Instance().ShouldRenderDebuggedParticles()) defineCount++;

    if (defineCount > 0)
    {
        D3D10_SHADER_MACRO *shader_Macros = new D3D10_SHADER_MACRO[defineCount+1];
        shader_Macros[defineCount].Name = NULL;
        shader_Macros[defineCount].Definition = NULL;
        int index = 0;
        if (ExplosionParameters::Instance().ShouldSimulate3D()) {
            shader_Macros[index].Name = "SAMPLE3D";
            shader_Macros[index].Definition = "1";
            index++;
        }
        if (ExplosionParameters::Instance().ShouldRenderDebuggedParticles()) {
            shader_Macros[index].Name = "RENDERDEBUG_POINTS";
            shader_Macros[index].Definition = "1";
            index++;
        }
        EffectManager::Instance().PrepareEffectWithMacros(this->effectFileName, device, shader_Macros);
        delete [] shader_Macros;
    }

    vp.Width = bufferSurfaceDesc->Width;
    vp.Height = bufferSurfaceDesc->Height;
    vp.MinDepth = 0;
    vp.MaxDepth = 1;
    vp.TopLeftX = 0;
    vp.TopLeftY = 0;

    return true;
}




void ParticleVisualizer::InitEffectVariables(ID3D10Device* device)
{
    ID3D10Effect* effect = EffectManager::Instance().GetEffect(this->effectFileName, device);
    worldViewProjectionMatrixVariable = effect->GetVariableByName("WorldViewPerspective")->AsMatrix();
    turbulenceInterpolationVariable = effect->GetVariableByName("turbulenceInterpolator")->AsScalar();
    timeVariable = effect->GetVariableByName("globalTime")->AsScalar();

    effect->GetVariableByName("TimeStep")->AsScalar()->SetFloat(ExplosionParameters::Instance().GetTimeStep());
    effect->GetVariableByName("TexelSizes")->AsVector()->SetFloatVector(ExplosionParameters::Instance().GetTexelSizes());

    renderToTextureID = simulator->GetTextureStore()->AddTexture("Render To", device);
}

void ParticleVisualizer::InitAdvanceParticlesTechnique(ID3D10Device* device)
{
    ID3D10Effect* effect = EffectManager::Instance().GetEffect(this->effectFileName, device);
    TextureStore* textureStore = simulator->GetTextureStore();
    advanceParticlesTechnique = new TechniqueOperation(effect, this->advanceParticlesTechniqueName, textureStore, device);

    advanceParticlesTechnique->AddSetShaderResourceTask("velocity", simulator->GetVelocityResultID());
}

void ParticleVisualizer::InitRenderParticlesTechnique(ID3D10Device* device)
{
    ID3D10Effect* effect = EffectManager::Instance().GetEffect(this->effectFileName, device);
    TextureStore* textureStore = simulator->GetTextureStore();
    renderParticlesTechnique = new TechniqueOperation(effect, this->renderParticlesTechniqueName, textureStore, device);

    renderParticlesTechnique->AddSetShaderResourceTask("density", simulator->GetDensityResultID());
}

void ParticleVisualizer::InitBlackBodyRadiationTable(ID3D10Device* device)
{
    ID3D10Effect* effect = EffectManager::Instance().GetEffect(this->effectFileName, device);
    ID3D10EffectShaderResourceVariable* texture = effect->GetVariableByName("ColorLookUpTexture")->AsShaderResource();

    const D3DXVECTOR4* colors = BlackbodyRadiationManager::Instance().GetAllColors();
    colorLookupTextureID = simulator->GetTextureStore()->AddTexture1D(
        "Blackbody Radiation", device, 256, colors
    );
    simulator->GetTextureStore()->SetAsShaderResource1D(colorLookupTextureID, texture);
}

void ParticleVisualizer::InitRandomValuesTable(ID3D10Device* device)
{
    ID3D10Effect* effect = EffectManager::Instance().GetEffect(this->effectFileName, device);
    ID3D10EffectShaderResourceVariable* textureVariable = effect->GetVariableByName("RandomValuesTexture")->AsShaderResource();

    const int numberCount = 1024;

    D3DXVECTOR4* values = new D3DXVECTOR4[numberCount];

    for (int i = 0; i < numberCount; i++) {
        values[i] = D3DXVECTOR4((MathUtils::randNormalized()+1)/2,
                                (MathUtils::randNormalized()+1)/2,
                                (MathUtils::randNormalized()+1)/2,
                                (MathUtils::randNormalized()+1)/2);
    }

    int randomValuesTextureID = simulator->GetTextureStore()->AddTexture1D(
        "Random Values Texture", device, numberCount, values
    );
    simulator->GetTextureStore()->SetAsShaderResource1D(randomValuesTextureID, textureVariable);

    delete [] values;
}

void ParticleVisualizer::InitKolmogorovTurbulenceTextures(ID3D10Device* device)
{
    WCHAR str[MAX_PATH];
    ID3D10Effect* effect;

    DXUTFindDXSDKMediaFileCch( str, MAX_PATH, L"..\\TurbulenceTexture2.dds" );
    D3DX10CreateShaderResourceViewFromFile( device, str, NULL, NULL, &turbulenceVolumeView1 );
    effect = EffectManager::Instance().GetEffect(this->effectFileName, device);
    effect->GetVariableByName("TurbulenceTexture1")->AsShaderResource()->SetResource(turbulenceVolumeView1);

    DXUTFindDXSDKMediaFileCch( str, MAX_PATH, L"..\\TurbulenceTexture2.dds" );
    D3DX10CreateShaderResourceViewFromFile( device, str, NULL, NULL, &turbulenceVolumeView2 );
    effect = EffectManager::Instance().GetEffect(this->effectFileName, device);
    effect->GetVariableByName("TurbulenceTexture2")->AsShaderResource()->SetResource(turbulenceVolumeView2);
}


void ParticleVisualizer::InitTextures(ID3D10Device* device)
{
    // Load the Particle Texture
    {
        WCHAR str[MAX_PATH];
        ID3D10Effect* effect = EffectManager::Instance().GetEffect(this->effectFileName, device);

        switch (particleSystemType)
        {
        case Fire:
            DXUTFindDXSDKMediaFileCch( str, MAX_PATH, L"..\\AnimatedExplosionParticleTexture2.dds" );
            D3DX10CreateShaderResourceViewFromFile( device, str, NULL, NULL, &volumeTextureView );
            effect->GetVariableByName("fireParticleTexture")->AsShaderResource()->SetResource(volumeTextureView);
            break;
        case Smoke:
            DXUTFindDXSDKMediaFileCch( str, MAX_PATH, L"..\\AnimatedSmokeParticleTexture2.dds" );
            D3DX10CreateShaderResourceViewFromFile( device, str, NULL, NULL, &volumeTextureView );
            effect->GetVariableByName("smokeParticleTexture")->AsShaderResource()->SetResource(volumeTextureView);
            break;
        }
    }
}

void ParticleVisualizer::InitBuffers(ID3D10Device* device)
{
    VertexPositionVelocityRMass* particles = new VertexPositionVelocityRMass[initialParticleCount];

    InitRandomParticlePositions(particles, initialParticleCount);
    this->CreateBuffers(particles, device);

    delete [] particles;
}

void ParticleVisualizer::InitRandomParticlePositions(VertexPositionVelocityRMass * particles, int particleCount)
{
    float maxRadius = ExplosionParameters::Instance().initialParticleAreaRadius;
    //float maxRadius = 0.30f;

    float baseX = (ExplosionParameters::Instance().GetSparkPositionX() + 1.0f) / 2.0f;
    float baseY = 1 - (ExplosionParameters::Instance().GetSparkPositionY() + 1.0f) / 2.0f;
    float baseZ = ExplosionParameters::Instance().GetSparkPositionZ() * 2.0f - 1.0f;

    baseY -= 0.0f;

    D3DXVECTOR3 center(baseX, baseY, baseZ);
    D3DXVECTOR3 velocity(0,0,0);
    float minMass = 100;
    float maxMass = 1000;

    // Debug-stasj
    float minX = 999;
    float maxX = -999;
    float minZ = 999;
    float maxZ = -999;

    for (int i = 0; i < particleCount; i++) {
        float radius = MathUtils::randomBetween(0, maxRadius);

        D3DXVECTOR3 vec(MathUtils::randNormalized(),
                        MathUtils::randNormalized(),
                        MathUtils::randNormalized());
        D3DXVECTOR3 norm;
        D3DXVec3Normalize(&norm, &vec);
        D3DXVec3Scale(&vec, &norm, radius);
        D3DXVec3Add(&norm, &vec, &center);

        float r = MathUtils::randomBetween(0,100);
        float mass;

        ExplosionParameters& p = ExplosionParameters::Instance();
        switch (particleSystemType)
        {
        case Fire:
            {
                if (r < p.fireParticleMassLightPercentage) {
                    mass = MathUtils::randomBetween(p.fireParticleMassLightBeginValue,
                                                    p.fireParticleMassLightEndValue);
                } else {
                    mass = MathUtils::randomBetween(p.fireParticleMassHeavyBeginValue,
                                                    p.fireParticleMassHeavyEndValue);
                }
            }
            break;
        case Smoke:
            {
                if (r < p.smokeParticleMassLightPercentage) {
                    mass = MathUtils::randomBetween(p.smokeParticleMassLightBeginValue,
                                                    p.smokeParticleMassLightEndValue);
                } else {
                    mass = MathUtils::randomBetween(p.smokeParticleMassHeavyBeginValue,
                                                    p.smokeParticleMassHeavyEndValue);
                }
            }
            break;
        }

        //norm.y = baseY - 0.2;

        // Debug-stasj
        minX = min(minX, norm.x);
        maxX = max(maxX, norm.x);
        minZ = min(minZ, norm.z);
        maxZ = max(maxZ, norm.z);

        float rad = MathUtils::randomBetween(0, 2*3.1415f);
        velocity.x = rad;

        particles[i] = VertexPositionVelocityRMass(norm, velocity, 1.0f / mass);
    }
}

void ParticleVisualizer::ReleaseGeneralGraphicsContent()
{
    SAFE_DELETE(renderParticlesTechnique);
    SAFE_DELETE(advanceParticlesTechnique);
    SAFE_RELEASE(inputLayout);
    SAFE_RELEASE(drawFromBuffer);
    SAFE_RELEASE(streamOutToBuffer);
    SAFE_RELEASE_BASE_RESOURCE(volumeTextureView);
    SAFE_RELEASE(volumeTextureView);
    SAFE_RELEASE_BASE_RESOURCE(turbulenceVolumeView1);
    SAFE_RELEASE(turbulenceVolumeView1);
    SAFE_RELEASE_BASE_RESOURCE(turbulenceVolumeView2);
    SAFE_RELEASE(turbulenceVolumeView2);
}

void ParticleVisualizer::Draw(ID3D10Device* device, double time, float elapsedTime)
{
    InitStuff(device);

    if (this->simulator->ShouldSimulate()) AdvanceParticles(device, time);
    RenderParticles(device);
}

void ParticleVisualizer::HandleTurbulenceInterpolation()
{
    float interpolationValue;
    int floor = (int)turbulenceInterpolatorCounter;
    if (floor%2 == 0) {
        interpolationValue = turbulenceInterpolatorCounter - floor;
    } else {
        interpolationValue = 1 - (turbulenceInterpolatorCounter - floor);
    }

    this->turbulenceInterpolationVariable->SetFloat(interpolationValue);

    turbulenceInterpolatorCounter += 0.001f;
}

void ParticleVisualizer::AdvanceParticles(ID3D10Device* device, double time)
{
    globalTime += 1.0f/60.0f;
    HandleTurbulenceInterpolation();

    // <Common to all draw calls in this method>
    device->IASetInputLayout(inputLayout);
    device->IASetPrimitiveTopology(D3D10_PRIMITIVE_TOPOLOGY_POINTLIST);
    // </Common to all draw calls in this method>

    // <Render the contents in the drawFromBuffer into the streamOutToBuffer after moving the vortices>
    //timeVariable->SetFloat((float)time);
    timeVariable->SetFloat(globalTime);
    device->OMSetRenderTargets(0,0,0);
    VertexBufferUtils::SetSingleVertexBuffer<VertexPositionVelocityRMass>(device, drawFromBuffer);
    VertexBufferUtils::SetSingleSOVertexBuffer(device, streamOutToBuffer);
    advanceParticlesTechnique->ExecuteAuto(device);

    device->SOSetTargets(0,0,0);
    std::swap(drawFromBuffer, streamOutToBuffer);
    // </Render the contents in the drawFromBuffer into the streamOutToBuffer after moving the vortices>
}

void ParticleVisualizer::RenderParticles(ID3D10Device* device)
{
    device->RSSetViewports( 1, &vp );

    ID3D10RenderTargetView* RTV = DXUTGetD3D10RenderTargetView();
    ID3D10DepthStencilView* DSV = DXUTGetD3D10DepthStencilView();
    device->OMSetRenderTargets(1,&RTV, DSV);

    //simulator->GetTextureStore()->SetAsRenderTarget(renderToTextureID, false, device);

    //float ClearColor[4] = { 0.3f, 0.3f, 0.3f, 0.3f }; // red, green, blue, alpha
    //device->ClearRenderTargetView( RTV, ClearColor );

    //ID3D10DepthStencilView* pDSV = DXUTGetD3D10DepthStencilView();
    //device->ClearDepthStencilView( pDSV, D3D10_CLEAR_DEPTH, 1.0, 0 );

    D3DXMATRIX worldViewProj;
    D3DXMATRIX temp;
    D3DXMATRIX temp2;

    const D3DXMATRIX* projection;
    projection = cameraProvider->GetCamera().GetProjMatrix();

    const D3DXMATRIX* view;
    view = cameraProvider->GetCamera().GetViewMatrix();

    float scale = static_cast<float>(ExplosionParameters::Instance().GetSimulationHeight()) /
                                     ExplosionParameters::Instance().GetSimulationWidth();

    D3DXMATRIX world;
    D3DXMatrixScaling(&temp, 4,4*scale,4);
    D3DXMatrixTranslation(&temp2, 2, -2*scale-2, 0);
    D3DXMatrixMultiply(&world, &temp, &temp2);

    D3DXMatrixMultiply(&temp, &world, view);
    D3DXMatrixMultiply(&worldViewProj, &temp, projection);
    worldViewProjectionMatrixVariable->SetMatrix((float*)worldViewProj);

    // <Render particles to the back-buffer>
    device->IASetInputLayout(inputLayout);
    device->IASetPrimitiveTopology(D3D10_PRIMITIVE_TOPOLOGY_POINTLIST);
    VertexBufferUtils::SetSingleVertexBuffer<VertexPositionVelocityRMass>(device, drawFromBuffer);

    //D3D10_QUERY_DESC queryDesc;
    //queryDesc.MiscFlags = 0;
    //queryDesc.Query = D3D10_QUERY_PIPELINE_STATISTICS;
    //ID3D10Query * query;
    //device->CreateQuery(&queryDesc, &query);
    //query->Begin();

    renderParticlesTechnique->ExecuteAuto(device);

    //query->End();

    //D3D10_QUERY_DATA_PIPELINE_STATISTICS queryData;

    //while( S_OK != query->GetData(&queryData, sizeof(D3D10_QUERY_DATA_PIPELINE_STATISTICS), 0) )
    //{
    //}
    //
    //
    //if (frame == 250 && particleSystemType == Smoke)
    //{
    //  UINT64 pixels = queryData.PSInvocations;
    //  pixels = pixels;

    //  stringstream str;
    //  str << "Pixel shader invocations: " << pixels << endl;
    //  string s = str.str();
    //  OutputDebugStringA(s.c_str());
    //  cout << s;

    //}
    //
    //frame++;
    //
    //query->Release();


    // </Render particles to the back-buffer>
}
