//--------------------------------------------------------------------------------------
// Structures
//--------------------------------------------------------------------------------------

struct VS_Pos3
{
	float3 Pos : POS;
};

struct GS_Pos4Tex3
{
	float4 Pos : SV_POSITION;
	float3 Tex : TEXCOORD;
};

struct PS_Pos4Tex3
{
	float4 Pos : SV_POSITION;
	float3 Tex : TEXCOORD;
};

//--------------------------------------------------------------------------------------
// States
//--------------------------------------------------------------------------------------

BlendState NoBlending
{
    AlphaToCoverageEnable = FALSE;
    BlendEnable[0] = FALSE;
};

DepthStencilState EnableDepth
{
    DepthEnable = TRUE;
    DepthWriteMask = ALL;
};

DepthStencilState DisableDepth
{
    DepthEnable = FALSE;
    DepthWriteMask = ZERO;
};

DepthStencilState DisableDepthWrite
{
    DepthEnable = TRUE;
    DepthWriteMask = ZERO;
};

DepthStencilState DisableDepthTest
{
    DepthEnable = TRUE;
    DepthWriteMask = ALL;
    DepthFunc = ALWAYS;
};

SamplerState samPoint
{
    Filter = MIN_MAG_MIP_POINT;
    AddressU = Clamp;
    AddressV = Clamp;
    AddressW = Clamp;    
};

SamplerState samLinear
{
    Filter = MIN_MAG_MIP_LINEAR;
    AddressU = Wrap;
    AddressV = Wrap;
    AddressW = Wrap;
};

//--------------------------------------------------------------------------------------
// Buffers
//--------------------------------------------------------------------------------------

cbuffer cbNeverChanges
{
    float3 MinTextureCoords;
    float3 MaxTextureCoords;
    float3 texelSizes;
};

cbuffer cbChangePerFrame
{
    matrix View;
	matrix World;
};

cbuffer cbChangeOnResize
{
    matrix Projection;
};

//--------------------------------------------------------------------------------------
// Textures
//--------------------------------------------------------------------------------------

Texture3D obstacleTexture;
Texture2D stoneWallTexture;

//--------------------------------------------------------------------------------------
// Helper Functions
//--------------------------------------------------------------------------------------

bool IsBoundaryCell(float3 tex)
{
	return tex.x > 0.0f && tex.x < 0.9f && 
	       tex.y > 0.0f && tex.y < 0.9f && 
	       tex.z > 0.0f && tex.z < 0.9f && 
	       obstacleTexture.SampleLevel(samPoint, tex, 0).x > 0.0;
}

//--------------------------------------------------------------------------------------
// Shaders
//--------------------------------------------------------------------------------------

GS_Pos4Tex3 VS(VS_Pos3 input)
{
	GS_Pos4Tex3 output = (GS_Pos4Tex3) 0;
	output.Tex = (input.Pos + 1) / 2;
	output.Pos = mul(float4(input.Pos, 1.0), World);
    output.Pos = mul(output.Pos, View);
    output.Pos = mul(output.Pos, Projection);
	return output;
}

[MaxVertexCount(24)]
void GS(point GS_Pos4Tex3 input[1], inout TriangleStream<PS_Pos4Tex3> stream)
{
	if (IsBoundaryCell(input[0].Tex)) {
		float size = 2.0f / 32.0f;
//		float doubleSize = 4 * 0.03125f;
		float doubleSize = 4 * texelSizes.x;
		

		float2 upperLeft  = float2(0.0f, size);
		float2 upperRight = float2(size, size);
		float2 lowerLeft  = float2(0.0f, 0.0f);
		float2 lowerRight = float2(size, 0.0f);

		float4 upperLeftNear  = float4(0.0f,       doubleSize, 0.0f,       0.0f);
		float4 upperRightNear = float4(doubleSize, doubleSize, 0.0f,       0.0f);
		float4 lowerLeftNear  = float4(0.0f,       0.0f,       0.0f,       0.0f);
		float4 lowerRightNear = float4(doubleSize, 0.0f,       0.0f,       0.0f);
		float4 upperLeftFar   = float4(0.0f,       doubleSize, doubleSize, 0.0f);
		float4 upperRightFar  = float4(doubleSize, doubleSize, doubleSize, 0.0f);
		float4 lowerLeftFar   = float4(0.0f,       0.0f,       doubleSize, 0.0f);
		float4 lowerRightFar  = float4(doubleSize, 0.0f,       doubleSize, 0.0f);

		upperLeftNear  = mul(upperLeftNear,  View);
		upperRightNear = mul(upperRightNear, View);
		lowerLeftNear  = mul(lowerLeftNear,  View);
		lowerRightNear = mul(lowerRightNear, View);
		upperLeftFar   = mul(upperLeftFar,  View);
		upperRightFar  = mul(upperRightFar, View);
		lowerLeftFar   = mul(lowerLeftFar,  View);
		lowerRightFar  = mul(lowerRightFar, View);

		upperLeftNear  = mul(upperLeftNear,  Projection);
		upperRightNear = mul(upperRightNear, Projection);
		lowerLeftNear  = mul(lowerLeftNear,  Projection);
		lowerRightNear = mul(lowerRightNear, Projection);
		upperLeftFar   = mul(upperLeftFar,  Projection);
		upperRightFar  = mul(upperRightFar, Projection);
		lowerLeftFar   = mul(lowerLeftFar,  Projection);
		lowerRightFar  = mul(lowerRightFar, Projection);

//        float3 pos = (input[0].Tex * 2.0f) - 1.0f;
        float3 tex = input[0].Tex;
//		tex.x = (pos.x - MinTextureCoords.x) / (MaxTextureCoords.x - MinTextureCoords.x);
//		tex.y = (pos.y - MinTextureCoords.y) / (MaxTextureCoords.y - MinTextureCoords.y);
//		tex.z = (pos.z - MinTextureCoords.z) / (MaxTextureCoords.z - MinTextureCoords.z);

		
		tex = (tex - float3(0.3, 0.3, 0.3)) / float3(0.4, 0.4, 0.4);
		
		
		float3 inc = texelSizes / float3(0.4, 0.4, 0.4);

		PS_Pos4Tex3 output = (PS_Pos4Tex3) 0;
		output.Tex = tex;


		output.Pos = input[0].Pos + upperLeftNear;
//		output.Tex = tex.xy + upperLeft * float2(1.75f, 1.0f);
		output.Tex.xy = tex.xy + float2(0, inc.y);
		stream.Append(output);
		
		output.Pos = input[0].Pos + upperRightNear;
		output.Tex.xy = tex.xy + float2(inc.x, inc.y);
		//output.Tex = tex.xy + upperRight * float2(1.75f, 1.0f);
		stream.Append(output);

		output.Pos = input[0].Pos + lowerLeftNear;
		output.Tex.xy = tex.xy + float2(0, 0);
		//output.Tex = tex.xy + lowerLeft * float2(1.75f, 1.0f);
		stream.Append(output);
		
		output.Pos = input[0].Pos + lowerRightNear;
		output.Tex.xy = tex.xy + float2(inc.x, 0);
		//output.Tex = tex.xy + lowerRight * float2(1.75f, 1.0f);
		stream.Append(output);

		stream.RestartStrip();

		output.Pos = input[0].Pos + lowerLeftNear;
		//output.Tex = tex.xz;
		stream.Append(output);
		
		output.Pos = input[0].Pos + lowerRightNear;
		//output.Tex = tex.xz;
		stream.Append(output);

		output.Pos = input[0].Pos + lowerLeftFar;
		//output.Tex = tex.xz;
		stream.Append(output);

		output.Pos = input[0].Pos + lowerRightFar;
		//output.Tex = tex.xz;
		stream.Append(output);

		stream.RestartStrip();

		output.Pos = input[0].Pos + upperRightFar;
		//output.Tex = tex.xy;
		stream.Append(output);

		output.Pos = input[0].Pos + upperLeftFar;
		//output.Tex = tex.xy;
		stream.Append(output);

		output.Pos = input[0].Pos + lowerRightFar;
		//output.Tex = tex.xy;
		stream.Append(output);

		output.Pos = input[0].Pos + lowerLeftFar;
		//output.Tex = tex.xy;
		stream.Append(output);

		stream.RestartStrip();

		output.Pos = input[0].Pos + upperLeftFar;
		output.Tex.xy = tex.xz + float2(0, 0);
		stream.Append(output);

		output.Pos = input[0].Pos + upperRightFar;
		//output.Tex = tex.xz + upperRight * float2(1.75f, 0.65f);
		output.Tex.xy = tex.xz + float2(inc.x, 0);
		stream.Append(output);

		output.Pos = input[0].Pos + upperLeftNear;
		//output.Tex = tex.xz + lowerLeft * float2(1.75f, 0.65f);
		output.Tex.xy = tex.xz + float2(0, inc.z);
		stream.Append(output);

		output.Pos = input[0].Pos + upperRightNear;
		//output.Tex = tex.xz + lowerRight * float2(1.75f, 0.65f);
		output.Tex.xy = tex.xz + float2(inc.x, inc.z);
		stream.Append(output);

		stream.RestartStrip();

		output.Pos = input[0].Pos + upperLeftFar;
		output.Tex.xy = tex.zy + float2(inc.z, inc.y);
		//output.Tex = tex.yz;
		stream.Append(output);

		output.Pos = input[0].Pos + upperLeftNear;
		output.Tex.xy = tex.zy + float2(0, inc.y);
		//output.Tex = tex.yz;
		stream.Append(output);

		output.Pos = input[0].Pos + lowerLeftFar;
		output.Tex.xy = tex.zy + float2(inc.z, 0);
		//output.Tex = tex.yz;
		stream.Append(output);

		output.Pos = input[0].Pos + lowerLeftNear;
		output.Tex.xy = tex.zy + float2(0, 0);
		//output.Tex = tex.yz;
		stream.Append(output);

		stream.RestartStrip();

		output.Pos = input[0].Pos + upperRightNear;
		//output.Tex = tex.yz;
		stream.Append(output);

		output.Pos = input[0].Pos + upperRightFar;
		//output.Tex = tex.yz;
		stream.Append(output);

		output.Pos = input[0].Pos + lowerRightNear;
		//output.Tex = tex.yz;
		stream.Append(output);

		output.Pos = input[0].Pos + lowerRightFar;
		//output.Tex = tex.yz;
		stream.Append(output);

		stream.RestartStrip();
	}
}

float4 PS(PS_Pos4Tex3 input) : SV_Target
{
	return stoneWallTexture.Sample(samLinear, input.Tex.xy);
	return float4(input.Tex.xy, 0,0);
}

technique10 RenderObstacleTechnique
{
	pass
	{
		SetVertexShader( CompileShader(vs_4_0, VS()) );
		SetGeometryShader( CompileShader(gs_4_0, GS()) );
		SetPixelShader( CompileShader(ps_4_0, PS()) );

        SetDepthStencilState( EnableDepth, 0 );
        SetBlendState( NoBlending, float4( 0.0f, 0.0f, 0.0f, 0.0f ), 0xFFFFFFFF );
	}
}
