//--------------------------------------------------------------------------------------
// Structures
//--------------------------------------------------------------------------------------


// Comment out if the texture used for smoke particles are 2D
#define SMOKE_TEXTURE_3D
#define FIRE_TEXTURE_3D


// SAMPLE3D gets defined programatically from within the c++ code
//#define SAMPLE3D





#define PI 3.1415926535

struct VS_Pos3
{
	float3 Pos : POS;
	float3 Velocity : VELOCITY;
	float rMass : RMASS;
};

struct GS_Pos4
{
	float4 RenderPos : SV_POSITION;
	float3 SamplePos : POSITION;
	float3 Velocity : VELOCITY;
	float rMass : RMASS;
};

struct PS_Pos4
{
	float4 Position : SV_POSITION;
	float2 Tex : TEXCOORD;
	float4 StaticColor : COLOR;
	float AnimationOffset : SIM_OFFSET;
};

//--------------------------------------------------------------------------------------
// States
//--------------------------------------------------------------------------------------

SamplerState samPoint
{
    AddressU = Clamp;
    AddressV = Clamp;
};

SamplerState samLinear
{
	Filter = MIN_MAG_MIP_LINEAR;
    AddressU = Clamp;
    AddressV = Clamp;
};

SamplerState samLinear3D
{
	Filter = MIN_MAG_MIP_LINEAR;
    AddressU = Clamp;
    AddressV = Clamp;
    AddressW = Clamp;
};

SamplerState samLinearWrap
{
	Filter = MIN_MAG_MIP_LINEAR;
    AddressU = Wrap;
    AddressV = Wrap;
    AddressW = Wrap;
};

SamplerState samPointWrap
{
    AddressU = Wrap;
};

SamplerState samVolume
{
    Filter = MIN_MAG_LINEAR_MIP_POINT;
    AddressU = Wrap;
    AddressV = Wrap;
    AddressW = Wrap;
};

DepthStencilState EnableDepth
{
    DepthEnable = TRUE;
    DepthWriteMask = ALL;
};

DepthStencilState DisableDepth
{
    DepthEnable = FALSE;
    DepthWriteMask = ZERO;
};

DepthStencilState DisableDepthWrite
{
    DepthEnable = TRUE;
    DepthWriteMask = ZERO;
};

DepthStencilState DisableDepthTest
{
    DepthEnable = TRUE;
    DepthWriteMask = ALL;
    DepthFunc = ALWAYS;
};

BlendState AdditiveBlending
{
    AlphaToCoverageEnable = FALSE;
    BlendEnable[0] = TRUE;
    SrcBlend = ONE;
    DestBlend = INV_SRC_COLOR;
    BlendOp = ADD;
    SrcBlendAlpha = ZERO;
    DestBlendAlpha = ZERO;
    BlendOpAlpha = ADD;
    RenderTargetWriteMask[0] = 0x0F;
};

BlendState AlphaBlending
{
    BlendEnable[0] = TRUE;
    SrcBlend = SRC_ALPHA;
    DestBlend = INV_SRC_ALPHA;
    BlendOp = ADD;
    SrcBlendAlpha = ZERO;
    DestBlendAlpha = ZERO;
    BlendOpAlpha = ADD;
};

BlendState SmokeBlending
{
    BlendEnable[0] = TRUE;
    SrcBlend = SRC_ALPHA;
    DestBlend = INV_SRC_ALPHA;
    BlendOp = ADD;
    SrcBlendAlpha = SRC_ALPHA;
    DestBlendAlpha = INV_SRC_ALPHA;
    BlendOpAlpha = ADD;
};

BlendState DefaultAlphaBlending
{
    AlphaToCoverageEnable = FALSE;
    BlendEnable[0] = TRUE;
    RenderTargetWriteMask[0] = 0x0F;
};

BlendState NoBlending
{
    AlphaToCoverageEnable = FALSE;
    BlendEnable[0] = FALSE;
};

//--------------------------------------------------------------------------------------
// Textures
//--------------------------------------------------------------------------------------

#ifdef SAMPLE3D
	Texture3D density;
	Texture3D velocity;
#else
	Texture2D density;
	Texture2D velocity;
#endif

Texture1D ColorLookUpTexture;
Texture1D RandomValuesTexture;

#ifdef FIRE_TEXTURE_3D
	Texture3D fireParticleTexture;
#else 
	Texture2D fireParticleTexture;
#endif

#ifdef SMOKE_TEXTURE_3D
	Texture3D smokeParticleTexture;
#else 
	Texture2D smokeParticleTexture;
#endif

Texture3D TurbulenceTexture1;
Texture3D TurbulenceTexture2;

//--------------------------------------------------------------------------------------
// Constant Buffers
//--------------------------------------------------------------------------------------

cbuffer cbNeverChanges{
	float TimeStep;
	float3 TexelSizes;
};

cbuffer cbChangesPerFrame
{
	float globalTime;
	float turbulenceInterpolator;
};

cbuffer cbChangesPerPrimitive
{
	matrix WorldViewPerspective;
	
	matrix World;
	matrix View;
	matrix Projection;
};


//--------------------------------------------------------------------------------------
// Rendering
//--------------------------------------------------------------------------------------

GS_Pos4 VS_Render(VS_Pos3 input)
{
	GS_Pos4 output;
	output.SamplePos = input.Pos;
	output.RenderPos = mul(float4(input.Pos, 1), WorldViewPerspective);
	output.Velocity = input.Velocity;
	output.rMass = input.rMass;
	return output;
}



float4 SampleDensitySpherical(float3 pos, Texture2D source)
{
	float2 tex;
	tex.y = pos.y;
	
	float2 vec = pos.xz - float2(0.5, 0);
	float len = length(vec);
	float2 dir = normalize(vec);
	
	tex.x = 0.5 - len;
	float4 left = source.SampleLevel(samLinear, tex, 0);
	
	tex.x = 0.5 + len;
	float4 right = source.SampleLevel(samLinear, tex, 0);
	
	return lerp(left, right, (1+dir.x)*0.5);
}


float4 SampleDensity(float3 pos) 
{
#ifdef SAMPLE3D
	pos.z = (pos.z + 1.0) / 2.0;
	return density.SampleLevel(samLinear3D, pos, 0); 
#else
	return SampleDensitySpherical(pos, density);
#endif
}


[MaxVertexCount(4)]
void GS_FireRender(point GS_Pos4 input[1], inout TriangleStream<PS_Pos4> outputStream)
{
	float4 densities = SampleDensity(input[0].SamplePos);

	PS_Pos4 output;

	// The color is constant over the quad for now
	float4 color = ColorLookUpTexture.SampleLevel(samLinear, densities.z*0.2f, 0);
	//color *= saturate(densities.y);
	//color *= smoothstep(0, 1, densities.y);
	color *= smoothstep(0.0, 1.3, densities.z);

	output.StaticColor = color * 3.8; 
	output.AnimationOffset = input[0].Velocity.x;
#ifdef RENDERDEBUG_POINTS
	float radius = 0.05;
#else
	float radius = 0.8;//*4.5;
#endif


	float rot = input[0].Velocity.x;
	float deg;
	
	deg = 3*PI/4+rot;
	output.Position = input[0].RenderPos + float4(radius*cos(deg), radius*sin(deg), 0, 0);
	output.Tex = float2(0,1);
	outputStream.Append(output);

	deg = PI/4+rot;
	output.Position = input[0].RenderPos + float4(radius*cos(deg), radius*sin(deg), 0, 0);
	output.Tex = float2(1,1);
	outputStream.Append(output);

	deg = 5*PI/4+rot;
	output.Position = input[0].RenderPos + float4(radius*cos(deg), radius*sin(deg), 0, 0);
	output.Tex = float2(0,0);
	outputStream.Append(output);

	deg = 7*PI/4+rot;
	output.Position = input[0].RenderPos + float4(radius*cos(deg), radius*sin(deg), 0, 0);
	output.Tex = float2(1,0);
	outputStream.Append(output);

/*
	output.Position = input[0].RenderPos + float4(-radius, radius, 0, 0);
	output.Tex = float2(0,0);
	outputStream.Append(output);

	output.Position = input[0].RenderPos + float4( radius, radius, 0, 0);
	output.Tex = float2(1,0);
	outputStream.Append(output);

	output.Position = input[0].RenderPos + float4(-radius,-radius, 0, 0);
	output.Tex = float2(0,1);
	outputStream.Append(output);

	output.Position = input[0].RenderPos + float4( radius,-radius, 0, 0);
	output.Tex = float2(1,1);
	outputStream.Append(output);
*/
	outputStream.RestartStrip();
}

[MaxVertexCount(4)]
void GS_SmokeRender(point GS_Pos4 input[1], inout TriangleStream<PS_Pos4> outputStream)
{
	float4 densities = SampleDensity(input[0].SamplePos);

	PS_Pos4 output;

	float exhaust = densities.y;
	float temp = densities.z;
	
	
	float tempScalar = 1 - smoothstep(0.0f, 4, temp);
	float intensity = smoothstep(0.0f, 0.25, exhaust) * tempScalar / 8.8175f;
	intensity *= smoothstep(0.2, 0.8, globalTime);

	float3 color = 0.6 + 1.4*ColorLookUpTexture.SampleLevel(samLinear, temp*0.1f, 0);
	
	output.StaticColor.xyz = color;
	output.StaticColor.a = intensity;
	output.AnimationOffset = input[0].Velocity.x;


#ifdef RENDERDEBUG_POINTS
	float radius = 0.05;
#else
	float radius = 1.2;//*1.4142;
	//float radius = 1.2*1.5;
#endif


	float rot = input[0].Velocity.x;
	float deg;
	
	deg = 3*PI/4+rot;
	output.Position = input[0].RenderPos + float4(radius*cos(deg), radius*sin(deg), 0, 0);
	output.Tex = float2(0,1);
	outputStream.Append(output);

	deg = PI/4+rot;
	output.Position = input[0].RenderPos + float4(radius*cos(deg), radius*sin(deg), 0, 0);
	output.Tex = float2(1,1);
	outputStream.Append(output);

	deg = 5*PI/4+rot;
	output.Position = input[0].RenderPos + float4(radius*cos(deg), radius*sin(deg), 0, 0);
	output.Tex = float2(0,0);
	outputStream.Append(output);

	deg = 7*PI/4+rot;
	output.Position = input[0].RenderPos + float4(radius*cos(deg), radius*sin(deg), 0, 0);
	output.Tex = float2(1,0);
	outputStream.Append(output);

/*
	output.Position = input[0].RenderPos + float4(-radius, radius, 0, 0);
	output.Tex = float2(0,1);
	outputStream.Append(output);

	output.Position = input[0].RenderPos + float4( radius, radius, 0, 0);
	output.Tex = float2(1,1);
	outputStream.Append(output);

	output.Position = input[0].RenderPos + float4(-radius,-radius, 0, 0);
	output.Tex = float2(0,0);
	outputStream.Append(output);

	output.Position = input[0].RenderPos + float4( radius,-radius, 0, 0);
	output.Tex = float2(1,0);
	outputStream.Append(output);
*/

	outputStream.RestartStrip();
}

float4 PS_FireRender(PS_Pos4 input) : SV_Target
{
#ifdef FIRE_TEXTURE_3D
	float4 sample = fireParticleTexture.Sample(samVolume, float3(input.Tex, globalTime/35));
#else
	float4 sample = fireParticleTexture.Sample(samVolume, input.Tex);
#endif

#ifdef RENDERDEBUG_POINTS
	return float4(1,0,0,0);
#else
	return input.StaticColor * sample * sample.a;
#endif
}

float4 PS_SmokeRender(PS_Pos4 input) : SV_Target
{	
#ifdef SMOKE_TEXTURE_3D
	float4 sample = smokeParticleTexture.Sample(samVolume, float3(input.Tex, input.AnimationOffset + globalTime/35));
#else
	float4 sample = smokeParticleTexture.Sample(samVolume, input.Tex);
#endif

#ifdef RENDERDEBUG_POINTS
	return 1;
#else
	return sample * input.StaticColor;
#endif

}

technique10 RenderFireParticles
{
	pass P0
	{
		SetVertexShader( CompileShader( vs_4_0, VS_Render() ) );
		SetGeometryShader( CompileShader( gs_4_0, GS_FireRender() ) );
		SetPixelShader( CompileShader( ps_4_0, PS_FireRender() ) );

        SetBlendState( AdditiveBlending, float4( 0.0f, 0.0f, 0.0f, 0.0f ), 0xFFFFFFFF );
//        SetDepthStencilState( DisableDepth, 0 );
        SetDepthStencilState( DisableDepthWrite, 0 );
	}
}

technique10 RenderSmokeParticles
{
	pass P0
	{
		SetVertexShader( CompileShader( vs_4_0, VS_Render() ) );
		SetGeometryShader( CompileShader( gs_4_0, GS_SmokeRender() ) );
		SetPixelShader( CompileShader( ps_4_0, PS_SmokeRender() ) );

        SetBlendState( AlphaBlending, float4( 0.0f, 0.0f, 0.0f, 0.0f ), 0xFFFFFFFF );
//        SetDepthStencilState( DisableDepth, 0 );
        SetDepthStencilState( DisableDepthWrite, 0 );

	}
}

//--------------------------------------------------------------------------------------
// Advancement
//--------------------------------------------------------------------------------------

float3 RandomDir(float offset)
{
	return RandomValuesTexture.SampleLevel(samPointWrap, globalTime + offset, 0).xyz;
}

/*
VS_Pos3 VS_Advance2D(VS_Pos3 input)
{

	float2 fieldVelocity = velocity.SampleLevel(samLinear, max(min(input.Pos.xy, 1-TexelSizes), TexelSizes), 0);
	input.Pos.xy += fieldVelocity * TimeStep * input.rMass;
	return input;
*/

/*
	// Similar to "Animating Suspended Particle Explosion"
	float2 fieldVelocity = velocity.SampleLevel(samLinear, max(min(input.Pos.xy, 1-TexelSizes), TexelSizes), 0);
	float2 force = 1.0f * (fieldVelocity-input.Velocity.xy) * length(fieldVelocity-input.Velocity.xy);		
	input.Velocity.xy += force * input.rMass * TimeStep;
	input.Pos.xy += input.Velocity * TimeStep;
	return input;
*/
/*
	float2 force = velocity.SampleLevel(samLinear, max(min(input.Pos.xy, 1-TexelSizes), TexelSizes), 0);
	input.Velocity.xy *= 0.99f;
	input.Velocity.xy += force * input.rMass * TimeStep;
	input.Pos.xyz += input.Velocity * TimeStep;
	return input;
*/
//}


float3 SampleVelocitySpherical(Texture2D source, float3 position) 
{
	float2 texcoord;
	texcoord.y = position.y;	

	float2 vec = position.xz - float2(0.5f, 0);
	float len = length(vec);
	float2 dir = normalize(vec);	
	
	texcoord.x = 0.5f - len;
	float2 leftSide = source.SampleLevel(samLinear, max(min(texcoord, 1-TexelSizes.xy), TexelSizes.xy), 0);
	leftSide.x *= -1;

	texcoord.x = 0.5f + len;
	float2 rightSide = source.SampleLevel(samLinear, max(min(texcoord, 1-TexelSizes.xy), TexelSizes.xy), 0);
	
	float2 value2d = lerp(leftSide, rightSide, (1 + dir.x)*0.5f);

	float3 value;
	value.y = value2d.y;
	value.xz = value2d.x * dir;

	return value;
}

float3 SampleVelocity(float3 pos)
{
#ifdef SAMPLE3D
	pos.z = (pos.z + 1.0) / 2.0;
	return velocity.SampleLevel(samLinear3D, pos, 0); 
#else
	return SampleVelocitySpherical(velocity, pos);
#endif
}

float3 GetTurbulenceVelocity(float3 pos) 
{
	pos /= 1.1;
	float4 left = TurbulenceTexture1.SampleLevel(samLinearWrap, pos, 0);
	float4 right = TurbulenceTexture2.SampleLevel(samLinearWrap, pos, 0);
	return lerp(left, right, turbulenceInterpolator).xyz;
}

VS_Pos3 VS_Advance3D(VS_Pos3 input)
{
	float3 fieldVelocity = SampleVelocity(input.Pos);
	
	input.Pos.xyz += fieldVelocity * TimeStep * TexelSizes.x * input.rMass;

	float3 turbis = GetTurbulenceVelocity(input.Pos) * TimeStep * input.rMass * TexelSizes.x * 10;
	turbis *= smoothstep(0, 5.0, length(fieldVelocity));
	input.Pos.xyz += turbis;

	return input;
	
/*
	// Similar to "Animating Suspended Particle Explosion"
	float3 force = 0.1f * (fieldVelocity-input.Velocity) * length(fieldVelocity-input.Velocity);		
	input.Velocity += force * input.rMass * TimeStep;
	input.Pos.xyz += input.Velocity;
	return input;
*/
/*
	// Old non-physical stuff
	input.Velocity -= 0.15f * input.Velocity * length(input.Velocity);
	input.Velocity += fieldVelocity * input.rMass * TimeStep;
	input.Pos.xyz += input.Velocity;
	return input;
*/

}


[MaxVertexCount(1)]
void GS_AdvanceParticles(point VS_Pos3 input[1], inout PointStream<VS_Pos3> outputStream) 
{
	if (input[0].Pos.x > 1  || input[0].Pos.x < 0 ||
		input[0].Pos.y > 1  || input[0].Pos.y < 0)
	{
/*
		float3 newPos = RandomDir(0); 
		input[0].Pos = newPos;
		input[0].Velocity = SampleSpherical(velocity, newPos);
*/
	}
	else
	outputStream.Append(input[0]);
}

GeometryShader advanceGS = ConstructGSWithSO( 
									CompileShader( gs_4_0, GS_AdvanceParticles() ), 
									"POS.xyz; VELOCITY.xyz; RMASS.x" );
technique10 AdvanceParticles
{
	pass P0
	{
		SetVertexShader( CompileShader( vs_4_0, VS_Advance3D() ) );
		SetGeometryShader( advanceGS );
		SetPixelShader( NULL );
		
        SetDepthStencilState(DisableDepth, 0);
	}
}

