#include "stdafx.h"

#include "DensityOperation2d.h"
#include "EffectManager.h"
#include "TechniqueOperation.h"
#include "HlslTextureOperation2d.h"
#include "ExplosionParameters.h"

DensityOperation2d::DensityOperation2d(ID3D10Device* device, TextureStore* textureStore, 
                                       int densityID, int velocityID,
                                       int densitySourcesID)
	:
	textureStore(textureStore),
	densityID(densityID),
	velocityID(velocityID),
	densitySourcesID(densitySourcesID),
	jacobiIterations(ExplosionParameters::Instance().GetDensityDiffusionJacobiIterations()),
	effectName("FluidSolver2d.fx")
{
    InitAddDensitySourcesOperation(device);
	InitAdvectionOperation(device);
	InitDiffusionOperation(device);
	InitEffectVariables(device);
}

DensityOperation2d::~DensityOperation2d()
{
}

void DensityOperation2d::InitAddDensitySourcesOperation(ID3D10Device* device)
{
    ID3D10Effect* effect = EffectManager::Instance().GetEffect(effectName, device);
    
	addDensitySourcesTechnique = auto_ptr<TechniqueOperation>(
	    new TechniqueOperation(effect, "AddDensitySourcesTechnique", textureStore, device)
	);
	addDensitySourcesOperation = auto_ptr<HlslTextureOperation>(
	    new HlslTextureOperation2d(
	        device,
	        ExplosionParameters::Instance().GetSimulationWidth(),
	        ExplosionParameters::Instance().GetSimulationHeight(),
	        addDensitySourcesTechnique.get(), 
	        addDensitySourcesTechnique.get(), 
	        USE_DISPLACED_BORDER
	    )
	);

	addDensitySourcesTechnique->AddSetShaderResourceTask("density", densityID);
	addDensitySourcesTechnique->AddSetShaderResourceTask("densitySources", densitySourcesID);

	addDensitySourcesOperation->SetOutputTextureID(densityID);
}

void DensityOperation2d::InitAdvectionOperation(ID3D10Device* device)
{
	ID3D10Effect* effect = EffectManager::Instance().GetEffect(effectName, device);
	
	advectionTechnique = auto_ptr<TechniqueOperation>(
	    new TechniqueOperation(effect, "AdvectionTechnique", textureStore, device)
	);
	advectionOperation = auto_ptr<HlslTextureOperation>(
	    new HlslTextureOperation2d(
	        device,
	        ExplosionParameters::Instance().GetSimulationWidth(),
	        ExplosionParameters::Instance().GetSimulationHeight(),
	        advectionTechnique.get(), 
	        advectionTechnique.get(), 
	        USE_DISPLACED_BORDER
	    )
	);

	advectionTechnique->AddSetShaderResourceTask("velocity", velocityID);
	advectionTechnique->AddSetShaderResourceTask("advectionSource", densityID);

	advectionOperation->SetOutputTextureID(densityID);
}

void DensityOperation2d::InitDiffusionOperation(ID3D10Device* device)
{
	diffusedDensityID = textureStore->AddTexture("Temp texture for the diffusion iterations", device);
	textureStore->ZeroTexture(diffusedDensityID, device);

	ID3D10Effect* effect = EffectManager::Instance().GetEffect(effectName, device);
	diffusionTechnique = auto_ptr<TechniqueOperation>(
									new TechniqueOperation(effect, "DiffusionTechnique", textureStore, device));

	diffusionOperation = auto_ptr<HlslTextureOperation>(new HlslTextureOperation2d(device,
		ExplosionParameters::Instance().GetSimulationWidth(),
		ExplosionParameters::Instance().GetSimulationHeight(),
		diffusionTechnique.get(), diffusionTechnique.get(), USE_DISPLACED_BORDER));

	diffusionTechnique->AddSetShaderResourceTask("diffusedDensity", diffusedDensityID);
	diffusionTechnique->AddSetShaderResourceTask("density", densityID);
	diffusionOperation->SetOutputTextureID(diffusedDensityID);
}

void DensityOperation2d::InitEffectVariables(ID3D10Device* device)
{
	ID3D10Effect* effect = EffectManager::Instance().GetEffect(effectName, device);
	ExplosionParameters& p = ExplosionParameters::Instance();

	float diffusionAlpha[4];
	float diffusionRBeta[4];

	diffusionAlpha[0] = 1.0f / (p.GetFuelDiffusion() * p.GetTimeStep());
	diffusionAlpha[1] = 1.0f / (p.GetSootDiffusion() * p.GetTimeStep());
	diffusionAlpha[2] = 1.0f / (p.GetTempDiffusion() * p.GetTimeStep());
	diffusionAlpha[3] = 0;

	diffusionRBeta[0] = 1.0f / (4 + diffusionAlpha[0]);
	diffusionRBeta[1] = 1.0f / (4 + diffusionAlpha[1]);
	diffusionRBeta[2] = 1.0f / (4 + diffusionAlpha[2]);
	diffusionRBeta[3] = 1.0f / (4 + diffusionAlpha[3]);

	effect->GetVariableByName("DiffusionAlpha")->AsVector()->SetFloatVector(diffusionAlpha);
	effect->GetVariableByName("DiffusionRBeta")->AsVector()->SetFloatVector(diffusionRBeta);
}

void DensityOperation2d::Run(float timestep, ID3D10Device* device) {
    addDensitySourcesOperation->Execute(device);

	for (int i = 0; i < jacobiIterations-1; i++) {
		diffusionOperation->Execute(device);
	}
	textureStore->ManualSwap(diffusedDensityID, densityID);

	advectionOperation->Execute(device);
}

void DensityOperation2d::SetJacobiIterations(int iterationCount) {
	jacobiIterations = iterationCount;
}
