#include "stdafx.h"

#include "DensityOperation3d.h"
#include "EffectManager.h"
#include "TechniqueOperation.h"
#include "HlslTextureOperation3d.h"
#include "ExplosionParameters.h"

DensityOperation3d::DensityOperation3d(ID3D10Device* device, TextureStore* textureStore, int velocityID, int densityID, int obstaclesID)
	:
	textureStore(textureStore),
	velocityID(velocityID),
	densityID(densityID),
	obstaclesID(obstaclesID),
	jacobiIterations(ExplosionParameters::Instance().GetDensityDiffusionJacobiIterations()),
	effectName("FluidSolver3d.fx")
{
	InitAdvectionOperation(device);
	InitDiffusionOperation(device);
	InitEffectVariables(device);
}

DensityOperation3d::~DensityOperation3d()
{
}


void DensityOperation3d::InitAdvectionOperation(ID3D10Device* device)
{
	ID3D10Effect* effect = EffectManager::Instance().GetEffect(effectName, device);
	advectionTechnique = auto_ptr<TechniqueOperation>(
									new TechniqueOperation(effect, "AdvectionTechnique", textureStore, device));
	advectionOperation = auto_ptr<HlslTextureOperation3d>(new HlslTextureOperation3d(device,
		ExplosionParameters::Instance().GetSimulationWidth(),
		ExplosionParameters::Instance().GetSimulationHeight(),
		ExplosionParameters::Instance().GetSimulationDepth(),
		advectionTechnique.get(), advectionTechnique.get()));

	advectionTechnique->AddSetShaderResourceTask("velocity", velocityID);
	advectionTechnique->AddSetShaderResourceTask("advectionSource", densityID);
	advectionTechnique->AddSetShaderResourceTask("obstacles", obstaclesID);

	advectionOperation->SetOutputTextureID(densityID);
}

void DensityOperation3d::InitDiffusionOperation(ID3D10Device* device)
{
	diffusionSolutionGuessID = textureStore->AddTexture("Temp texture for the diffusion iterations", device);
	textureStore->ZeroTexture(diffusionSolutionGuessID, device);

	ID3D10Effect* effect = EffectManager::Instance().GetEffect(effectName, device);
	diffusionTechnique = auto_ptr<TechniqueOperation>(
									new TechniqueOperation(effect, "DiffusionTechnique", textureStore, device));

	diffusionOperation = auto_ptr<HlslTextureOperation3d>(new HlslTextureOperation3d(device,
		ExplosionParameters::Instance().GetSimulationWidth(),
		ExplosionParameters::Instance().GetSimulationHeight(),
		ExplosionParameters::Instance().GetSimulationDepth(),
		diffusionTechnique.get(), diffusionTechnique.get()));

	diffusionTechnique->AddSetShaderResourceTask("newDensity", diffusionSolutionGuessID);
	diffusionTechnique->AddSetShaderResourceTask("density", densityID);
	diffusionTechnique->AddSetShaderResourceTask("obstacles", obstaclesID);

	diffusionOperation->SetOutputTextureID(diffusionSolutionGuessID);
}

void DensityOperation3d::InitEffectVariables(ID3D10Device* device)
{
	ID3D10Effect* effect = EffectManager::Instance().GetEffect(effectName, device);
	ExplosionParameters& p = ExplosionParameters::Instance();

	float diffusionAlpha[4];
	float diffusionRBeta[4];

	diffusionAlpha[0] = 1.0f / (p.GetFuelDiffusion() * p.GetTimeStep());
	diffusionAlpha[1] = 1.0f / (p.GetSootDiffusion() * p.GetTimeStep());
	diffusionAlpha[2] = 1.0f / (p.GetTempDiffusion() * p.GetTimeStep());
	diffusionAlpha[3] = 0;

	diffusionRBeta[0] = 1.0f / (6 + diffusionAlpha[0]);
	diffusionRBeta[1] = 1.0f / (6 + diffusionAlpha[1]);
	diffusionRBeta[2] = 1.0f / (6 + diffusionAlpha[2]);
	diffusionRBeta[3] = 1.0f / (6 + diffusionAlpha[3]);

	effect->GetVariableByName("DiffusionAlpha")->AsVector()->SetFloatVector(diffusionAlpha);
	effect->GetVariableByName("DiffusionRBeta")->AsVector()->SetFloatVector(diffusionRBeta);
}

void DensityOperation3d::Run(float timestep, ID3D10Device* device) {

	for (int i = 0; i < jacobiIterations-1; i++) {
		diffusionOperation->Execute(device);
	}
	textureStore->ManualSwap(diffusionSolutionGuessID, densityID);

	advectionOperation->Execute(device);

	device->OMSetRenderTargets(0,0,0);
	advectionTechnique->ApplyTechnique();
}

void DensityOperation3d::SetJacobiIterations(int iterationCount) {
	jacobiIterations = iterationCount;
}
