//--------------------------------------------------------------------------------------
// Structures
//--------------------------------------------------------------------------------------

struct VS_Pos3Tex2
{
	float3 Position : POS;
	float2 Texcoord : TEXCOORD0;	
};

struct PS_Pos4Tex2
{
	float4 Position : SV_POSITION;	
	float2 Texcoord : TEXCOORD0;	
};

//--------------------------------------------------------------------------------------
// States
//--------------------------------------------------------------------------------------

SamplerState samPoint
{
    AddressU = Clamp;
    AddressV = Clamp;
};

SamplerState samLinear
{
	Filter = MIN_MAG_MIP_LINEAR;
    AddressU = Clamp;
    AddressV = Clamp;
};

BlendState NoBlending
{
    AlphaToCoverageEnable = FALSE;
    BlendEnable[0] = FALSE;
};

//--------------------------------------------------------------------------------------
// Textures
//--------------------------------------------------------------------------------------

Texture2D density;
Texture2D densitySources;
Texture2D diffusedDensity;

Texture2D velocity;
Texture2D velocityForces;

Texture2D advectionSource;

Texture2D divergence;
Texture2D divergenceForces;

Texture2D pressure;

//--------------------------------------------------------------------------------------
// Constant buffers
//--------------------------------------------------------------------------------------

cbuffer cbNeverChanges{
	float TimeStep;
	float2 texelSizes;

	float4 DiffusionAlpha;
	float4 DiffusionRBeta;

	float4 ProjectionAlpha;
	float4 ProjectionRBeta;
};

//--------------------------------------------------------------------------------------
// General vertex shader
//--------------------------------------------------------------------------------------

PS_Pos4Tex2 VS_General(VS_Pos3Tex2 input) {
	PS_Pos4Tex2 output = (PS_Pos4Tex2)0;
	output.Position = float4(input.Position,1);
	output.Texcoord = input.Texcoord;
	return output;
}

//--------------------------------------------------------------------------------------
// Add velocity forces
//--------------------------------------------------------------------------------------

float4 PS_AddVelocityForces(PS_Pos4Tex2 input) : SV_Target
{
    return velocity.Sample(samPoint, input.Texcoord) + 
           TimeStep * velocityForces.Sample(samPoint, input.Texcoord);
}

technique10 AddVelocityForcesTechnique
{
    pass
    {
		SetVertexShader( CompileShader( vs_4_0, VS_General() ) );
        SetGeometryShader( NULL );
        SetPixelShader( CompileShader( ps_4_0, PS_AddVelocityForces() ) );
        
        SetBlendState( NoBlending, float4( 0.0f, 0.0f, 0.0f, 0.0f ), 0xFFFFFFFF );        
    }
}

//--------------------------------------------------------------------------------------
// Add density sources
//--------------------------------------------------------------------------------------

float4 PS_AddDensitySources(PS_Pos4Tex2 input) : SV_Target
{
    return density.Sample(samPoint, input.Texcoord) + 
           TimeStep * densitySources.Sample(samPoint, input.Texcoord);
}

technique10 AddDensitySourcesTechnique
{
    pass
    {
		SetVertexShader( CompileShader( vs_4_0, VS_General() ) );
        SetGeometryShader( NULL );
        SetPixelShader( CompileShader( ps_4_0, PS_AddDensitySources() ) );
        
        SetBlendState( NoBlending, float4( 0.0f, 0.0f, 0.0f, 0.0f ), 0xFFFFFFFF );        
    }
}

//--------------------------------------------------------------------------------------
// Diffusion
//--------------------------------------------------------------------------------------

float4 PS_DiffJacobi2D( PS_Pos4Tex2 input, 
					uniform float4 alpha, uniform float4 rBeta, 
					uniform Texture2D x, uniform Texture2D b) : SV_Target
{	
	float4 xNeighbourSum = float4(0,0,0,0);
	xNeighbourSum += x.Sample(samPoint, input.Texcoord - float2(texelSizes.x, 0));
	xNeighbourSum += x.Sample(samPoint, input.Texcoord + float2(texelSizes.x, 0));
	xNeighbourSum += x.Sample(samPoint, input.Texcoord - float2(0, texelSizes.y));
	xNeighbourSum += x.Sample(samPoint, input.Texcoord + float2(0, texelSizes.y));

	float4 bCenter = b.Sample(samPoint, input.Texcoord);
	return (alpha * bCenter + xNeighbourSum) * rBeta;
}

technique10 DiffusionTechnique
{
	pass P0 
	{
		SetVertexShader( CompileShader( vs_4_0, VS_General() ) );
        SetGeometryShader( NULL );
        SetPixelShader( CompileShader( ps_4_0, PS_DiffJacobi2D(DiffusionAlpha, DiffusionRBeta, diffusedDensity, density ) ) );

        SetBlendState( NoBlending, float4( 0.0f, 0.0f, 0.0f, 0.0f ), 0xFFFFFFFF );        
	}
}

//--------------------------------------------------------------------------------------
// Advection
//--------------------------------------------------------------------------------------

float4 PS_Advection(PS_Pos4Tex2 input) : SV_Target
{
	float2 pos = input.Texcoord - TimeStep * velocity.Sample(samPoint, input.Texcoord) * texelSizes;
    return advectionSource.Sample(samLinear, pos);
}

float4 PS_AdvectionBorder(PS_Pos4Tex2 input) : SV_Target
{
	return -PS_Advection(input);
}

technique10 AdvectionTechnique
{
	pass P0 
	{
		SetVertexShader( CompileShader( vs_4_0, VS_General() ) );
        SetGeometryShader( NULL );
        SetPixelShader( CompileShader( ps_4_0, PS_Advection() ) );

        SetBlendState( NoBlending, float4( 0.0f, 0.0f, 0.0f, 0.0f ), 0xFFFFFFFF );        
	}
}

technique10 AdvectionBorderTechnique
{
	pass P0 
	{
		SetVertexShader( CompileShader( vs_4_0, VS_General() ) );
        SetGeometryShader( NULL );
        SetPixelShader( CompileShader( ps_4_0, PS_AdvectionBorder() ) );

        SetBlendState( NoBlending, float4( 0.0f, 0.0f, 0.0f, 0.0f ), 0xFFFFFFFF );        
	}
}

//--------------------------------------------------------------------------------------
// Calculate divergence
//--------------------------------------------------------------------------------------

float4 PS_CalculateDivergence(PS_Pos4Tex2 input) : SV_Target
{
    float4 left   = velocity.Sample(samPoint, input.Texcoord - float2(texelSizes.x ,0.0f));
    float4 right  = velocity.Sample(samPoint, input.Texcoord + float2(texelSizes.x ,0.0f));
    float4 bottom = velocity.Sample(samPoint, input.Texcoord - float2(0.0f, texelSizes.y));
    float4 top    = velocity.Sample(samPoint, input.Texcoord + float2(0.0f, texelSizes.y));

	float divergence = ((right.x-left.x) + (top.y-bottom.y)) / 2;

	return divergence - divergenceForces.Sample(samPoint, input.Texcoord).x;
}

technique10 CalculateDivergenceTechnique
{
	pass P0 
	{
		SetVertexShader( CompileShader( vs_4_0, VS_General() ) );
        SetGeometryShader( NULL );
        SetPixelShader( CompileShader( ps_4_0, PS_CalculateDivergence() ) );
        
        SetBlendState( NoBlending, float4( 0.0f, 0.0f, 0.0f, 0.0f ), 0xFFFFFFFF );
	}
}

//--------------------------------------------------------------------------------------
// Calculate pressure
//--------------------------------------------------------------------------------------

float4 PS_PressureJacobi2D( PS_Pos4Tex2 input, 
					uniform float4 alpha, uniform float4 rBeta, 
					uniform Texture2D x, uniform Texture2D b) : SV_Target
{
	float4 xNeighbourSum = float4(0,0,0,0);
	xNeighbourSum += x.Sample(samPoint, input.Texcoord - float2(texelSizes.x, 0));
	xNeighbourSum += x.Sample(samPoint, input.Texcoord + float2(texelSizes.x, 0));
	xNeighbourSum += x.Sample(samPoint, input.Texcoord - float2(0, texelSizes.y));
	xNeighbourSum += x.Sample(samPoint, input.Texcoord + float2(0, texelSizes.y));

	float4 bCenter = b.Sample(samPoint, input.Texcoord);
	return (alpha * bCenter + xNeighbourSum) * rBeta;
}

technique10 CalculatePressureTechnique
{
	pass P0 
	{
		SetVertexShader( CompileShader( vs_4_0, VS_General() ) );
        SetGeometryShader( NULL );
        SetPixelShader( CompileShader( ps_4_0, PS_PressureJacobi2D(ProjectionAlpha, ProjectionRBeta, pressure, divergence) ) );

        SetBlendState( NoBlending, float4( 0.0f, 0.0f, 0.0f, 0.0f ), 0xFFFFFFFF );
	}
}

//--------------------------------------------------------------------------------------
// Subtract divergence
//--------------------------------------------------------------------------------------

float4 PS_SubtractDivergence(PS_Pos4Tex2 input) : SV_Target
{
    float left   = pressure.Sample(samPoint, input.Texcoord - float2(texelSizes.x ,0.0f)).x;
    float right  = pressure.Sample(samPoint, input.Texcoord + float2(texelSizes.x ,0.0f)).x;
    float bottom = pressure.Sample(samPoint, input.Texcoord - float2(0.0f, texelSizes.y)).x;
    float top    = pressure.Sample(samPoint, input.Texcoord + float2(0.0f, texelSizes.y)).x;
    
    float4 vel = velocity.Sample(samPoint, input.Texcoord);
    
    vel.x -= (right - left) / 2;
    vel.y -= (top - bottom) / 2;

	return vel;		
}

technique10 SubtractDivergenceTechnique
{
	pass P0 
	{
		SetVertexShader( CompileShader( vs_4_0, VS_General() ) );
        SetGeometryShader( NULL );
        SetPixelShader( CompileShader( ps_4_0, PS_SubtractDivergence() ) );
        
        SetBlendState( NoBlending, float4( 0.0f, 0.0f, 0.0f, 0.0f ), 0xFFFFFFFF );
	}
}
