#define CENTERCELL input.Texcoord
#define LEFTCELL float3(input.Texcoord.x-texelSizes.x, input.Texcoord.y, input.Texcoord.z)
#define RIGHTCELL float3(input.Texcoord.x+texelSizes.x, input.Texcoord.y, input.Texcoord.z)
#define TOPCELL float3(input.Texcoord.x, input.Texcoord.y+texelSizes.y, input.Texcoord.z)
#define BOTTOMCELL float3(input.Texcoord.x, input.Texcoord.y-texelSizes.y, input.Texcoord.z)
#define BACKCELL float3(input.Texcoord.x, input.Texcoord.y, input.Texcoord.z-texelSizes.z)
#define FRONTCELL float3(input.Texcoord.x, input.Texcoord.y, input.Texcoord.z+texelSizes.z)

//--------------------------------------------------------------------------------------
// Structures
//--------------------------------------------------------------------------------------

struct VS_Pos3Tex3
{
	float3 Position : POS;
	float3 Texcoord : TEXCOORD;	
};

struct GS_Pos4Tex3SliceID
{
	float4 Position : SV_POSITION;	
	float3 Texcoord : TEXCOORD;	
    uint SliceIndex : SLICE;
};

struct PS_Pos4Tex3SliceID
{
	float4 Position : SV_POSITION;	
	float3 Texcoord : TEXCOORD;	
    uint SliceIndex : SV_RenderTargetArrayIndex;
};

//--------------------------------------------------------------------------------------
// States
//--------------------------------------------------------------------------------------

SamplerState samPoint
{
	Filter = MIN_MAG_MIP_POINT;
    AddressU = Clamp;
    AddressV = Clamp;
    AddressW = Clamp;
};

SamplerState samLinear
{
	Filter = MIN_MAG_MIP_LINEAR;
    AddressU = Clamp;
    AddressV = Clamp;
    AddressW = Clamp;
};

BlendState NoBlending
{
    AlphaToCoverageEnable = FALSE;
    BlendEnable[0] = FALSE;
};

//--------------------------------------------------------------------------------------
// Textures
//--------------------------------------------------------------------------------------

Texture3D density;
Texture3D newDensity;
Texture3D velocity;
Texture3D obstacles;

Texture3D advectionSource;
Texture3D thermalExpansion;
Texture3D divergence;
Texture3D poissonSolution;

//--------------------------------------------------------------------------------------
// Constant Buffers
//--------------------------------------------------------------------------------------

cbuffer cbNeverChanges
{
	float3 simulationDims;
	float TimeStep;
	float3 texelSizes;

	float4 DiffusionAlpha;
	float4 DiffusionRBeta;
};

//--------------------------------------------------------------------------------------
// General vertex shader
//--------------------------------------------------------------------------------------

GS_Pos4Tex3SliceID VS_Through(VS_Pos3Tex3 input) 
{
	GS_Pos4Tex3SliceID output;
	output.Position = float4(input.Position, 1.0);
	output.SliceIndex = input.Texcoord.z;
	output.Texcoord = input.Texcoord;
	output.Texcoord.z = (output.Texcoord.z + 0.5) / simulationDims.z;
	return output; 
}

//--------------------------------------------------------------------------------------
// Geometry Shaders
//--------------------------------------------------------------------------------------

[maxvertexcount (3)]
void GS_TriangleThrough(triangle GS_Pos4Tex3SliceID input[3], inout TriangleStream<PS_Pos4Tex3SliceID> triStream)
{
    PS_Pos4Tex3SliceID output;
    output.SliceIndex = input[0].SliceIndex;

    for (int v=0; v<3; v++) {
        output.Position     = input[v].Position; 
        output.Texcoord     = input[v].Texcoord;
        triStream.Append(output);
    }

    triStream.RestartStrip( );
}

//--------------------------------------------------------------------------------------
// Helper Functions
//--------------------------------------------------------------------------------------

bool IsBoundaryCell(float3 tex)
{
	return obstacles.SampleLevel(samPoint, tex, 0).x > 0.0;
}

//--------------------------------------------------------------------------------------
// Diffusion
//--------------------------------------------------------------------------------------

float4 PS_DiffusionSolver(PS_Pos4Tex3SliceID input) : SV_Target
{	
	float4 left = newDensity.Sample(samPoint, LEFTCELL);
	float4 right = newDensity.Sample(samPoint, RIGHTCELL);
	float4 top = newDensity.Sample(samPoint, TOPCELL);
	float4 bottom = newDensity.Sample(samPoint, BOTTOMCELL);
	float4 front = newDensity.Sample(samPoint, FRONTCELL);
	float4 back = newDensity.Sample(samPoint, BACKCELL);
	
	float4 newCenter = newDensity.Sample(samPoint, input.Texcoord);
	float4 oldCenter = density.Sample(samPoint, input.Texcoord);

	if (IsBoundaryCell(LEFTCELL))	left = newCenter;
	if (IsBoundaryCell(RIGHTCELL))	right = newCenter;
	if (IsBoundaryCell(TOPCELL))	top = newCenter;
	if (IsBoundaryCell(BOTTOMCELL))	bottom = newCenter;
	if (IsBoundaryCell(FRONTCELL))	front = newCenter;
	if (IsBoundaryCell(BACKCELL))	back = newCenter;

	float4 sum = (left + right + top + bottom + front + back);
	return (sum + DiffusionAlpha * oldCenter) * DiffusionRBeta;
}

technique10 DiffusionTechnique
{
	pass
	{
        SetVertexShader( CompileShader( vs_4_0, VS_Through() ));
        SetGeometryShader( CompileShader( gs_4_0, GS_TriangleThrough() ));
        SetPixelShader( CompileShader( ps_4_0,  PS_DiffusionSolver() ));

        SetBlendState( NoBlending, float4( 0.0f, 0.0f, 0.0f, 0.0f ), 0xFFFFFFFF );                
	}
}

//--------------------------------------------------------------------------------------
// Advection
//--------------------------------------------------------------------------------------

float4 PS_Advection(PS_Pos4Tex3SliceID input) : SV_Target
{
	if (IsBoundaryCell(input.Texcoord)) return 0;
	
	float3 oldPos = input.Texcoord - TimeStep * velocity.SampleLevel(samPoint, input.Texcoord, 0).xyz * texelSizes;	
	return advectionSource.SampleLevel(samLinear, oldPos, 0);
}

technique10 AdvectionTechnique
{
	pass
	{
        SetVertexShader(CompileShader( vs_4_0,  VS_Through() ));
        SetGeometryShader( CompileShader( gs_4_0, GS_TriangleThrough() ));
        SetPixelShader( CompileShader( ps_4_0,  PS_Advection()   ));

        SetBlendState( NoBlending, float4( 0.0f, 0.0f, 0.0f, 0.0f ), 0xFFFFFFFF );                
	}
}

//--------------------------------------------------------------------------------------
// Calculate divergence
//--------------------------------------------------------------------------------------

float4 PS_CalculateDivergence(PS_Pos4Tex3SliceID input) : SV_Target
{
	float4 left = velocity.Sample(samPoint, LEFTCELL);
	float4 right = velocity.Sample(samPoint, RIGHTCELL);
	float4 top = velocity.Sample(samPoint, TOPCELL);
	float4 bottom = velocity.Sample(samPoint, BOTTOMCELL);
	float4 front = velocity.Sample(samPoint, FRONTCELL);
	float4 back = velocity.Sample(samPoint, BACKCELL);
	
	float4 center = velocity.Sample(samPoint, input.Texcoord);
	
	// To enforce zero flow over boundaries between interior and exterior cells, we use the inverse velocity as at the exterior
	// Cells ABC: Divergence = (C+B)/2 - (B+A)2 = (C-A)/2
	// if C is exterior, the expression (C+B)/2 should evaluate to zero. This can be achieved by setting
	// C = -B, produsing the new expression for divergence (at boundaries): Divergence = (-B - A)/2

	if (IsBoundaryCell(LEFTCELL)) left = center * float4(-1, 1, 1, 1);
	if (IsBoundaryCell(RIGHTCELL)) right = center * float4(-1, 1, 1, 1);
	if (IsBoundaryCell(TOPCELL)) top = center * float4(1, -1, 1, 1);
	if (IsBoundaryCell(BOTTOMCELL)) bottom = center * float4(1, -1, 1, 1);
	if (IsBoundaryCell(FRONTCELL)) front = center * float4(1, 1, -1, 1);
	if (IsBoundaryCell(BACKCELL)) back = center * float4(1, 1, -1, 1);
	
	float divergence = 0.5 * ((right.x-left.x) + (top.y-bottom.y) + (front.z-back.z));
	return divergence - thermalExpansion.Sample(samPoint, input.Texcoord);
}

technique10 CalculateDivergenceTechnique
{
	pass
	{
        SetVertexShader(CompileShader( vs_4_0,  VS_Through() ));
        SetGeometryShader( CompileShader( gs_4_0, GS_TriangleThrough() ));
        SetPixelShader( CompileShader( ps_4_0,  PS_CalculateDivergence()   ));

        SetBlendState( NoBlending, float4( 0.0f, 0.0f, 0.0f, 0.0f ), 0xFFFFFFFF );                
	}
}

//--------------------------------------------------------------------------------------
// Projection
//--------------------------------------------------------------------------------------

float4 PS_PoissonSolver(PS_Pos4Tex3SliceID input) : SV_Target
{
	float left = poissonSolution.Sample(samPoint, LEFTCELL).x;
	float right = poissonSolution.Sample(samPoint, RIGHTCELL).x;
	float top = poissonSolution.Sample(samPoint, TOPCELL).x;
	float bottom = poissonSolution.Sample(samPoint, BOTTOMCELL).x;
	float front = poissonSolution.Sample(samPoint, FRONTCELL).x;
	float back = poissonSolution.Sample(samPoint, BACKCELL).x;

	float center = poissonSolution.Sample(samPoint, input.Texcoord).x;
	float divergence = divergence.Sample(samPoint, input.Texcoord);
	
	// Neumann boundaries (derivatives at the boundaries should be zero, i.e equal to the closest interior cell)
	if (IsBoundaryCell(LEFTCELL))	left = center;
	if (IsBoundaryCell(RIGHTCELL))	right = center;
	if (IsBoundaryCell(TOPCELL))	top = center;
	if (IsBoundaryCell(BOTTOMCELL))	bottom = center;
	if (IsBoundaryCell(FRONTCELL))	front = center;
	if (IsBoundaryCell(BACKCELL))	back = center;
	
	return (left + right + top + bottom + front + back - divergence) / 6.0;
}

technique10 CalculatePressureTechnique
{
	pass
	{
        SetVertexShader(CompileShader( vs_4_0,  VS_Through() ));
        SetGeometryShader( CompileShader( gs_4_0, GS_TriangleThrough() ));
        SetPixelShader( CompileShader( ps_4_0,  PS_PoissonSolver()   ));

        SetBlendState( NoBlending, float4( 0.0f, 0.0f, 0.0f, 0.0f ), 0xFFFFFFFF );                
	}
}

//--------------------------------------------------------------------------------------
// Subtract divergence
//--------------------------------------------------------------------------------------

float4 PS_SubtractGradient(PS_Pos4Tex3SliceID input) : SV_Target
{
	if (IsBoundaryCell(input.Texcoord))	return 0;

	float left = poissonSolution.SampleLevel(samPoint, LEFTCELL, 0).x;
	float right = poissonSolution.SampleLevel(samPoint, RIGHTCELL, 0).x;
	float top = poissonSolution.SampleLevel(samPoint, TOPCELL, 0).x;
	float bottom = poissonSolution.SampleLevel(samPoint, BOTTOMCELL, 0).x;
	float front = poissonSolution.SampleLevel(samPoint, FRONTCELL, 0).x;
	float back = poissonSolution.SampleLevel(samPoint, BACKCELL, 0).x;

	float4 center = poissonSolution.SampleLevel(samPoint, input.Texcoord, 0);

	// We assume the pressure gradient at boundaries are zero
	if (IsBoundaryCell(LEFTCELL))	left = center;
	if (IsBoundaryCell(RIGHTCELL))	right = center;
	if (IsBoundaryCell(TOPCELL))	top = center;
	if (IsBoundaryCell(BOTTOMCELL))	bottom = center;
	if (IsBoundaryCell(FRONTCELL))	front = center;
	if (IsBoundaryCell(BACKCELL))	back = center;

	float4 velocity = velocity.Sample(samPoint, input.Texcoord);

	return velocity - 0.5 * float4(right-left, top-bottom, front-back, 0);	
}

technique10 SubtractDivergenceTechnique
{
	pass
	{
        SetVertexShader(CompileShader( vs_4_0,  VS_Through() ));
        SetGeometryShader( CompileShader( gs_4_0, GS_TriangleThrough() ));
        SetPixelShader( CompileShader( ps_4_0,  PS_SubtractGradient()   ));

        SetBlendState( NoBlending, float4( 0.0f, 0.0f, 0.0f, 0.0f ), 0xFFFFFFFF );                
	}
}
