#include "stdafx.h"

#include "VelocityOperation2d.h"
#include "EffectManager.h"
#include "TechniqueOperation.h"
#include "HlslTextureOperation2d.h"
#include "ExplosionParameters.h"

VelocityOperation2d::VelocityOperation2d(ID3D10Device* device, TextureStore* textureStore,
                                         int velocityID, int divergenceForcesID, int velocityForcesID)
    :
    textureStore(textureStore),
    velocityID(velocityID),
    divergenceForcesID(divergenceForcesID),
	velocityForcesID(velocityForcesID),
    jacobiIterations(ExplosionParameters::Instance().GetPressureGradientJacobiIterations()),
    effectName("FluidSolver2d.fx")
{
    InitAddVelocityForcesOperation(device);
    InitAdvectionOperation(device);
    InitCalculateDivergenceOperation(device);
    InitCalculatePressureOperation(device);
    InitSubtractGradientOperation(device);

    InitEffectVariables(device);
}

VelocityOperation2d::~VelocityOperation2d()
{
}

void VelocityOperation2d::InitAddVelocityForcesOperation(ID3D10Device* device)
{
    ID3D10Effect* effect = EffectManager::Instance().GetEffect(effectName, device);
    
	addVelocityForcesTechnique = auto_ptr<TechniqueOperation>(
	    new TechniqueOperation(effect, "AddVelocityForcesTechnique", textureStore, device)
	);
	addVelocityForcesOperation = auto_ptr<HlslTextureOperation>(
	    new HlslTextureOperation2d(
	        device,
	        ExplosionParameters::Instance().GetSimulationWidth(),
	        ExplosionParameters::Instance().GetSimulationHeight(),
	        addVelocityForcesTechnique.get(), 
	        addVelocityForcesTechnique.get(), 
	        USE_DISPLACED_BORDER
	    )
	);

	addVelocityForcesTechnique->AddSetShaderResourceTask("velocity", velocityID);
	addVelocityForcesTechnique->AddSetShaderResourceTask("velocityForces", velocityForcesID);

	addVelocityForcesOperation->SetOutputTextureID(velocityID);
}

void VelocityOperation2d::InitAdvectionOperation(ID3D10Device* device)
{
    ID3D10Effect* effect = EffectManager::Instance().GetEffect(effectName, device);
    advectionTechnique = auto_ptr<TechniqueOperation>(
                                    new TechniqueOperation(effect, "AdvectionTechnique", textureStore, device));
    advectionBorderTechnique = auto_ptr<TechniqueOperation>(
                                    new TechniqueOperation(effect, "AdvectionBorderTechnique", textureStore, device));

    advectionOperation = auto_ptr<HlslTextureOperation2d>(new HlslTextureOperation2d(device,
        ExplosionParameters::Instance().GetSimulationWidth(),
        ExplosionParameters::Instance().GetSimulationHeight(),
        advectionTechnique.get(), advectionBorderTechnique.get(), USE_DISPLACED_BORDER));

    advectionTechnique->AddSetShaderResourceTask("velocity", velocityID);
    advectionTechnique->AddSetShaderResourceTask("advectionSource", velocityID);
    advectionBorderTechnique->AddSetShaderResourceTask("velocity", velocityID);
    advectionBorderTechnique->AddSetShaderResourceTask("advectionSource", velocityID);

    advectionOperation->SetOutputTextureID(velocityID);
}

void VelocityOperation2d::InitCalculateDivergenceOperation(ID3D10Device* device)
{
    divergenceID = textureStore->AddTexture("Divergence", device);

    ID3D10Effect* effect = EffectManager::Instance().GetEffect(effectName, device);
    calculateDivergenceTechnique = auto_ptr<TechniqueOperation>(
                                            new TechniqueOperation(effect, "CalculateDivergenceTechnique",
                                            textureStore, device));

    calculateDivergenceOperation = auto_ptr<HlslTextureOperation2d>(
        new HlslTextureOperation2d(
            device,
            ExplosionParameters::Instance().GetSimulationWidth(),
            ExplosionParameters::Instance().GetSimulationHeight(),
            calculateDivergenceTechnique.get(), 
            NULL, 
            IGNORE_BORDER
        )
    );

    calculateDivergenceTechnique->AddSetShaderResourceTask("velocity", velocityID);
    calculateDivergenceTechnique->AddSetShaderResourceTask("divergenceForces", divergenceForcesID);

    calculateDivergenceOperation->SetOutputTextureID(divergenceID);
}

void VelocityOperation2d::InitCalculatePressureOperation(ID3D10Device* device)
{
    pressureID = textureStore->AddTexture("Temp texture for the projection iterations", device);
    textureStore->ZeroTexture(pressureID, device);

    ID3D10Effect* effect = EffectManager::Instance().GetEffect(effectName, device);
    calculatePressureTechnique = auto_ptr<TechniqueOperation>(
        new TechniqueOperation(effect, "CalculatePressureTechnique", textureStore, device)
    );

    calculatePressureOperation = auto_ptr<HlslTextureOperation2d>(
        new HlslTextureOperation2d(
            device,
            ExplosionParameters::Instance().GetSimulationWidth(),
            ExplosionParameters::Instance().GetSimulationHeight(),
            calculatePressureTechnique.get(),
            calculatePressureTechnique.get(),
            USE_DISPLACED_BORDER
        )
    );

    calculatePressureTechnique->AddSetShaderResourceTask("pressure", pressureID);
    calculatePressureTechnique->AddSetShaderResourceTask("divergence", divergenceID);

    calculatePressureOperation->SetOutputTextureID(pressureID);
}

void VelocityOperation2d::InitSubtractGradientOperation(ID3D10Device* device)
{
    ID3D10Effect* effect = EffectManager::Instance().GetEffect(effectName, device);
    subtractGradientTechnique = auto_ptr<TechniqueOperation>(
                                    new TechniqueOperation(effect, "SubtractDivergenceTechnique", textureStore, device));

    subtractGradientOperation = auto_ptr<HlslTextureOperation2d>(new HlslTextureOperation2d(device,
        ExplosionParameters::Instance().GetSimulationWidth(),
        ExplosionParameters::Instance().GetSimulationHeight(),
        subtractGradientTechnique.get(), subtractGradientTechnique.get(), USE_DISPLACED_BORDER));

    subtractGradientTechnique->AddSetShaderResourceTask("pressure", pressureID);
    subtractGradientTechnique->AddSetShaderResourceTask("velocity", velocityID);

    subtractGradientOperation->SetOutputTextureID(velocityID);
}

void VelocityOperation2d::InitEffectVariables(ID3D10Device* device)
{
    ID3D10Effect* effect = EffectManager::Instance().GetEffect(effectName, device);
    ExplosionParameters& p = ExplosionParameters::Instance();

    float projectionAlpha[4];
    float projectionRBeta[4];

    projectionAlpha[0] = -1.0f;
    projectionAlpha[1] = -1.0f;
    projectionAlpha[2] = 0;
    projectionAlpha[3] = 0;

    projectionRBeta[0] = 1.0f / 4.0f;
    projectionRBeta[1] = 1.0f / 4.0f;
    projectionRBeta[2] = 0;
    projectionRBeta[3] = 0;

    effect->GetVariableByName("ProjectionAlpha")->AsVector()->SetFloatVector(projectionAlpha);
    effect->GetVariableByName("ProjectionRBeta")->AsVector()->SetFloatVector(projectionRBeta);
}

void VelocityOperation2d::Run(float timestep, ID3D10Device* device) {
    addVelocityForcesOperation->Execute(device);

    advectionOperation->Execute(device); 

    calculateDivergenceOperation->Execute(device); 

    textureStore->ZeroTexture(pressureID, device);
    for (int i = 0; i < jacobiIterations-1; i++) {
        calculatePressureOperation->Execute(device);
    }

    subtractGradientOperation->Execute(device);
}

void VelocityOperation2d::SetJacobiIterations(int iterationCount) {
    jacobiIterations = iterationCount;
}
