#include "stdafx.h"

#include "VelocityOperation3d.h"
#include "EffectManager.h"
#include "TechniqueOperation.h"
#include "HlslTextureOperation3d.h"
#include "ExplosionParameters.h"

VelocityOperation3d::VelocityOperation3d(ID3D10Device* device, TextureStore* textureStore,
                                     int velocityID, int thermalExpansionID, int obstaclesID )
    :
    textureStore(textureStore),
    velocityID(velocityID),
    thermalExpansionID(thermalExpansionID),
    obstaclesID(obstaclesID),
    jacobiIterations(ExplosionParameters::Instance().GetPressureGradientJacobiIterations()),
    effectName("FluidSolver3d.fx")
{
        InitAdvectionOperation(device);

        InitCalculateDivergenceOperation(device);
        InitCalculatePressureOperation(device);
        InitSubtractGradientOperation(device);

        InitEffectVariables(device);
}

VelocityOperation3d::~VelocityOperation3d()
{
}

void VelocityOperation3d::InitAdvectionOperation(ID3D10Device* device)
{
    ID3D10Effect* effect = EffectManager::Instance().GetEffect(effectName, device);
    advectionTechnique = auto_ptr<TechniqueOperation>(
                                    new TechniqueOperation(effect, "AdvectionTechnique", textureStore, device));
    //advectionBorderTechnique = auto_ptr<TechniqueOperation>(
    //                              new TechniqueOperation(effect, "AdvectionBorderTechnique", textureStore, device));
    advectionOperation = auto_ptr<HlslTextureOperation3d>(new HlslTextureOperation3d(device,
        ExplosionParameters::Instance().GetSimulationWidth(),
        ExplosionParameters::Instance().GetSimulationHeight(),
        ExplosionParameters::Instance().GetSimulationDepth(),
        advectionTechnique.get(), NULL));

    advectionTechnique->AddSetShaderResourceTask("velocity", velocityID);
    advectionTechnique->AddSetShaderResourceTask("advectionSource", velocityID);
    advectionTechnique->AddSetShaderResourceTask("obstacles", obstaclesID);
    //advectionBorderTechnique->AddSetShaderResourceTask("velocity", velocityID);
    //advectionBorderTechnique->AddSetShaderResourceTask("advectionSource", velocityID);

    advectionOperation->SetOutputTextureID(velocityID);
}

void VelocityOperation3d::InitCalculateDivergenceOperation(ID3D10Device* device)
{
    ID3D10Effect* effect = EffectManager::Instance().GetEffect(effectName, device);
    calculateDivergenceTechnique = auto_ptr<TechniqueOperation>(
                                            new TechniqueOperation(effect, "CalculateDivergenceTechnique",
                                            textureStore, device));

    calculateDivergenceOperation = auto_ptr<HlslTextureOperation3d>(new HlslTextureOperation3d(device,
        ExplosionParameters::Instance().GetSimulationWidth(),
        ExplosionParameters::Instance().GetSimulationHeight(),
        ExplosionParameters::Instance().GetSimulationDepth(),
        calculateDivergenceTechnique.get(), NULL));

    calculateDivergenceTechnique->AddSetShaderResourceTask("velocity", velocityID);
    calculateDivergenceTechnique->AddSetShaderResourceTask("thermalExpansion", thermalExpansionID);
    calculateDivergenceTechnique->AddSetShaderResourceTask("obstacles", obstaclesID);
    calculateDivergenceOperation->SetOutputTextureID(thermalExpansionID);
}

void VelocityOperation3d::InitCalculatePressureOperation(ID3D10Device* device)
{
    projectionSolutionGuessID = textureStore->AddTexture("Temp texture for the projection iterations", device);
    textureStore->ZeroTexture(projectionSolutionGuessID, device);

    ID3D10Effect* effect = EffectManager::Instance().GetEffect(effectName, device);
    calculatePressureTechnique = auto_ptr<TechniqueOperation>(
        new TechniqueOperation(effect, "CalculatePressureTechnique", textureStore, device)
    );

    calculatePressureOperation = auto_ptr<HlslTextureOperation3d>(
        new HlslTextureOperation3d(
            device,
            ExplosionParameters::Instance().GetSimulationWidth(),
            ExplosionParameters::Instance().GetSimulationHeight(),
            ExplosionParameters::Instance().GetSimulationDepth(),
            calculatePressureTechnique.get(),
            calculatePressureTechnique.get()
        )
    );

    calculatePressureTechnique->AddSetShaderResourceTask("poissonSolution", projectionSolutionGuessID);
    calculatePressureTechnique->AddSetShaderResourceTask("divergence", thermalExpansionID);
    calculatePressureTechnique->AddSetShaderResourceTask("obstacles", obstaclesID);

    calculatePressureOperation->SetOutputTextureID(projectionSolutionGuessID);
}

void VelocityOperation3d::InitSubtractGradientOperation(ID3D10Device* device)
{
    ID3D10Effect* effect = EffectManager::Instance().GetEffect(effectName, device);
    subtractGradientTechnique = auto_ptr<TechniqueOperation>(
                                    new TechniqueOperation(effect, "SubtractDivergenceTechnique", textureStore, device));

    subtractGradientOperation = auto_ptr<HlslTextureOperation3d>(new HlslTextureOperation3d(device,
        ExplosionParameters::Instance().GetSimulationWidth(),
        ExplosionParameters::Instance().GetSimulationHeight(),
        ExplosionParameters::Instance().GetSimulationDepth(),
        subtractGradientTechnique.get(), subtractGradientTechnique.get()));

    subtractGradientTechnique->AddSetShaderResourceTask("poissonSolution", projectionSolutionGuessID);
    subtractGradientTechnique->AddSetShaderResourceTask("velocity", velocityID);
    subtractGradientTechnique->AddSetShaderResourceTask("obstacles", obstaclesID);

    subtractGradientOperation->SetOutputTextureID(velocityID);
}

void VelocityOperation3d::InitEffectVariables(ID3D10Device* device)
{
    //ID3D10Effect* effect = EffectManager::Instance().GetEffect(effectName, device);
    //ExplosionParameters& p = ExplosionParameters::Instance();

    //float projectionAlpha[4];
    //float projectionRBeta[4];
    //
    //projectionAlpha[0] = -p.GetH()*p.GetH();
    //projectionAlpha[1] = -p.GetH()*p.GetH();
    //projectionAlpha[2] = 0;
    //projectionAlpha[3] = 0;

    //projectionRBeta[0] = 1.0f / 4.0f;
    //projectionRBeta[1] = 1.0f / 4.0f;
    //projectionRBeta[2] = 0;
    //projectionRBeta[3] = 0;

    //effect->GetVariableByName("ProjectionAlpha")->AsVector()->SetFloatVector(projectionAlpha);
    //effect->GetVariableByName("ProjectionRBeta")->AsVector()->SetFloatVector(projectionRBeta);
}

void VelocityOperation3d::Run(float timestep, ID3D10Device* device) {
    advectionOperation->Execute(device); // output velocity
    calculateDivergenceOperation->Execute(device); // output thermalexpansion

    textureStore->ZeroTexture(projectionSolutionGuessID, device);

    for (int i = 0; i < jacobiIterations-1; i++) {
        calculatePressureOperation->Execute(device); // output (and input) projectionSolutionGuess
    }

    subtractGradientOperation->Execute(device); // output velocity
}

void VelocityOperation3d::SetJacobiIterations(int iterationCount) {
    jacobiIterations = iterationCount;
}
