#include "stdafx.h"

#include "HlslTextureOperation2d.h"
#include "VertexBufferUtils.h"

HlslTextureOperation2d::HlslTextureOperation2d(ID3D10Device* device, int width, int height, 
										   TechniqueOperation* interiorTechnique, TechniqueOperation* borderTechnique,  
										   BorderOperation borderOperation)
:	width(width), height(height), 
	interiorTech(interiorTechnique),
	borderTech(borderTechnique),
	textureStore(NULL),
	hasTestedSwapping(false),
	interiorVB(NULL),
	borderVB(NULL),
	interiorInputLayout(NULL),
	borderInputLayout(NULL)
{
	assert(interiorTech && "HlslOperation must have an interior technique");
	assert((borderOperation == IGNORE_BORDER || borderTech)  && "Missing border technique");
	if (borderOperation == IGNORE_BORDER) borderTech = NULL;

	textureStore = interiorTech->GetTextureStore();
	
	assert((!borderTech || textureStore == borderTech->GetTextureStore()) && "The techniques used must have the same TextureStore instance" );
	
	if (interiorTech) InitInteriorVertexBuffer(device, borderOperation);
	if (borderTech) InitBorderVertexBuffer(device, borderOperation);
}

HlslTextureOperation2d::~HlslTextureOperation2d()
{
	SAFE_RELEASE(interiorVB);
	SAFE_RELEASE(borderVB);
	SAFE_RELEASE(interiorInputLayout);
	SAFE_RELEASE(borderInputLayout);
}

void HlslTextureOperation2d::Execute(ID3D10Device* device)
{
	assert(textureStore!=NULL && "You have forgotten to call SetTextureStore with a valid parameter");
	SetOutputTargets(device);

	if (interiorTech) {
		ActivateInteriorBuffer(device);
		this->interiorTech->Execute(this->interiorVertexCount, device);
	}

	if (borderTech)
	{
		ActivateBorderBuffer(device);
		this->borderTech->Execute(this->borderVertexCount, device);
	}

	textureStore->Swap();
}

void HlslTextureOperation2d::InitInteriorVertexBuffer(ID3D10Device* device, BorderOperation borderOp)
{
	float xOffset = (borderTech || borderOp == IGNORE_BORDER) ? 2.0f / this->width : 0;
	float yOffset = (borderTech || borderOp == IGNORE_BORDER) ? 2.0f / this->height : 0;

	VertexPositionTexture vertices[] = 
	{
		VertexPositionTexture(D3DXVECTOR3(-1+xOffset,-1+yOffset,0), D3DXVECTOR2(xOffset/2, 1-yOffset/2)),
		VertexPositionTexture(D3DXVECTOR3(-1+xOffset,+1-yOffset,0), D3DXVECTOR2(xOffset/2, yOffset/2)), 
		VertexPositionTexture(D3DXVECTOR3(+1-xOffset,-1+yOffset,0), D3DXVECTOR2(1-xOffset/2, 1-yOffset/2)),
		VertexPositionTexture(D3DXVECTOR3(+1-xOffset,+1-yOffset,0), D3DXVECTOR2(1-xOffset/2, yOffset/2))
	};
	this->interiorVertexCount = sizeof(vertices)/sizeof(VertexPositionTexture);	
	this->interiorVB = VertexBufferUtils::CreateDefaultVertexBuffer(device, vertices, this->interiorVertexCount);
	this->interiorPrimitiveTopology = D3D10_PRIMITIVE_TOPOLOGY_TRIANGLESTRIP;
	this->interiorInputLayout = VertexBufferUtils::CreateInputLayout<VertexPositionTexture>
		(device, interiorTech->GetFirstPass());

}

void HlslTextureOperation2d::InitBorderVertexBuffer(ID3D10Device* device, BorderOperation borderOp)
{
	float xOffset = 1.0f / this->width;
	float yOffset = 1.0f / this->height;

	if (borderOp == USE_DISPLACED_BORDER) {	
		VertexPositionTexture vertices[] = 
		{
			VertexPositionTexture(D3DXVECTOR3(-1+xOffset,-1+yOffset,0), D3DXVECTOR2(xOffset*3/2, 1-yOffset/2)),
			VertexPositionTexture(D3DXVECTOR3(-1+xOffset,+1-yOffset,0), D3DXVECTOR2(xOffset*3/2, yOffset/2)), 

			VertexPositionTexture(D3DXVECTOR3(-1+xOffset,+1-yOffset,0), D3DXVECTOR2(xOffset/2, yOffset*3/2)), 
			VertexPositionTexture(D3DXVECTOR3(+1-xOffset,+1-yOffset,0), D3DXVECTOR2(1-xOffset/2, yOffset*3/2)),

			VertexPositionTexture(D3DXVECTOR3(+1-xOffset,+1-yOffset,0), D3DXVECTOR2(1-xOffset*3/2, yOffset/2)),
			VertexPositionTexture(D3DXVECTOR3(+1-xOffset,-1+yOffset,0), D3DXVECTOR2(1-xOffset*3/2, 1-yOffset/2)),

			VertexPositionTexture(D3DXVECTOR3(+1-xOffset,-1+yOffset,0), D3DXVECTOR2(1-xOffset/2, 1-yOffset*3/2)),
			VertexPositionTexture(D3DXVECTOR3(-1+xOffset,-1+yOffset,0), D3DXVECTOR2(xOffset/2, 1-yOffset*3/2))
		};
		this->borderVertexCount = sizeof(vertices)/sizeof(VertexPositionTexture);	
		this->borderVB = VertexBufferUtils::CreateDefaultVertexBuffer(device, vertices, this->borderVertexCount);
	} else {
		VertexPositionTexture vertices[] = 
		{
			VertexPositionTexture(D3DXVECTOR3(-1+xOffset,-1+yOffset,0), D3DXVECTOR2(xOffset*3/2, 1-yOffset/2)),
			VertexPositionTexture(D3DXVECTOR3(-1+xOffset,+1-yOffset,0), D3DXVECTOR2(xOffset*3/2, yOffset/2)), 

			VertexPositionTexture(D3DXVECTOR3(-1+xOffset,+1-yOffset,0), D3DXVECTOR2(xOffset/2, yOffset*3/2)), 
			VertexPositionTexture(D3DXVECTOR3(+1-xOffset,+1-yOffset,0), D3DXVECTOR2(1-xOffset/2, yOffset*3/2)),

			VertexPositionTexture(D3DXVECTOR3(+1-xOffset,+1-yOffset,0), D3DXVECTOR2(1-xOffset*3/2, yOffset/2)),
			VertexPositionTexture(D3DXVECTOR3(+1-xOffset,-1+yOffset,0), D3DXVECTOR2(1-xOffset*3/2, 1-yOffset/2)),

			VertexPositionTexture(D3DXVECTOR3(+1-xOffset,-1+yOffset,0), D3DXVECTOR2(1-xOffset/2, 1-yOffset*3/2)),
			VertexPositionTexture(D3DXVECTOR3(-1+xOffset,-1+yOffset,0), D3DXVECTOR2(xOffset/2, 1-yOffset*3/2))
		};
		this->borderVertexCount = sizeof(vertices)/sizeof(VertexPositionTexture);	
		this->borderVB = VertexBufferUtils::CreateDefaultVertexBuffer(device, vertices, this->borderVertexCount);
	}

	this->borderPrimitiveTopology = D3D10_PRIMITIVE_TOPOLOGY_LINELIST;
	this->borderInputLayout = VertexBufferUtils::CreateInputLayout<VertexPositionTexture>
		(device, borderTech->GetFirstPass());
}

void HlslTextureOperation2d::ActivateInteriorBuffer( ID3D10Device* device )
{
	device->IASetInputLayout(this->interiorInputLayout);
	device->IASetPrimitiveTopology(this->interiorPrimitiveTopology);
	VertexBufferUtils::SetSingleVertexBuffer<VertexPositionTexture>(device, this->interiorVB);
}

void HlslTextureOperation2d::ActivateBorderBuffer( ID3D10Device* device )
{
	device->IASetInputLayout(this->borderInputLayout);
	device->IASetPrimitiveTopology(this->borderPrimitiveTopology);
	VertexBufferUtils::SetSingleVertexBuffer<VertexPositionTexture>(device, this->borderVB);
}

void HlslTextureOperation2d::ClearOutputTextures()
{
	renderTargetInfo.clear();
}

void HlslTextureOperation2d::SetOutputTargets(ID3D10Device* device)
{
	if (!this->hasTestedSwapping) {
		CheckRenderTargetSwapping();
		this->hasTestedSwapping = true;
	}
	textureStore->SetAsRenderTargets(renderTargetInfo, device);
}

void HlslTextureOperation2d::CheckRenderTargetSwapping() {
	for (unsigned int i = 0; i < renderTargetInfo.size(); ++i) {
		pair<int, bool> & info = renderTargetInfo[i];
		info.second = interiorTech->IsIDShaderResource(info.first) || (borderTech && borderTech->IsIDShaderResource(info.first));
	}
}
