#include "stdafx.h"

#include "HlslTextureOperation3d.h"
#include "VertexBufferUtils.h"
#include "VertexDefinition.h"

HlslTextureOperation3d::HlslTextureOperation3d(ID3D10Device* device, int width, int height, int depth,
						 TechniqueOperation* interiorTechnique, TechniqueOperation* borderTechnique)
	:
	width(width), 
	height(height), 
	depth(depth), 
	interiorTech(interiorTechnique), 
	borderTech(borderTechnique),
	textureStore(NULL),
	hasTestedSwapping(false),
	interiorVB(NULL),
	borderVB(NULL),
	interiorInputLayout(NULL),
	borderInputLayout(NULL)

{
	this->textureStore = interiorTechnique->GetTextureStore();
	InitInteriorVertexBuffer(device);
}

HlslTextureOperation3d::~HlslTextureOperation3d()
{
	SAFE_RELEASE(interiorVB);
	SAFE_RELEASE(borderVB);
	SAFE_RELEASE(interiorInputLayout);
	SAFE_RELEASE(borderInputLayout);
}

void HlslTextureOperation3d::InitInteriorVertexBuffer(ID3D10Device* device)
{
	this->interiorVertexCount = 6 * (this->depth-2);
	VertexPositionTexture3D* vertexData = new VertexPositionTexture3D[this->interiorVertexCount];

	int index = 0;
	for (int z = 1; z < this->depth-1; z++) {
		CreateInteriorSliceVertices(vertexData, index, z);
		index += 6;
	}

	this->interiorPrimitiveTopology = D3D10_PRIMITIVE_TOPOLOGY_TRIANGLELIST;
	this->interiorInputLayout = VertexBufferUtils::CreateInputLayout<VertexPositionTexture3D>(device, interiorTech->GetFirstPass());
	this->interiorVB = VertexBufferUtils::CreateDefaultVertexBuffer(device, vertexData, this->interiorVertexCount);
	
	delete [] vertexData;
}



void HlslTextureOperation3d::CreateInteriorSliceVertices(VertexPositionTexture3D* vertices, int startIndex, int z) 
{
	float xOffset =  2.0f / this->width;
	float yOffset = 2.0f / this->height;

	VertexPositionTexture3D bottomleft(D3DXVECTOR3(-1+xOffset,-1+yOffset,0), D3DXVECTOR3(xOffset/2, 1-yOffset/2, (float)z));
	VertexPositionTexture3D topleft(D3DXVECTOR3(-1+xOffset,+1-yOffset,0), D3DXVECTOR3(xOffset/2, yOffset/2, (float)z));
	VertexPositionTexture3D bottomright(D3DXVECTOR3(+1-xOffset,-1+yOffset,0), D3DXVECTOR3(1-xOffset/2, 1-yOffset/2, (float)z));
	VertexPositionTexture3D topright(D3DXVECTOR3(+1-xOffset,+1-yOffset,0), D3DXVECTOR3(1-xOffset/2, yOffset/2, (float)z));

	vertices[startIndex++] = bottomleft;
	vertices[startIndex++] = topleft;
	vertices[startIndex++] = bottomright;
	vertices[startIndex++] = bottomright;
	vertices[startIndex++] = topleft;
	vertices[startIndex++] = topright;
}
void HlslTextureOperation3d::ActivateInteriorBuffer( ID3D10Device* device )
{
	device->IASetInputLayout(this->interiorInputLayout);
	device->IASetPrimitiveTopology(this->interiorPrimitiveTopology);
	VertexBufferUtils::SetSingleVertexBuffer<VertexPositionTexture3D>(device, this->interiorVB);
}

void HlslTextureOperation3d::Execute(ID3D10Device* device)
{
	assert(textureStore!=NULL && "You have forgotten to call SetTextureStore with a valid parameter");
	SetOutputTargets(device);

	if (interiorTech) {
		ActivateInteriorBuffer(device);
		this->interiorTech->Execute(this->interiorVertexCount, device);
	}

	//if (borderTech)
	//{
	//	ActivateBorderBuffer(device);
	//	this->borderTech->Execute(this->borderVertexCount, device, !isInputIdentical);
	//}

	textureStore->Swap();
}

void HlslTextureOperation3d::SetOutputTargets(ID3D10Device* device)
{
	if (!this->hasTestedSwapping) {
		CheckRenderTargetSwapping();
		this->hasTestedSwapping = true;
	}
	textureStore->SetAsRenderTargets(renderTargetInfo, device);
}

void HlslTextureOperation3d::CheckRenderTargetSwapping() {
	for (unsigned int i = 0; i < renderTargetInfo.size(); ++i) {
		pair<int, bool> & info = renderTargetInfo[i];
		info.second = interiorTech->IsIDShaderResource(info.first) || (borderTech && borderTech->IsIDShaderResource(info.first));
	}
}
