#include "stdafx.h"

#include "TextureStore.h"
#include "TextureUtils.h"

#include <sstream>

TextureStore::TextureStore( int width, int height, ID3D10Device* device, int numberOfTempsNeeded )
	: width(width), height(height) 
{
	//InitTempTextures(numberOfTempsNeeded, device);
	InitializeViewPort();
}

void TextureStore::InitTempTextures(int numberOfTempsNeeded, ID3D10Device* device) 
{
	const float ClearColor[4] = {0,0,0,0};

	for (int i = 0; i < numberOfTempsNeeded; ++i) {
		int id = AddTexture("Temporary Texture", device);
		ClearRenderTarget(id, ClearColor, device);
		availableTemps.push_back(id);
	}
}

TextureStore::~TextureStore()
{
	for (unsigned int i = 0; i < textures1d.size(); ++i) {		
		Texture1DRefs& refs = textures1d[i];
		SAFE_RELEASE(refs.renderTargetView);
		SAFE_RELEASE(refs.shaderResourceView);
		SAFE_RELEASE(refs.texture);
	}
	for (unsigned int i = 0; i < resources.size(); ++i) {		
		ResourceRefs& refs = resources[i];
		SAFE_RELEASE(refs.renderTargetView);
		SAFE_RELEASE(refs.shaderResourceView);
		SAFE_RELEASE(refs.texture);
	}
}

int TextureStore::AddTexture1D(string description, ID3D10Device* device, int width, const D3DXVECTOR4* initialData)
{
	Texture1DRefs refs;
	{
		ZeroMemory(&refs, sizeof(refs));
		refs.texture = TextureUtils::CreateTexture1D(
			device, width, initialData, D3D10_BIND_SHADER_RESOURCE
		);
		refs.desc = description;
	}

	device->CreateShaderResourceView(refs.texture, NULL, &refs.shaderResourceView);

	this->textures1d.push_back(refs);
	return (int)(textures1d.size()-1);
}

void TextureStore::SetAsRenderTarget(int id, bool isSwapNeeded, ID3D10Device* device)
{
	vector< pair<int, bool> > renderTargets;
	renderTargets.push_back(pair<int, bool>(id, isSwapNeeded));

	SetAsRenderTargets(renderTargets, device);
}

void TextureStore::Swap()
{
    while (!tempsInUseAs.empty()) {
        pair<int,int> tempIsUseAs = tempsInUseAs.front();
        tempsInUseAs.pop_front();

        const ResourceRefs ref = resources[tempIsUseAs.first];
        resources[tempIsUseAs.first] = resources[tempIsUseAs.second];
        resources[tempIsUseAs.second] = ref;

        availableTemps.push_back(tempIsUseAs.first);
    }
}

void TextureStore::SetAsShaderResource1D(int id, ID3D10EffectShaderResourceVariable* variable)
{
	variable->SetResource(this->textures1d[id].shaderResourceView);
}

void TextureStore::SetAsShaderResource(int id, ID3D10EffectShaderResourceVariable* variable)
{
	variable->SetResource(this->resources[id].shaderResourceView);
}

void TextureStore::ClearRenderTarget(int id, const float RGBA[], ID3D10Device* device)
{
	device->ClearRenderTargetView(resources[id].renderTargetView, RGBA);
}

ID3D10Texture1D* TextureStore::GetTexture1D(int id)
{
	return textures1d[id].texture;
}

ID3D10Resource* TextureStore::GetResource(int id)
{
	return resources[id].texture;
}

//ID3D10Texture2D* TextureStore::GetTexture2D(int id)
//{
//	return textures2d[id].texture;
//}

ID3D10ShaderResourceView* TextureStore::GetShaderResourceView1D(int id)
{
	return textures1d[id].shaderResourceView;
}

ID3D10ShaderResourceView* TextureStore::GetShaderResourceView(int id)
{
	return resources[id].shaderResourceView;
}

void TextureStore::InitializeViewPort()
{
	this->viewPort.Width = this->width;
	this->viewPort.Height = this->height;
	this->viewPort.MaxDepth = 1.0;
	this->viewPort.MinDepth = 0.0;
	this->viewPort.TopLeftX = 0;
	this->viewPort.TopLeftY = 0;
}

void TextureStore::SetAsRenderTargets(const vector< pair<int, bool> >& targets, ID3D10Device* device)
{
	static ID3D10RenderTargetView *renderTargetViews[8];

	for (int i = 0; i < 8; i++) renderTargetViews[i] = 0;
	
	assert(tempsInUseAs.size() == 0 && "Illegal SetAsRenderTargets. A Call to Swap() is missing");
	
	for (unsigned int i = 0; i < targets.size(); ++i) {
		const pair<int,bool> & target = targets[i];
		int u;
		if (!target.second) {
			u = target.first;
			renderTargetViews[i] = resources[target.first].renderTargetView;
		} else {
			assert(availableTemps.size() && "Not enough temp textures available");
	
			int tempId = availableTemps.front();
			availableTemps.pop_front();

			tempsInUseAs.push_back(pair<int,int>(tempId, target.first));
			renderTargetViews[i] = resources[tempId].renderTargetView;
			u = tempId;
		}
		//string s = textures[u].desc;
		//stringstream ss;
		//ss << "Render target bound at " << i << " " << s << "\r\n";
		//s = ss.str();
		//OutputDebugStringA(s.c_str());
	}
	device->OMSetRenderTargets(8, renderTargetViews, NULL);
	//device->OMSetRenderTargets((UINT)targets.size(), renderTargetViews, NULL);
	device->RSSetViewports(1,&this->viewPort);		

	//ID3D10RenderTargetView *actualRenderTargetViews[8];
	//device->OMGetRenderTargets(8,actualRenderTargetViews, NULL);

}

void TextureStore::ZeroTexture(int id, ID3D10Device* device)
{
	static float col[] = {0,0,0,0};
	ID3D10RenderTargetView* RTV = resources[id].renderTargetView;
	device->ClearRenderTargetView(RTV, col);
}

void TextureStore::ManualSwap(int id1, int id2)
{
	assert(tempsInUseAs.size() == 0 && "Illegal ManualSwap. There are temps in use.");

	const ResourceRefs ref = resources[id1];
	resources[id1] = resources[id2];
	resources[id2] = ref;
}

//ID3D10Resource* TextureStore::GetStagingTexture(ID3D10Device* device) 
//{
//	if (stagingTexture == NULL) {
//		stagingTexture = TextureUtils::CreateSimpleTexture2D(
//			device, 
//			width, 
//			height, 
//			0, D3D10_USAGE_STAGING, D3D10_CPU_ACCESS_READ);	
//	}
//	return stagingTexture;
//}
