#include "stdafx.h"

#include "MainFrame.h"
#include <algorithm>

MainFrame::MainFrame() : shouldShutDown(false), shutDownErrorCode(0)
{
	//OutputDebugString(L"MainFrame\n");
	MainFrameInitializer initializer;
	initializer.InitializeMainFrame(*this);
}
MainFrame::~MainFrame()
{
	//OutputDebugString(L"~MainFrame\n");

	for (std::list<ReferencedObject*>::iterator it = this->referencedObjects.begin(); it != this->referencedObjects.end(); ++it) {
		ReferencedObject* o = *it;
		SAFE_DELETE(*it);
	}
}

HRESULT MainFrame::OnDeviceCreated(ID3D10Device* pd3dDevice, const DXGI_SURFACE_DESC *pBufferSurfaceDesc) 
{ 
	//OutputDebugString(L"MainFrame::OnDeviceCreated\n");
	
	std::list<DrawableGameComponent*>::iterator it = this->drawableGameComponents.begin();
	while (it != this->drawableGameComponents.end()) {
		try
		{
			DrawableGameComponent* drawableComponent = *it;
			++it;
			bool hr = drawableComponent->LoadGeneralGraphicsContent(pd3dDevice, pBufferSurfaceDesc);
			if (!hr) return DXUTERR_CREATINGDEVICEOBJECTS;
		}
		catch (const std::exception& e)
		{
			MessageBoxA( NULL, e.what(), "Exception", MB_OK );
			return DXUTERR_CREATINGDEVICEOBJECTS;
		}
	}

	return S_OK;
}
void MainFrame::OnDeviceDestroyed() 
{
	//OutputDebugString(L"MainFrame::OnDeviceDestroyed\n");

	std::list<DrawableGameComponent*>::iterator it = this->drawableGameComponents.begin();
	while (it != this->drawableGameComponents.end()) {
		DrawableGameComponent* drawableComponent = *it;
		++it;
		drawableComponent->ReleaseGeneralGraphicsContent();
	}
}

HRESULT MainFrame::OnSwapChainResized(ID3D10Device* pd3dDevice, IDXGISwapChain *pSwapChain, const DXGI_SURFACE_DESC* pBufferSurfaceDesc) 
{
	//OutputDebugString(L"MainFrame::OnSwapChainResized\n");

	std::list<DrawableGameComponent*>::iterator it = this->drawableGameComponents.begin();;
	while (it != this->drawableGameComponents.end()) {
		DrawableGameComponent* drawableComponent = *it;
		++it;
		drawableComponent->LoadBackBufferDependantGraphicsContent(pd3dDevice, pSwapChain, pBufferSurfaceDesc);
	}

	return S_OK;
}
void MainFrame::OnSwapChainReleasing()
{
	//OutputDebugString(L"MainFrame::OnSwapChainReleasing\n");

	std::list<DrawableGameComponent*>::iterator it = this->drawableGameComponents.begin();
	while (it != this->drawableGameComponents.end()) {
		DrawableGameComponent* drawableComponent = *it;
		++it;
		drawableComponent->ReleaseBackBufferDependantGraphicsContent();
	}
}

void MainFrame::OnFrameUpdate( double time, float elapsedTime )
{
	//OutputDebugString(L"MainFrame::OnFrameUpdate\n");
	if (shouldShutDown) { 
		DXUTShutdown(shutDownErrorCode);
		return;
	}
	
	std::list<GameComponent*>::iterator it = this->gameComponents.begin();
	while (it != this->gameComponents.end()) {
		GameComponent* gameComponent = *it;
		++it;
		gameComponent->Update(time, elapsedTime);
	}
	ClearDeadObjects();
}

void MainFrame::OnFrameRender(ID3D10Device* pd3dDevice, double time, float elapsedTime) 
{
	//OutputDebugString(L"MainFrame::OnFrameRender\n");
	if (shouldShutDown) { 
		DXUTShutdown(shutDownErrorCode);
		return;
	}
	
	std::list<DrawableGameComponent*>::iterator it = this->drawableGameComponents.begin();
	while (it != this->drawableGameComponents.end()) {
		DrawableGameComponent* drawableGameComponent = *it;
		++it;
		if (drawableGameComponent->IsActive())
			drawableGameComponent->Draw(pd3dDevice, time, elapsedTime);
	}
	ClearDeadObjects();
}

void MainFrame::OnKeyDown(UINT nChar, bool keyDown, bool altDown) 
{
	std::list<IKeyHandler*>::iterator it;
	for (it = keyHandlers.begin(); it != keyHandlers.end(); ++it)
	{
		(*it)->HandleKeyEvent(nChar, keyDown, altDown);
	}
}

void MainFrame::OnMouseEvent() 
{
	//OutputDebugString(L"MainFrame::OnMouseEvent\n");
}

bool gameCompSortOp (GameComponent* one, GameComponent* two) {
	return one->GetUpdateOrder() < two->GetUpdateOrder();
}

bool drawableGameCompSortOp (DrawableGameComponent* one, DrawableGameComponent* two) {
	return one->GetDrawOrder() < two->GetDrawOrder();
}

void MainFrame::AddComponent(GameComponent* component) 
{
	if (component == NULL) return;	
	
	DrawableGameComponent* drawableComponent = dynamic_cast<DrawableGameComponent*>(component);
	if (drawableComponent != NULL) {
		this->drawableGameComponents.push_back(drawableComponent);
	}
	
	this->gameComponents.push_back(component);

	this->AddObjectReference(component);

	this->SortComponents();
}

void MainFrame::SortComponents() {
	drawableGameComponents.sort(drawableGameCompSortOp);
	gameComponents.sort(gameCompSortOp);
}

void MainFrame::AddMessageHandler( MessageHandler* handler )
{
	this->messageHandlers.push_back(handler);
	this->AddObjectReference(handler);
}

void MainFrame::AddKeyHandler(IKeyHandler* handler)
{
	this->keyHandlers.push_back(handler);
	this->AddObjectReference(handler);
}


void MainFrame::MsgProc( HWND hWnd, UINT uMsg, WPARAM wParam, LPARAM lParam )
{
	std::list<MessageHandler*>::iterator it;
	for (it = messageHandlers.begin(); it != messageHandlers.end(); ++it)
	{
		bool noFurtherProcessing = (*it)->HandleMessage(hWnd, uMsg, wParam, lParam);
		if (noFurtherProcessing) return;
	}
}

void MainFrame::AddObjectReference( ReferencedObject* objectToBeResponsibleOf )
{
	if (find( this->referencedObjects.begin(), 
			  this->referencedObjects.end(), 
			  objectToBeResponsibleOf) == referencedObjects.end())
	{
		this->referencedObjects.push_back(objectToBeResponsibleOf);
	}	
}

void MainFrame::ShutDown( int errorCode /*= 0*/ )
{
	shouldShutDown = true;
	shutDownErrorCode = errorCode;
}

void MainFrame::RemoveComponent( GameComponent* component )
{
	DrawableGameComponent* drawableComponent = dynamic_cast<DrawableGameComponent*>(component);
	if (drawableComponent != NULL) drawableGameComponents.remove(drawableComponent);
	
	gameComponents.remove(component);
	referencedObjects.remove(component);
	deadObjects.push_back(component);	
}

void MainFrame::ClearDeadObjects()
{
	while (!deadObjects.empty()) {
		ReferencedObject* obj = deadObjects.front();
		deadObjects.pop_front();
		
		obj->CleanUp();
		delete obj;
	}
}