using System;
using System.Collections.Generic;
using System.Text;
using SimpleFFTWWrapper;
using KarlsTools;

namespace KolmogorovTurbulenceGenerator
{
    abstract class Generator
    {
        protected double epsilon;
        protected double k_intertial;
        protected double sigma;

        protected Generator(double epsilon, double k_intertial, double sigma)
        {
            this.epsilon = epsilon;
            this.k_intertial = k_intertial;
            this.sigma = sigma;
        }

        public abstract void GenerateTurbulenceTexture(string filename, bool viewFriendly);
    }
    
    class Generator2D : Generator
    {
        private int width;
        private int height;
        
        public Generator2D(int width, int height, double epsilon, double k_intertial, double sigma) : 
            base(epsilon, k_intertial, sigma)
        {
            this.width = width;
            this.height = height;
        }

        public override void GenerateTurbulenceTexture(string filename, bool viewFriendly)
        {
            FFTWComplex2D transformX = new FFTWComplex2D(this.width, this.height, FFTTransformType.BACKWARD);
            FFTWComplex2D transformY = new FFTWComplex2D(this.width, this.height, FFTTransformType.BACKWARD);

            this.InitFrequencies(transformX, transformY);

            transformX.transform();
            transformY.transform();

            Texture2DSaver texture = this.CreateTexture(transformX, transformY, viewFriendly);
            texture.SaveTexture(filename);
        }


        private void InitFrequencies(FFTWComplex2D transformX, FFTWComplex2D transformY)
        {
            KolmogorovFrequencyCalculator k = new KolmogorovFrequencyCalculator(this.epsilon, this.k_intertial, this.sigma);
            transformX.zero();
            transformY.zero();

            for (int x = 0; x <= transformX.Nx / 2; x++)
            {
                for (int y = 0; y <= transformX.Ny / 2; y++)
                {
                    Complex u1, u2, u3;

                    int NminX = (transformX.Nx - x) % transformX.Nx;
                    int NminY = (transformX.Ny - y) % transformX.Ny;

                    k.calculateU(x, y, 0, 0, out u1, out u2, out u3);
                    transformX.setInput(x, y, u1.Real(), u1.Imag());
                    transformX.setInput(NminX, NminY, u1.Real(), -u1.Imag());
                    transformY.setInput(x, y, u2.Real(), u2.Imag());
                    transformY.setInput(NminX, NminY, u2.Real(), -u2.Imag());

                    k.calculateU(NminX, y, 0, 0, out u1, out u2, out u3);
                    transformX.setInput(NminX, y, u1.Real(), u1.Imag());
                    transformX.setInput(x, NminY, u1.Real(), -u1.Imag());
                    transformY.setInput(NminX, y, u2.Real(), u2.Imag());
                    transformY.setInput(x, NminY, u2.Real(), -u2.Imag());
                }
            }

            int[] tabX = new int[] { 0, transformX.Nx / 2 };
            int[] tabY = new int[] { 0, transformX.Ny / 2 };
            foreach (int x in tabX)
            {
                foreach (int y in tabY)
                {
                    transformX.setInput(x, y, transformX.getRealInput(x, y), 0);
                    transformY.setInput(x, y, transformY.getRealInput(x, y), 0);
                }
            }
        }

        private Texture2DSaver CreateTexture(FFTWComplex2D transformX, FFTWComplex2D transformY, bool viewFriendly)
        {
            Texture2DSaver texture = new Texture2DSaver(this.width, this.height);

            float maxAbs = 0.0f;
            for (int x = 0; x < texture.Width; x++)
            {
                for (int y = 0; y < texture.Height; y++)
                {
                    maxAbs = Math.Max(maxAbs, Math.Abs((float)transformX.getRealOutput(x, y)));
                    maxAbs = Math.Max(maxAbs, Math.Abs((float)transformY.getRealOutput(x, y)));
                }
            }

            for (int x = 0; x < texture.Width; x++)
            {
                for (int y = 0; y < texture.Height; y++)
                {
                    if (Math.Abs(transformX.getImagOutput(x, y)) > 1e-16 ||
                        Math.Abs(transformY.getImagOutput(x, y)) > 1e-16)
                    {
                        throw new Exception("The output contains non-real elements. This should not happen!");
                    }
                    
                    float r = (float) transformX.getRealOutput(x,y);
                    float g = (float) transformY.getRealOutput(x, y);
                    r /= maxAbs;
                    g /= maxAbs;

                    if (viewFriendly)
                    {
                        r = (r + 1.0f) / 2.0f;
                        g = (g + 1.0f) / 2.0f;
                        texture.SetValues(x, y, r, r, r);
                    } 
                    else 
                    {
                        texture.SetValues(x, y, r, g, 0);                        
                    }
                }
            }
            return texture;
        }
    }

    class Generator3D : Generator
    {
        private int width;
        private int height;
        private int depth;

        public Generator3D(int width, int height, int depth, double epsilon, double k_intertial, double sigma)
            :
            base(epsilon, k_intertial, sigma)
        {
            this.width = width;
            this.height = height;
            this.depth = depth;
        }

        public override void GenerateTurbulenceTexture(string filename, bool viewFriendly)
        {
            FFTWComplex3D transformX = new FFTWComplex3D(this.width, this.height, this.depth, FFTTransformType.BACKWARD);
            FFTWComplex3D transformY = new FFTWComplex3D(this.width, this.height, this.depth, FFTTransformType.BACKWARD);
            FFTWComplex3D transformZ = new FFTWComplex3D(this.width, this.height, this.depth, FFTTransformType.BACKWARD);

            this.InitFrequencies(transformX, transformY, transformZ);

            transformX.transform();
            transformY.transform();
            transformZ.transform();

            Texture3DSaver texture = this.CreateTexture(transformX, transformY, transformZ, viewFriendly);
            texture.SaveTexture(filename);
        }


        private void InitFrequencies(FFTWComplex3D transformX, FFTWComplex3D transformY, FFTWComplex3D transformZ)
        {
            KolmogorovFrequencyCalculator k = new KolmogorovFrequencyCalculator(this.epsilon, this.k_intertial, this.sigma);
            transformX.zero();
            transformY.zero();
            transformZ.zero();

            for (int x = 0; x <= transformX.Nx / 2; x++)
            {
                for (int y = 0; y <= transformX.Ny / 2; y++)
                {
                    for (int z = 0; z <= transformX.Nz / 2; z++)
                    {
                        Complex u1, u2, u3;

                        int NminX = (transformX.Nx - x) % transformX.Nx;
                        int NminY = (transformX.Ny - y) % transformX.Ny;
                        int NminZ = (transformX.Nz - z) % transformX.Nz;

                        // +++ and ---
                        k.calculateU(x, y, z, 0, out u1, out u2, out u3);
                        
                        transformX.setInput(x, y, z, u1.Real(), u1.Imag());
                        transformX.setInput(NminX, NminY, NminZ, u1.Real(), -u1.Imag());
                        
                        transformY.setInput(x, y, z, u2.Real(), u2.Imag());
                        transformY.setInput(NminX, NminY, NminZ, u2.Real(), -u2.Imag());

                        transformZ.setInput(x, y, z, u3.Real(), u3.Imag());
                        transformZ.setInput(NminX, NminY, NminZ, u3.Real(), -u3.Imag());
                        
                        // ++- and --+
                        k.calculateU(x, y, NminZ, 0, out u1, out u2, out u3);
                        transformX.setInput(x, y, NminZ, u1.Real(), u1.Imag());
                        transformX.setInput(NminX, NminY, z, u1.Real(), -u1.Imag());

                        transformY.setInput(x, y, NminZ, u2.Real(), u2.Imag());
                        transformY.setInput(NminX, NminY, z, u2.Real(), -u2.Imag());

                        transformZ.setInput(x, y, NminZ, u3.Real(), u3.Imag());
                        transformZ.setInput(NminX, NminY, z, u3.Real(), -u3.Imag());

                        // +-+ and -+-
                        k.calculateU(x, NminY, z, 0, out u1, out u2, out u3);
                        transformX.setInput(x, NminY, z, u1.Real(), u1.Imag());
                        transformX.setInput(NminX, y, NminZ, u1.Real(), -u1.Imag());

                        transformY.setInput(x, NminY, z, u2.Real(), u2.Imag());
                        transformY.setInput(NminX, y, NminZ, u2.Real(), -u2.Imag());

                        transformZ.setInput(x, NminY, z, u3.Real(), u3.Imag());
                        transformZ.setInput(NminX, y, NminZ, u3.Real(), -u3.Imag());

                        // +-- and -++
                        k.calculateU(x, NminY, NminZ, 0, out u1, out u2, out u3);
                        transformX.setInput(x, NminY, NminZ, u1.Real(), u1.Imag());
                        transformX.setInput(NminX, y, z, u1.Real(), -u1.Imag());

                        transformY.setInput(x, NminY, NminZ, u2.Real(), u2.Imag());
                        transformY.setInput(NminX, y, z, u2.Real(), -u2.Imag());

                        transformZ.setInput(x, NminY, NminZ, u3.Real(), u3.Imag());
                        transformZ.setInput(NminX, y, z, u3.Real(), -u3.Imag());
                    }
                }
            }

            int[] tabX = new int[] { 0, transformX.Nx / 2 };
            int[] tabY = new int[] { 0, transformX.Ny / 2 };
            int[] tabZ = new int[] { 0, transformX.Nz / 2 };
            foreach (int x in tabX)
            {
                foreach (int y in tabY)
                {
                    foreach (int z in tabZ)
                    {
                        transformX.setInput(x, y, z, transformX.getRealInput(x, y, z), 0);
                        transformY.setInput(x, y, z, transformY.getRealInput(x, y, z), 0);
                        transformZ.setInput(x, y, z, transformZ.getRealInput(x, y, z), 0);
                    }
                }
            }
        }

        private Texture3DSaver CreateTexture(FFTWComplex3D transformX, FFTWComplex3D transformY, FFTWComplex3D transformZ, bool viewFriendly)
        {
            Texture3DSaver texture = new Texture3DSaver(this.width, this.height, this.depth);

            float maxAbs = 0.0f;
            for (int x = 0; x < texture.Width; x++)
            {
                for (int y = 0; y < texture.Height; y++)
                {
                    for (int z = 0; z < texture.Depth; z++) 
                    {
                        maxAbs = Math.Max(maxAbs, Math.Abs((float)transformX.getRealOutput(x, y, z)));
                        maxAbs = Math.Max(maxAbs, Math.Abs((float)transformY.getRealOutput(x, y, z)));
                        maxAbs = Math.Max(maxAbs, Math.Abs((float)transformZ.getRealOutput(x, y, z)));
                    }
                }
            }

            for (int x = 0; x < texture.Width; x++)
            {
                for (int y = 0; y < texture.Height; y++)
                {
                    for (int z = 0; z < texture.Depth; z++)
                    {
                        if (Math.Abs(transformX.getImagOutput(x, y, z)) > 1e-16 ||
                            Math.Abs(transformY.getImagOutput(x, y, z)) > 1e-16 ||
                            Math.Abs(transformZ.getImagOutput(x, y, z)) > 1e-16)
                        {
                            throw new Exception("The output contains non-real elements. This should not happen!");
                        }

                        float r = (float)transformX.getRealOutput(x, y, z);
                        float g = (float)transformY.getRealOutput(x, y, z);
                        float b = (float)transformZ.getRealOutput(x, y, z);
                        r /= maxAbs;
                        g /= maxAbs;
                        b /= maxAbs;

                        if (viewFriendly)
                        {
                            r = (r + 1.0f) / 2.0f;
                            g = (g + 1.0f) / 2.0f;
                            b = (b + 1.0f) / 2.0f;
                            texture.SetValues(x, y, z, r, r, r);
                        }
                        else
                        {
                            texture.SetValues(x, y, z, r, g, b);
                        }
                    }
                }
            }
            return texture;
        }
    }
}
