using System;
using System.Collections.Generic;
using System.Text;

using KarlsTools;

namespace KolmogorovTurbulenceGenerator
{
    class KolmogorovFrequencyCalculator
    {
        private double epsilon; 
        private double k_intertial; 
        private double sigma;
        
        private Random random = new Random();

        public KolmogorovFrequencyCalculator(double epsilon, double k_intertial, double sigma)
        {
            this.epsilon = epsilon;
            this.k_intertial = k_intertial;
            this.sigma = sigma;
        }

        public void calculateU(int i, int j, int k, int l, out Complex U1, out Complex U2, out Complex U3)
        {
            Complex X1 = calculateX();
            Complex X2 = calculateX();
            Complex X3 = calculateX();

            U1 = calculateH11(i, j, k, l) * X1;
            U2 = calculateH21(i, j, k, l) * X1 + calculateH22(i, j, k, l) * X2;
            U3 = calculateH31(i, j, k, l) * X1 + calculateH32(i, j, k, l) * X2 + calculateH33(i, j, k, l) * X3;
        }

        private Complex calculateX()
        {
            double r = randomGaussSample();
            double theta = randomUniformSample();

            return new Complex(Math.Cos(2 * Math.PI * theta), Math.Sin(2 * Math.PI * theta)) * r;

        }

        private bool useLast = false;
        private double last;
        private double randomGaussSample()
        {
            if (useLast)
            {
                useLast = false;
                return last;
            }
            else
            {
                double x1,x2,w;
                do 
                {
                    x1 = 2.0 * random.NextDouble() - 1.0;
                    x2 = 2.0 * random.NextDouble() - 1.0;
                    w = x1 * x1 + x2 * x2;
                } while(w >= 1.0);

                w = Math.Sqrt((-2.0 * Math.Log(w)) / w);
                useLast = true;
                last = x1 * w;
                return x2 * w;
            }
        }

        private double randomUniformSample()
        {
            return random.NextDouble();
        }


        private Complex calculateH11(int i, int j, int k, int l)
        {
            return Complex.Sqrt(new Complex(phi(i, j, k, l, 1, 1), 0));
        }

        private Complex calculateH21(int i, int j, int k, int l)
        {
            Complex divisor = calculateH11(i, j, k, l);
            if (Complex.Abs(divisor) == 0.0)
            {
                return divisor;
            }
            else
            {
                return Complex.Divide(new Complex(phi(i, j, k, l, 2, 1), 0), divisor);
            }
        }

        private Complex calculateH31(int i, int j, int k, int l)
        {
            Complex divisor = calculateH11(i, j, k, l);
            if (Complex.Abs(divisor) == 0.0)
            {
                return divisor;
            }
            else
            {
                return Complex.Divide(new Complex(phi(i, j, k, l, 3, 1), 0), divisor);
            }
        }

        private Complex calculateH22(int i, int j, int k, int l)
        {
            Complex h21 = calculateH21(i, j, k, l);

            return Complex.Sqrt(new Complex(phi(i, j, k, l, 2, 2), 0) - h21 * h21);
        }

        private Complex calculateH32(int i, int j, int k, int l)
        {
            Complex divisor = calculateH22(i, j, k, l);
            if (Complex.Abs(divisor) == 0.0)
            {
                return divisor;
            }
            else
            {
                Complex phi32 = new Complex(phi(i, j, k, l, 3, 2), 0);
                return Complex.Divide(phi32 - calculateH31(i, j, k, l) * calculateH21(i, j, k, l), divisor);
            }
        }

        private Complex calculateH33(int i, int j, int k, int l)
        {
            Complex h31 = calculateH31(i, j, k, l);
            Complex h32 = calculateH32(i, j, k, l);
            Complex phi33 = new Complex(phi(i, j, k, l, 3, 3), 0);
            return Complex.Sqrt(phi33 - h31 * h31 - h32 * h32);
        }

        private double phi(int i, int j, int k, int l, int a, int b) 
        {
            double k_len = Math.Sqrt(i * i + j * j + k * k);
            if (k_len < this.k_intertial)
            {
                return 0.0;
            }
            else
            {
                double E = calculateE(k_len, l);
                double fraction = E / (4 * Math.PI * Math.Pow(k_len, 4));

                double[] k_tab = new double[] { i, j, k };
                double multiplier = (a == b ? k_len * k_len : 0) - k_tab[a - 1] * k_tab[b - 1];
                return fraction * multiplier;
            }
        }

        private double calculateE(double k, int l)
        {
            if (k < this.k_intertial)
            {
                return 0.0;
            }
            else
            {
                return calculateKolmogorovEnergy(k) * calculateGauss(k, l);
            }
        }

        private double calculateKolmogorovEnergy(double k)
        {
            if (k < this.k_intertial)
            {
                return 0.0;
            }
            else
            {
                return 1.5 * Math.Pow(this.epsilon, 3.0/2.0) * Math.Pow(k, -5.0 / 2.0);
            }
        }

        private double calculateGauss(double k, int l)
        {
            return (1.0 / (Math.Sqrt(2 * Math.PI) * k * this.sigma)) * 
                Math.Pow(Math.E, -l * l / (2 * k * k * this.sigma * this.sigma));
        }

    }
}
